#lang racket/base

(module interactions racket/base (provide (prefix-out configuration: language-version))
  (define language-version "Language version: 2019-07-05."))

(module ui racket/base (provide (prefix-out configuration: (combine-out (all-defined-out)
                                                                        language-name
                                                                        language-settings-label)))

  (require (submod "english.rkt" ui)
           (only-in "snack/button.rkt" view read-bitmap)
           racket/local)

  (define-values (view:help
                  view:shrink
                  view:line-comment
                  view:insert-image
                  view:explode
                  view:format
                  view:hug
                  view:definitions
                  view:layout
                  view:record
                  view:interactions
                  view:history-up
                  view:history-down)
    
    (local [(define (icon-view label name position)
              (view label (read-bitmap (collection-file-path (string-append name "-icon.png")
                                                             "csc104" "media"))
                    (+ position 50)))]
      
      (values (icon-view message:help                 "help" (- 102 50))
              (icon-view message:shrink             "shrink" 3.5)
              (icon-view message:line-comment      "comment"  1)
              (icon-view message:insert-image "insert-image"  3)
              (icon-view message:explode            "explode" 2)
              (icon-view message:format           "reindent" 1.5)
              (icon-view message:hug                   "hug"  4)
              (icon-view message:definitions             "D"  5)
              (icon-view message:layout             "layout"  6)
              (icon-view message:record             "record" 6.5)
              (icon-view message:interactions            "I"  7)
              (icon-view message:history:up       "previous"  8)
              (icon-view message:history:down         "next"  9)))))


(module step-display racket/base (require (submod "snack/contract.rkt" core))
  
  (provide (all-from-out (submod "english.rkt" step-display))
           (all-from-out (submod "shared-settings.rkt" step))
           (contract-out [a-space-width (→ positive-natural/c)]
                         [sane-minimum-space-width positive-natural/c]
                         [typical-font-aspect-ratio positive-real/c]
                         [maximum-steps      positive-natural/c]
                         [maximum-step-lines positive-natural/c]
                         [maximum-step-width positive-natural/c]))

  (require (prefix-in message: (submod "english.rkt" step-display))
           (only-in (submod "shared-settings.rkt" step)
                    preference:display:spacing:compact
                    preference:interaction:no-wait)
           (only-in framework editor:get-current-preferred-font-size))

  (define-values (typical-font-aspect-ratio
                  sane-minimum-space-width)
    (values 0.6
            8))
  
  (define (a-space-width)
    (local-require (only-in racket/math exact-floor))
    (max sane-minimum-space-width (exact-floor (* (editor:get-current-preferred-font-size)
                                                  typical-font-aspect-ratio))))
  
  (define-values (maximum-steps
                  maximum-step-lines
                  maximum-step-width)
    (values 50
            20
            (- 102 (string-length "#;{}")))))
