#lang racket/base

(provide compthink-eval)

(require (only-in racket/sandbox call-in-sandbox-context sandbox-output sandbox-error-output)
         racket/pretty (only-in file/convertible convertible?))

(define (set-up-printing)
  
  (print-boolean-long-form #true)
  ;
  (pretty-print-show-inexactness #true)
  (pretty-print-exact-as-decimal #true)
  ;
  (print-as-expression #false) ; Remove initial quote for lists in errors.
  (define (prefix-list l) (if (list? l) (cons 'list (map prefix-list l)) l))
  ; This might not need the ‘1’, wrote it before (print-as-expression #false).
  (current-print (λ (v) (unless (void? v) (pretty-print (prefix-list v) (current-output-port) 1))))

  (error-value->string-handler (let ([old (error-value->string-handler)])
                                 (λ (v n) (old (prefix-list v) n))))

  ; Indicate we handle images ourselves, which is by ‘write-special’.
  (pretty-print-size-hook
   (let ([prev (pretty-print-size-hook)])
     ; Non-#false return triggers ‘pretty-print-print-hook’.
     (λ (v w? op) (if (convertible? v) 1 (prev v w? op)))))
  (pretty-print-print-hook
   (let ([prev (pretty-print-print-hook)])
     (λ (v w? op) (if (convertible? v) (write-special v op) (prev v w? op))))))

; Produce a new evaluator with each def evaluated.
; ‘make-img-eval’ has an implicit parameter for pure image results: ‘extra-margin’.
(define-syntax-rule (compthink-eval def ...)
  (let ()
    (define me (make-img-eval))
    (for-each me '((require racket/pretty)
                   (current-print pretty-print-handler)
                   (pretty-print-columns 65)
                   def
                   ...))
    (for-each (λ (a-thunk) (call-in-sandbox-context me a-thunk))
              (list (λ () (error-print-source-location #f))
                    (λ () (sandbox-output 'string))
                    (λ () (sandbox-error-output 'string))
                    set-up-printing
                    (λ () (namespace-require 'csc104/language/without-gui))))
    me))

(module img-eval racket/base ; teachpack/2htdp/scribblings/img-eval
  (provide make-img-eval set-eval-handler-to-tweak-images)
  
  #| make-img-eval : -> (any/c . -> . any)
     Calls @racket[make-base-eval] to create an evaluator, and then initializes it with
      @racket[set-eval-handler-to-tweak-images].

     set-eval-handler-to-tweak-images : (any/c . -> . any) -> void?

     Loads @racketmodname[2htdp/image] and @racketmodname[lang/posn] into the given evaluator
      and then sets its eval-handler to increase the bounding box for 2htdp/image images by 1.
  
     It does this only for images that the result of evaluation directly (not for example,
      when lists that contain images are the result of evaluation). |#

  (require scribble/eval)

  (define (make-img-eval)
    (define img-eval (make-base-eval))
    (set-eval-handler-to-tweak-images img-eval)
    img-eval)

  (define (set-eval-handler-to-tweak-images img-eval)
    (interaction-eval #:eval img-eval (require 2htdp/image))
    (interaction-eval #:eval img-eval (require lang/posn))
    (img-eval '(define extra-margin (make-parameter 0)))
    (img-eval `(let ([ce (current-eval)])
                 (define (adjust-image exp i)
                   (if (image? i) (let ([em (extra-margin)])
                                    (overlay/xy i 
                                                (- em) (- em)
                                                (rectangle (+ (image-width i) 1 em em)
                                                           (+ (image-height i) 1 em em)
                                                           'solid (color 255 0 0 0))))
                       i))
                 (current-eval (λ (exp) (adjust-image exp (ce exp))))))))

(require 'img-eval)

