#lang racket/base (require racket/require
                           (submod "../snack/contract.rkt" core)
                           (submod "../snack/syntax.rkt" core)
                           "../snack/boolean.rkt"
                           (except-in "../snack/false.rkt" λ-case♯)
                           "../snack/match.rkt"
                           (submod "../snack/match.rkt" data)
                           "../snack/function.rkt"
                           (submod "../snack/list.rkt" core)
                           (only-in "../snack/list.rkt" filter-not)
                           "../snack/lambda.rkt"
                           racket/splicing racket/local)
(provide (contract-out [space-width (parameter/c positive-natural/c)]
                       [chunks (any/c natural/c . → . chunked/c)]))

(require (only-in "private/common.rkt" new next); only pattern matches.
         (submod "token.rkt" layout)
         (multi-in "private/layout" ["strings-and-images.rkt"]))

(define-macro (define/c . parts) (define . parts))

(define chunked/c (listof (listof token/c)))
(struct marked () #:transparent)
(define layout/c (or/c token/c marked?))
#;
(contracts [force? (parameter/c boolean?)]
           [(chunks any/c natural/c) → (♯ chunked/c)]
           [(markup any/c natural/c) → (♯ layout/c)]
           [(marked→chunked layout/c [(listof token/c)]) →∗ chunked/c]
           [(number-of-lines layout/c) → natural/c])

(define/c (chunks definition/expression maximum-width)
  (app♯ (∘ simplify-chunks marked→chunked) (markup definition/expression maximum-width)))

(define/c force? (make-parameter #true))

(define (simplify-chunks chunks)
  (define non-trivial (map (λ• (filter-not (λ-♯match `"") •)) chunks))
  (map nextify non-trivial))
(define (nextify tokens)
  (match tokens
    [`(next-start ,t next-end . ,rest) (list* (dereference t) (nextify rest))]
    [`(next-start ,t          . ,rest) (list* (next-start t)  (nextify rest))]
    [`(,t   next-end          . ,rest) (list* (next-end t)    (nextify rest))]
    [`(,t . ,rest)                     (list* t (nextify rest))]
    [`() `()]))

(define-syntax-parser struct/contract
  [(_ (name super) ([id contract] ...)) #'(struct name super [id ...])]
  [(_ name ([id contract] ...)) #'(struct name [id ...])])
(struct/contract (stack   marked) ([markeds (listof layout/c)]))
(struct/contract (flow    marked) ([marked₀ layout/c #;unstacked]
                                   [marked₁ layout/c]))
(struct/contract (delimit marked) ([open (or/c "(" "[" " " 'next-start)]
                                   [marked layout/c]
                                   [close (or/c ")" "]" "" 'next-end)]))
(define/c (marked→chunked marked [column (list)])
  (define ⟲ (λ• (marked→chunked • column)))
  (match marked
    [(flow marked₀ marked₁) (with-data*
                                ([(column ...) column]
                                 [((line₀ ...)) (⟲ marked₀)]
                                 [((line₁ ...)
                                   lines ...) (marked→chunked marked₁ (data (column ... line₀ ...)))])
                              (data ((line₀ ... line₁ ...)
                                     lines ...)))]
    [(delimit open marked close) (with-data ([(lines ...
                                               (line ...)) (⟲ (flow open marked))] [close close])
                                   (data (lines ...
                                          (line ... close))))]
    [(stack markeds) (with-data* ([(marked₀
                                    markeds ...) markeds]
                                  [(line
                                    ...) (⟲ (data marked₀))]
                                  [(((line′ ...)
                                     ...)
                                    ...) (map (λ• (marked→chunked • (list))) (data (markeds
                                                                                    ...)))]
                                  [(column ...) (invisible (filter-not symbol? column))])
                       (data (line
                              ...
                              (column ... line′ ...)
                              ...
                              ...)))]
    [string/image (with-data ([string/image string/image]) (data ((string/image))))]))

(define/c number-of-lines (∘ length marked→chunked))

(module layout racket/base (provide (all-defined-out))
  (require (submod "../snack/syntax.rkt" syntax-classes/core))
  (define-syntax-class Block-layout #:attributes (name (e 1))
    #:datum-literals (block) (pattern (name:block ~! e ...)))
  (define-syntax-class Fun-layout #:attributes (header body)
    #:datum-literals (fun) (pattern ((~and (_:fun ~! ~rest _) header) body)))
  (define-syntax-class If-layout #:attributes (name (clause 1))
    #:datum-literals (if) (pattern (name:if ~! clause ...)))
  (define-syntax-class For-layout #:attributes (name (clause 1))
    #:datum-literals (accumulate) (pattern (name:accumulate ~! clause ...)))
  (define-syntax-class Define-layout #:attributes (name variable? bind body)
    #:datum-literals (define)
    (pattern (name:define ~! (~and bind (~or* (~and _:id (~bind (variable? #true))) _))
                          body)))
  (define-syntax-class With-layout #:attributes (name with? (definition 1) body)
    #:datum-literals (with local)
    (pattern (~or* (name:with  ~! (~bind (with?  #true))  definition ...+ body) 
                   (name:local ~! (~bind (with? #false)) [definition ...] body))))
  (define-syntax-class For-Clause-layout #:attributes (name with? (part 1))
    #:datum-literals (each with)
    (pattern (~or* (name:with id:id ~! (~bind (with?  #true)) ~rest parts) 
                   (name:each id:id ~! (~bind (with? #false)) ~rest parts))
             #:with (part ...) #'(id . parts))))
(require 'layout)

(define/c (markup definition/expression maximum-width)
  #;(contracts [|{{⋯{| (parameter/c strings-and-images/c)]
               [|}⋯}}| (parameter/c natural/c)]
               [≡? (parameter/c (or/c .5 0) (or/c 1 .5 0))]
               [(fit-atomic string-or-image/c) → (♯ string-or-image/c)])
  (define/c |{{⋯{| (make-parameter (list)))
  (define/c |}⋯}}| (make-parameter 0))
  (define/c ≡? (make-parameter 1))
  ; Base case: does a line fit?
  (define/c (fit-atomic atom)
    (and (or (force?) ((|| (list atom)) . <= . (maximum-width . - . (|| (|{{⋯{|)) (|}⋯}}|))))
         atom))
  (define-macro (try marker:expr marker′:expr) (or (parameterize ([force? #false]) marker) marker′))
  (splicing-local [(define-macro (surround open:str marker:expr close:str)
                     (app♯ delimit open (parameterize ([|{{⋯{| (list*-right (|{{⋯{|) open)]
                                                       [|}⋯}}| (+ (string-length close) (|}⋯}}|))])
                                          marker) close))]
    (define-macro (⦅ marker:expr (~datum ⦆)) (surround "(" marker ")"))
    (define-macro (〚 marker:expr (~datum 〛)) (surround "[" marker "]"))
    (define-macro (␣ marker:expr)             (surround " " marker "")))
  (define-macro (=₎ #:by-convention? by-convention? marker:expr marker′:expr)
    (let*♯ ([_ (>= (≡?) (if by-convention? .5 1))] [marked (parameterize ([|}⋯}}| 0]) marker)]
                                                   [marked′ marker′]
                                                   [#:result (stack (list marked marked′))])))
  (splicing-local [(define-macro (try∗ marker:expr marker′:expr)
                     (let* ([m (try marker #false)]
                            [m′ marker′])
                       (if (m . and . (m′ . ⇒ . (> (number-of-lines m′) (number-of-lines m))))
                           m m′)))]
    
    (define-macro (-≈₎/=₎ marker (~optional (~and #:indent-under indent-under)
                                            #:defaults [(indent-under #'#false)])
                          marker′:expr)
      (try∗ (-≈₎ marker marker′) (=₎ #:by-convention? #false
                                     marker (if 'indent-under (␣ marker′) marker′)))))
  (define-macro (-≈₎/-_₎ marker:expr marker′:expr) (-≈₎/=₎ marker #:indent-under marker′))
  (define-macro (=_₎ marker:expr marker′:expr) (=₎ #:by-convention? #true marker (␣ marker′)))
  (define-macro (-≈₎ marker:expr marker′:expr)
    (let*♯ ([marked (parameterize ([≡? 0]) marker)])
           (app♯ flow marked (␣ (parameterize ([|{{⋯{| (append (|{{⋯{|) (flat-line marked))])
                                  marker′)))))
  (define/match₁ flat-line
    [(flow marked marked′) (append (flat-line marked) (flat-line marked′))]
    [(delimit 'next-start marked 'next-end) (flat-line marked)]
    [(delimit open marked close) (*list* open (flat-line marked) close)]
    [string/image (list string/image)])

  (local [(struct/contract if-clause ([condition-result (list/c any/c any/c)]))]
    (let ⟲ ([definition/expression/clause definition/expression])
      ; Recursion on dynamic number of expressions, need to be inside ⟲ [did I mean have ⟲ in scope?]
      (define (≡₎ #:by-convention? by-convention? es)
        (match es
          ['() (fit-atomic "")]
          [`(,e) (⟲ e)]
          [`(,es ... ,e) (=₎ #:by-convention? by-convention?
                             (app♯ stack (map♯ ⟲ es)) (⟲ e))]))
      (define-macro (---₎/-≈₎/=₎ es:expr)
        (local [(define/match₁ ---₎
                  [`(,e) (parameterize ([≡? (min .5 (≡?))]) (⟲ e))]
                  [`(,e . ,es) (-≈₎ (⟲ e) (---₎ es))])]
          (try (---₎ es) (-≈₎/=₎ (⟲ (first es)) (≡₎ #:by-convention? #false (rest es))))))
      (match definition/expression/clause
        [(new  d/e/c) (⟲ d/e/c)]
        [(next d/e/c) (app♯ delimit 'next-start (⟲ d/e/c) 'next-end)]
        [(? list?)
         (define-macro (⟲′ pattern:id) (⟲ (D pattern)))
         (data-parse
          definition/expression/clause
          
          [e:For-Clause-layout
           (〚 (if (attribute e.with?)
                  (-≈₎/-_₎ (⟲′ e.name) (≡₎ #:by-convention? #true (D (e.part ...))))
                  (---₎/-≈₎/=₎ definition/expression/clause)) 〛)]
          
          [e (⦅ (data-parse
                 (D e)
                            
                 [(~and b:Block-layout (_)) (⟲′ b.name)] ; could be handled by (f) presumably
                 [b:Block-layout   (-≈₎/-_₎ (⟲′ b.name) (≡₎ #:by-convention? #true
                                                            (D (b.e ...))))]
                            
                 [e:For-layout (-≈₎/-_₎ (⟲′ e.name) (≡₎ #:by-convention? #true
                                                        (D (e.clause ...))))]
                 [e:With-layout
                  (=_₎ (-≈₎/-_₎ (⟲′ e.name)
                                (if (attribute e.with?)
                                    (≡₎ #:by-convention? #true
                                        (D (e.definition ...)))
                                    (〚 (≡₎ #:by-convention? #true
                                           (D (e.definition ...))) 〛)))
                       (⟲′ e.body))]
                            
                 [e:Define-layout (try (and (attribute e.variable?)
                                            (-≈₎ (⟲′ e.name) (-≈₎ (⟲′ e.bind) (⟲′ e.body))))
                                       (=_₎ (-≈₎/-_₎ (⟲′ e.name) (⟲′ e.bind))
                                            (⟲′ e.body)))]
                            
                 [e:Fun-layout (=_₎ (⟲′ e.header)
                                    (␣ (⟲′ e.body)))]
                            
                 [e:If-layout (-≈₎/-_₎ (⟲′ e.name) (≡₎ #:by-convention? #true
                                                       (map if-clause (D (e.clause ...)))))]
                            
                 [(f) (⟲′ f)]
                            
                 [_ (---₎/-≈₎/=₎ definition/expression/clause)]) ⦆)])]
        
        [(if-clause condition-result) (〚 (---₎/-≈₎/=₎ condition-result) 〛)]
        
        [atom
         (define (de-syntax v) (if (syntax? v) (syntax-e v) v))
         (fit-atomic (atom→token (de-syntax atom)))]))))
