#lang racket/base (require (submod "../snack/contract.rkt" core)
                           (submod "../snack/list.rkt" core)
                           "../snack/functional.rkt"
                           "../snack/conditional.rkt")

(module* step-form #false (provide (struct-out wrapped)
                                   (contract-out [token/c predicate/c]
                                                 [next-start? predicate/c]
                                                 [next-end? predicate/c]
                                                 [dereference? predicate/c]
                                                 [number-token? (any/c . → . any/c)]
                                                 [boolean-token? predicate/c]
                                                 [string-token? predicate/c]
                                                 [box-token? predicate/c]
                                                 [void-token? predicate/c]
                                                 [identifier-token? predicate/c]
                                                 [delimiter-token? predicate/c]
                                                 [image-token? predicate/c])
           
                                   freeze rectangle image-width image-height overlay color above))

(module* layout #false (provide (contract-out [atom→token ((not/c list?) . → . (or/c string? image?))]
                                              [token/c predicate/c]
                                              [dereference (identifier-token? . → . dereference?)]
                                              [next-start  (delimiter-token? . → . next-start?)]
                                              [next-end    (delimiter-token? . → . next-end?)])
                                image-width image?))

(require (only-in (combine-in 2htdp/private/image-more mrlib/image-core)
                  freeze image? image-width rectangle image-height overlay color above))

(define-values
  (boolean-token? number-token? string-token? delimiter-token? image-token? box-token? void-token?
                  #;space-token?)
  (values (λ• (member? • '("#true" "#false"))) 
          (∧ string? string->number) 
          (∧ string? (λ• (regexp-match-exact? "\".*\"" •))) 
          (λ• (member? • '("(" "[" ")" "]")))
          image?
          Box?
          void?
          #;(∨ (∧ string? (λ• (equal? "" (string-trim •))))
               (∧ image?  (λ• (zero? (image-height •)))))))

(define identifier-token?
  (∧ string? (¬ (∨ delimiter-token? boolean-token? string-token? number-token?))))

(define-syntax struct/contract
  (syntax-rules ()
    [(_ (name super) ([id contract] ...) option ...) (struct name super [id ...] option ...)]
    [(_ name ([id contract] ...) option ...) (struct name [id ...] option ...)]))

(struct/contract wrapped ([content (or/c delimiter-token? identifier-token?)]) #:transparent)
(struct/contract (dereference wrapped) () #:transparent)
(struct/contract (next-start  wrapped) () #:transparent)
(struct/contract (next-end    wrapped) () #:transparent)
(require (only-in "../language/box.rkt" Box?))
(define token/c (or/c string? image? 'next-start 'next-end wrapped?))

(define (atom→token atom) (type-case atom
                            [string?  (string-append "\"" atom "\"")]
                            [boolean? (if atom "#true" "#false")]
                            [symbol?  (symbol->string atom)]
                            [number?  (string-append (if (inexact? atom) "#i" "")
                                                     (number->string atom))]
                            [Box? (format "~v" atom)]
                            [void? "#void"]
                            [#:assert image? atom]))
