#lang racket/base (provide tool@)

(require (only-in (combine-in "private/language-ui.rkt"
                              "private/comments-tool.rkt"
                              "private/repl-link-tool.rkt"
                              "private/keybindings-tool.rkt")
                  compthink-ui@
                  compthink-comments@
                  compthink-link@
                  compthink-keybindings@))

#;(require (lib "2htdp/universe.rkt")
           (prefix-in #%%: (lib "csc104/language/without-gui.rkt")))

(require (only-in "shared-settings.rkt" suppress-type-prefix))

(require (only-in lang/private/continuation-mark-key teaching-languages-continuation-mark-key)
         (only-in lang/htdp-langs-save-file-prefix htdp-save-file-prefix)
         (only-in "run-teaching-program.rkt" create-empty-module expand-teaching-program)
         (only-in "private/htdp-stepper-debugger.rkt"
                  dummy-htdp-language-mixin debugger-language-settings-mixin)
         (prefix-in et: errortrace/stacktrace))

(require (submod "test-display.rkt" langs))

(require (only-in "private/settings-panel.rkt" setup-config-panel)
         (only-in "private/structs.rkt" define-settings-struct))

(require "private/preferences.rkt")

(require (only-in "rewrite-error.rkt" exn→rewritten-message)
         (submod "private/print.rkt" langs))

(require "snack/string-constants.rkt"
         (only-in racket/gui/base current-eventspace queue-callback style-delta%)
         (for-syntax racket/base))

(require racket/class
         (only-in racket/list second) (only-in racket/string string-join) racket/splicing)

(define test:render-value/format:width 40)

(require racket/unit drracket/tool)
(define-unit tool@ (import drracket:tool^) (export drracket:tool-exports^)
  (invoke-unit compthink-keybindings@ (import drracket:tool^))
  (invoke-unit compthink-link@        (import drracket:tool^))
  (invoke-unit compthink-ui@          (import drracket:tool^))
  (invoke-unit compthink-comments@    (import drracket:tool^))

  (define drs-eventspace (current-eventspace))

  (define (phase1) (void))
  
  (define (phase2) (set-our-preferences!) ; Set some preferences if first time ever loading the tool.
    
    ; Compthink language DrRacket registration.
    (define compthink-language%
      (debugger-language-settings-mixin
       ((drracket:language:get-default-mixin)
        (language-extension
         (drracket:language:module-based-language->language-mixin
          (module-based-language-extension
           (drracket:language:simple-module-based-language->module-based-language-mixin
            (dummy-htdp-language-mixin
             (class drracket:language:simple-module-based-language% (init-field reader-module)
               (super-new))))))))))
    (define compthink-language (instantiate compthink-language% ()
                                 ; Leaving in for some flexibility:
                                 [reader-module '(lib "2018-fall-reader.rkt" "csc104")] ; save/load
                                 ; All the drracket:language:simple-module-based-language% inits.
                                 [language-position (list (string-constant teaching-languages)
                                                          "University of Toronto" "CSC 104")]
                                 [language-numbers '(-500 -501 1)]
                                 [one-line-summary "CSC 104 2018 Fall"]
                                 [language-id "csc104:2018-fall"]
                                 [language-url "http://www.cs.toronto.edu/~csc104h"]
                                 [reader read-syntax]
                                 [module '(lib "csc104/2018-fall.rkt")]))
    (drracket:language-configuration:add-language ; Register for DrRacket's “Choose Language”.
     #:allow-executable-creation? #false compthink-language))
  (splicing-local [; Default settings
                   (define case-sensitive   #true)
                   (define printing-style   'print #;'constructor)
                   (define fraction-style   'repeating-decimal)
                   (define show-sharing     #false)
                   (define insert-newlines  #true)
                   (define annotations      'none)]
    (define-settings-struct compthink-lang-settings drracket:language:simple-settings
      ([prefix-types? #false])
      case-sensitive printing-style fraction-style show-sharing insert-newlines annotations))

  ; Change default settings, config panel, set more parameters during `on-execute', override printing.
  (define (module-based-language-extension super%)
    (class* super% ()
      
      (define/override (default-settings) (make-compthink-lang-settings #false))
      (define/override (default-settings?   s) (not (compthink-lang-settings-prefix-types? s)))
      (define/override (marshall-settings   s) (compthink-lang-settings->hash s))
      (define/override (unmarshall-settings m) (cond [(hash? m) (hash->compthink-lang-settings m)]
                                                     [else (default-settings)]))
      
      ; Set up “Show Details” pane in “Language > Choose Language...” for “CSC 104”.
      (define/override (config-panel parent)
        (define get/set (setup-config-panel parent))
        (case-lambda [() (apply make-compthink-lang-settings (get/set))]
                     [(settings) (get/set (compthink-lang-settings-prefix-types? settings))]))

      (define/override (on-execute settings run-in-user-thread)
        (define drs-namespace (current-namespace))
        (define scheme-test-module-name ((current-module-name-resolver)
                                         '(lib "test-engine/racket-tests.ss") #false #false #true))
        #;(define universe-module-name ((current-module-name-resolver)
                                        '(lib "2htdp/universe.rkt") #f #f #t))
        #;(define w/o-gui-name ((current-module-name-resolver)
                                '(lib "csc104/language/without-gui.rkt") #f #f #t))
        (define tests-on? (preferences:get 'test-engine:enable?))
        (run-in-user-thread
         (λ ()
           (when (getenv "PLTDRHTDPNOCOMPILED") (use-compiled-file-paths '()))

           (read-decimal-as-inexact #false)
           (read-accept-dot         #false)
           (read-accept-infix-dot   #false)
           (read-accept-quasiquote  #false)
           (read-accept-bar-quote   #false)
           
           (current-eval (add-annotation #false (current-eval)))

           ; To activate custom indent hook.
           (port-count-lines-enabled #true) (port-count-lines! (current-output-port))
           
           (error-display-handler compthink-error-display-handler)
           (error-value->string-handler (λ (x y) (compthink-error-value->string settings x y)))
           (error-print-source-location #false)

           #;(namespace-attach-module drs-namespace universe-module-name)
           #;(namespace-attach-module drs-namespace w/o-gui-name)

           (namespace-attach-module drs-namespace scheme-test-module-name)
           (namespace-require scheme-test-module-name)
           ;; hack: the test-engine code knows about the test~object name; we do, too
           (namespace-set-variable-value! 'test~object (build-test-engine))
           (scheme-test-data (list (drracket:rep:current-rep) drs-eventspace test-display%))
           (test-execute tests-on?) (test-silence #false)
           (test-format
            (make-formatter
             (λ (v o) (render-value/format v settings o test:render-value/format:width))))))
        
        (super on-execute settings run-in-user-thread)

        (run-in-user-thread ; Set global-port-print-handler after superclass since super sets it too.
         (λ () (define our-print-setup-parameters (drracket:language:make-setup-printing-parameters))
           (global-port-print-handler
            (λ (value port [depth 0])
              (compthink-render our-print-setup-parameters value settings port 'infinity))))))

      #| PRINTING VALUES |#
      ; For DrRacket languages.
      ; “This method is used to print values into a port, for display to a user.
      ;  The final argument is a maximum width to use (in characters) when formatting the value.
      ;  This method is expected to format the value by inserting newlines in appropriate places
      ;  and is expected to render a newline after the value.”
      ; We also give it to the tester.
      (define/override (render-value/format v settings port width)
        (compthink-render drracket:language:setup-printing-parameters v settings port width))
      (define/override (render-value v settings port)
        (compthink-render drracket:language:setup-printing-parameters v settings port 'infinity))
      ; How our settings take effect: used in the two methods that everything ends up delegating to.
      (define/private (set-printing-parameters settings thunk)
        (define (print-types?) (and (compthink-lang-settings-prefix-types? settings)
                                    (not (suppress-type-prefix))))
        (compthink:set-printing-parameters print-types? thunk))
      ; For racket/base's ‘error-value->string-handler’, set above in ‘on-execute’.
      ; Related: (error-display-handler compthink-error-display-handler) above,
      ;           implemented below in errortrace unit.
      (define/private (compthink-error-value->string settings v len)
        (parameterize ([suppress-type-prefix #true])
          (define sp (open-output-string))
          (set-printing-parameters settings (λ () (print v sp)))
          (flush-output sp) ; is this necessary?
          (define s (get-output-string sp))
          (truncate s len)))
      ; Main value rendering, for previous three methods, and ‘global-print-handler’ in ‘on-execute’.
      ; Indirectly via ‘print’ in the case of ‘compthink-error-value->string’, although I'm not sure
      ;  exactly what the indirection accomplishes and even if it could be harmful now.
      ; Not sure about the relationship with ‘compthink-error-display-handler’.
      (define/private (compthink-render setup-printing-parameters value settings port width)
        (define value₁ (prefix-list value))
        (setup-printing-parameters ;; set drracket's printing parameters
         (λ () (set-printing-parameters
                settings ;; then adjust the settings for the teaching languages
                (λ () (call-with-values
                       (λ () (drracket:language:simple-module-based-language-convert-value
                              value₁ settings))
                       (λ (converted-value [write? #true])
                         (define pretty-out (if write? pretty-write pretty-print))
                         (cond [(drracket:language:simple-settings-insert-newlines settings)
                                (if (number? width)
                                    (parameterize ([pretty-print-columns  width])
                                      (pretty-out converted-value port))
                                    (pretty-out converted-value port))]
                               [else (parameterize ([pretty-print-columns 'infinity])
                                       (pretty-out converted-value port))
                                     (newline port)]))))))
         settings width))

      (super-new)))

  (define (language-extension %)
    (class % (inherit get-module default-settings #;get-htdp-style-delta get-reader)
      (inherit-field reader-module)

      ; Alter capabilities: help context, and dock test menu.
      (define/augment (capability-value key)
        (case key
          [(drscheme:help-context-term) "L:csc104/2018-fall"]
          [(tests:test-menu tests:dock-menu) #true]
          [else (inner (drracket:language:get-capability-default key) capability-value key)]))
      
      ; Reading, inclduding metadata handling.
      (define/override (get-reader-module) reader-module)
      (define/override (get-metadata modname settings)
        (string-append (string-join htdp-save-file-prefix "\n" #:after-last "\n")
                       (format "#reader~s~s\n"
                               reader-module
                               `((modname ,modname)
                                 (compthink-settings ,(compthink-lang-settings->hash settings))))))
      (define/override (get-metadata-lines) 3)
      (define/override (metadata->settings metadata)
        (define (massage-metadata metadata)
          (if (and (list? metadata) (andmap (λ (x) (and (pair? x) (symbol? (car x)))) metadata))
              metadata
              '()))
        (define (metadata->table metadata)
          (with-handlers ([exn:fail:read? (λ (x) #false)])
            (define p (open-input-string metadata)) ; Skip to reader, skip module, then read metadata.
            (regexp-match #rx"\n#reader" p) (read p) (read p)))
        (define table (massage-metadata (metadata->table metadata)))
        (define s (assoc 'compthink-settings table))
        (if (and s (hash? (second s))) (hash->compthink-lang-settings (second s)) (default-settings)))

      ; Dummy program effect for a new tab that defaults csc104, or before run of opened program.
      ; Not sure how the wrapping occurs [test object, etc].
      (define/override (first-opened settings)
        (define-values (mod name) (create-empty-module (get-module)))
        (eval mod) (dynamic-require `',name #f) (current-namespace (module->namespace `',name))) 

      ; Interactions Header.
      (define the-style-delta (make-object style-delta%))
      (define/override (get-style-delta) the-style-delta)
      (define/override (extra-repl-information settings port)
        (parameterize ([current-output-port port])
          (display-extra-repl (drracket:rep:get-welcome-delta)
                              (drracket:rep:get-dark-green-delta)
                              (compthink-lang-settings-prefix-types? settings))))
        
      ; Read and wrap Definitions or Interactions code.
      ; • Definitions. Not affected by any of our settings.
      (define/override (front-end/complete-program port settings)
        (expand-teaching-program port (get-reader) (get-module)))
      ; • Interactions. HtDP wraps it with a reset of the test window and then a call to test,
      ;    because test reruns all recorded tests and appends the results.
      (define/override (front-end/interaction port settings)
        (define-values (t start? done?)
          (values (super front-end/interaction port settings) #true #false))
        (λ () (cond [start? (set! start? #false) #'(#%plain-app reset-tests)]
                    [done? eof]
                    [else (define ans (parameterize ([read-accept-lang #false]) (t)))
                          (cond [(eof-object? ans) (set! done? #true) #`(test)]
                                [else ans])])))

      (super-new)))

  #| errortrace |#

  ; compthink-languages-error-display-handler : (string (union any/c exn) → void) → string exn → void
  (define (compthink-error-display-handler msg exn)
    (test-silence #true)
    (if (exn? exn)
        (display (exn→rewritten-message exn) (current-error-port))
        (eprintf "uncaught exception: ~e" exn))
    (eprintf "\n")
    ; So error annotations inserted in next line aren't erased if this output is after insertion.
    (flush-output (current-error-port))
    (let ([rep (drracket:rep:current-rep)])
      (when (and (is-a? rep drracket:rep:text<%>) (eq? (send rep get-err-port) (current-error-port)))
        (define to-highlight
          (cond [(exn:srclocs? exn) ((exn:srclocs-accessor exn) exn)]
                [(exn? exn) (define cms (continuation-mark-set->list
                                         (exn-continuation-marks exn)
                                         teaching-languages-continuation-mark-key))
                            (or (for/first ([mark (or cms '())]
                                            #:when (and mark (or (path?   (car mark))
                                                                 (symbol? (car mark)))))
                                  (list (apply make-srcloc mark)))
                                '())]
                [else '()]))
        (parameterize ([current-eventspace drs-eventspace])
          (queue-callback ; Need to make sure user's eventspace still the same and still running here?
           (λ () (send rep highlight-errors to-highlight #false)))))))
  ;
  ; Member of stacktrace-imports^.  syntax syntax exact-nonnegative-integer → syntax
  ; Make sure marks associated with teaching-languages-continuation-mark-key are debug-source type.
  (splicing-local [(define base-phase (variable-reference->module-base-phase (#%variable-reference)))]
    (define (with-mark source-stx expr phase)
      (let ([source (syntax-source source-stx)]
            [line (syntax-line source-stx)] [col (syntax-column source-stx)]
            [start-position (syntax-position source-stx)] [span (syntax-span source-stx)])
        (if (and (or (symbol? source) (path? source)) (number? start-position) (number? span))
            (with-syntax ([expr expr]
                          [mark (list source line col start-position span)]
                          [teaching-languages-continuation-mark-key
                           teaching-languages-continuation-mark-key]
                          [wcm (syntax-shift-phase-level #'with-continuation-mark
                                                         (- phase base-phase))]
                          [quot (syntax-shift-phase-level #'quote (- phase base-phase))])
              #`(wcm (quot teaching-languages-continuation-mark-key) (quot mark) expr))
            expr))))

  #| Disable profiling from ‘stacktrace-imports^’ from ‘stacktrace@’. |#
  (define-values (profiling-enabled
                  profile-key initialize-profile-point register-profile-start register-profile-done)
    (values (λ () #false)
            (gensym) void (λ _ #false) void))
  
  #| Test Coverage. Disabled for now. |#
  (define-values (test-coverage-enabled initialize-test-coverage-point test-covered)
    (values (make-parameter #false) void (λ _ #false)))
  
  (define-values/invoke-unit et:stacktrace@
    (import et:stacktrace-imports^) (export (prefix et: et:stacktrace^)))
  
  #| Add debugging and test coverage information to sexp and call oe. |#
  (define ((add-annotation tracing? oe) exp) ; boolean (sexp → value) → sexp → value
    (oe (if (compiled-expression? (if (syntax? exp) (syntax-e exp) exp))
            exp
            (let ([et-annotated (et:annotate-top (expand exp) (namespace-base-phase))])
              (if tracing? (drracket:tracing:annotate (expand et-annotated)) et-annotated))))))
