#lang racket/base (provide compthink:quote
                           compthink:#%top-interaction
                           compthink:#%top compthink:#%app
                           compthink:fun compthink:define compthink:with compthink:local
                           compthink:check compthink:assuming
                           compthink:if compthink:else compthink:and compthink:or
                           compthink:time
                           compthink:for compthink:for/list)

(provide (for-syntax close
                     callable fun-shape define-name-shape
                     or/and-shape if-shape else-shape
                     with/local-shape for-form)
         with-local-function)

#| Runtime |#

(require (only-in "current-definitions.rkt" add-definition) ; Shared state.
         (submod "../functions/custom.rkt" forms)
         (only-in test-engine/racket-tests check-expect))

#| Expansion Time |#

(require "../../snack/syntax.rkt" 
         (for-syntax "../../snack/syntax.rkt" (submod "../../snack/syntax.rkt" syntax-classes/core)))

(begin-for-syntax

  (define-macro (§ template) (syntax/loc this-syntax template))
  (define-macro (§∘ {~optional loc} template) (quasisyntax/loc {~? loc this-syntax} template))

  (require "syntax-error-context.rkt") ; Shared state for syntax errors.
  (define form-name (syntax-parser [(~is id) this-syntax] [(head . _) (form-name (α head))]))

  (define-macro (parser/⊥ [part ...+] ...+ (~optional (~seq #:arity ⌢)))
    (λ (stx) (parameterize [(context stx)]
               (local-require (submod "syntax-error.rkt" miscellaneous))
               (syntax-parse stx [part ...] ...
                 (~? [(_ . :sequence) (⌢ (α length))])
                 [_ (⌢:bare/weird)]))))  

  (require racket/local "../../snack/boolean.rkt" "../../snack/function.rkt") ; More specific imports.
  
  ; Some standard definitions.
  (define-syntax-class sequence #:attributes (length) (ρ (e ...) #:attr length (length (α e))))
  (require "common.rkt"))

(define-syntax-parser define-syntax-parser/⊥
  [(_ id:id [part ...+] ...+
      (~optional (~seq #:arity ⌢))
      (~optional (~seq #:error-module module:expr)))
   (§ (define-syntax id (let () (~? (local-require module))
                          (parser/⊥ [part ...] ... (~? (~@ #:arity ⌢))))))]
  [(_ (id:id . pattern) template directive ...)
   (§ (define-syntax-parser/⊥ id [(_ . pattern) template] directive ... ))])

#| Punctuation |#

(define-syntax-parser/⊥ (compthink:quote . parts)
  (local [(define form (form-name this-syntax))]
    (local-require (submod "syntax-error.rkt" miscellaneous))
    (if (= (syntax-span form) 1) (raise-syntax-error '|' | (message:unusable-character) form)
        (§ (base:quote . parts)))))

#| #%top |#

(define-syntax-parser compthink:#%top
  [(_ . id) (cond [(identifier-binding   (α id)) (§ (base:#%top . id))]
                  [(syntax-source-module (α id)) (§ (base:#%app values (inward id)))]
                  [else                          (§ (base:#%app top:checked-identifier #'id))])])
(define-syntax-parser inward
  [(_ id) (local-require (submod "syntax-error.rkt" miscellaneous))
          (if (identifier-binding (α id)) #'id (parameterize ([context (α id)]) (⌢:top:unbound)))])

#| #%top-interaction |#

(define-syntax-parser compthink:#%top-interaction
  [(_ (~and our-step (~or* (~datum step) (~datum steps))) . parts)
   (§ (base:#%top-interaction our-step #:interactions . parts))]
  [(_ . parts) (§ (base:#%top-interaction . parts))])

(begin-for-syntax

  ; forms.rkt : parameters parameter body
  ; interpreter/patterns.rkt : name parameter body
  (define-syntax-class fun-shape #:attributes (name parameters (parameter 1) body)
    (ρ :binder+body #:when (and (α named-function?) (binding? (α name) #'compthink:fun))))

  ; forms.rkt : all attributes
  ; interpreter/patterns.rkt : name parameters
  (define-syntax-class define-name-shape
    #:attributes (binder name named-variable? named-function? parameters (parameter 1) body)
    (ρ ((~binding? #'compthink:define) . :binder+body) #:when (α named?)))

  (define-syntax-class or/and-shape #:attributes (name neu (condition 1))
    (ρ ((~or* (~and (~binding? name  #'compthink:or) (~bind (neu #false)))
              (~and (~binding? name #'compthink:and) (~bind (neu  #true))))
        condition ...)))

  (define-syntax-class else-shape (ρ (~binding? #'compthink:else)))

  (define-syntax-class if-shape
    #:attributes (name clauses (condition 1) (result 1) else-name else-result)
    (ρ ((~binding? name #'compthink:if)
        ~rest (~and clauses ([(~and (~is-not else-shape) condition:expr) result:expr]
                             ...+
                             [else-name:else-shape else-result:expr])))))

  
  ; forms.rkt: all attributes
  ; Always succeeds for  (_ . _) .
  (define-syntax-class binder+body #:attributes (binder name named? named-variable? named-function?
                                                        parameters (parameter 1)
                                                        body-parts body)
    (ρ ((~and binder (~or* (name . parameters) name)) . body-parts)
       #:attr named? (identifier? (α name))
       #:attr named-variable? (identifier? (α binder))
       #:attr named-function? (and (α named?) (α parameters))
       #:with (~maybe (parameter ...)) (α parameters)
       #:with (~maybe (body)) (α body-parts)))

  ; Fully valid separated from larger context.
  (define-syntax-class binder-body #:attributes (name named-function? parameters (parameter 1) body)
    (ρ :binder+body #:when (and (α body) (or (α named-variable?)
                                             (and (α named-function?)
                                                  (pair? (α parameter))
                                                  (andmap identifier? (α parameter))
                                                  (distinct? (cons (α name) (α parameter))))))))

  (define-syntax-class with/local-shape #:attributes {name [definition 1] .body body with?}
    (ρ (~or* ((~binding? name #'compthink:local) (~bind (with? #false)) [definition ...] . local-body)
             ((~binding? name #'compthink:with)  (~bind (with?  #true))  definition ...+ with-body))
       #:attr .body (if (α with-body) #'(with-body) #'local-body)
       #:with {~maybe (body)} {α .body}))

  (provide Block List Variable Single Each-clause With-clause for-form for/list-form)
  ;
  (define an-accumulate-expression "an accumulate expression")
  ;
  (define-syntax-class for-form #:description an-accumulate-expression
    #:attributes (name parameters namer e-id s a-id⒮ a⒮ body (clause 1) es first-type)
    (ρ ((~binding? name #'compthink:for)
        (~and (~seq (~or* clause:With-clause clause:Each-clause clause)...)
              (~seq (~alt (~once with:With-clause #:name a-with-clause)
                          (~once each:Each-clause #:name an-each-clause)
                          (~once (~describe an-update-expression body:expr)
                                 #:name an-update-expression))
                    ...)))
       #:with (a-id⒮ a⒮) #'(with.a-id⒮ with.a⒮)
       #:with (e-id s) #'(each.e-id each.s)
       #:with parameters #'(e-id . with.a-id⒮.elements)
       #:attr namer (λ (parts) (slist* #'name parts))
       #:do [(define non-#false (λ (l) (filter values l)))]
       #:attr first-type (car (non-#false (α clause.type)))
       #:with es              (non-#false (α clause.expression))))
  ;
  (define-syntax-class With-clause #:no-delimit-cut
    #:attributes (a-id⒮ a⒮ a-id⒮.elements with-builder expression type)
    #:description a-with-clause
    (ρ ((~binding? with-name #'compthink:with) ~! (~describe #:opaque a-variable-or-list-of-variables
                                                             (~or* (~and _:Variable a-id⒮:Single)
                                                                   (~and (_:id ...) a-id⒮:List)))
                                               (~or* a⒮:List a⒮))
       #:attr with-builder (syntax-parser [a⒮ #'(with-name a-id⒮ a⒮)])
       #:with expression #'a⒮ #:attr type 'with))
  ;
  (define-syntax-class Each-clause #:no-delimit-cut #:datum-literals (each)
    #:attributes (e-id s each-builder expression type)
    #:description an-each-clause
    (ρ (each-name:each ~! (~describe #:opaque a-variable e-id:Variable) s:expr)
       #:attr each-builder (syntax-parser [s (build-Each-clause #:name #'each-name #'e-id #'s)])
       #:with expression #'s #:attr type 'each))
  (define (build-Each-clause #:name [name 'each] e-id s) (slist name e-id s))
  ;
  ; note: form name is always prefixed to the error.
  (define-syntax-class for/list-form #:description #false
    #:attributes (name parameters namer e-id s condition body (clause 1))
    (ρ (name (~and (~seq clause ...)
                   (~seq (~alt (~once :Each-clause #:name an-each-clause)
                               (~optional :If-clause #:name an-if-clause)
                               (~once (~describe an-element-expression body:expr)
                                      #:name an-element-expression))
                         ...)))
       #:with parameters #'(e-id)
       #:attr namer (λ (parts) (slist* #'name parts))))
  ;
  (define-syntax-class If-clause #:no-delimit-cut
    #:attributes (condition) ; otherwise _ seems to be an attribute
    #:description an-if-clause
    (ρ ((~binding? name #'compthink:if)
        ~! (~describe #:opaque a-condition-expression condition:expr))))
  ;
  (define-syntax-class Variable #:description a-variable (ρ (~describe #:opaque #false (~is id))))
  ;
  (define-syntax-class List #:description "a list" #:datum-literals (list)
    (ρ (name:list ~rest (~and elements:sequence (element ...)))
       #:attr builder (syntax-parser [(element ...) #'(name element ...)])
       #:attr the-length (length (α element))))
  ;
  (define-syntax-class Single (ρ e #:with elements #'(e)))
  ;
  (define a-variable "a variable")
  (define a-variable-or-list-of-variables "a variable or list of variables")
  (define an-element-expression "a body expression")
  (define an-if-clause "an if condition clause")
  (define a-condition-expression "a condition expression")
  (define a-with-clause "a  with  clause")
  (define an-each-clause "an  each  clause")
  (define an-update-expression "a body expression")

  (require racket/block)
  (define-syntax-class Block #:attributes (name es)
    (ρ (name:id ~rest es:sequence) #:when (free-identifier=? (α name) #'block)))  
  
  #| close |#

  ; model: each binding form has syntax-class with attribute ‘parameters’, all expressions are
  ;  assumed to be under their scope, syntax has been checked by the time ‘close’ is called.

  (define-syntax-class with/local:binds #:attributes {parameters}
    (ρ with/local:with/local-shape
       #:with {definition:define-name-shape ...} (α with/local.definition)
       #:with parameters (α definition.name)))

  (define-syntax-class for-shape:binds #:attributes {parameters} (ρ :for-form))
  (define-syntax-class for/list:binds  #:attributes {parameters} (ρ :for/list-form))

  (define close (syntax-parser [(~or* :fun-shape :define-name-shape :with/local:binds
                                      :for-shape:binds :for/list:binds
                                      (~is sequence))
                                #:with {id ...} #'{~? parameters {}}
                                #:with {part ...} (stx-map close this-syntax)
                                #'(base:let ([id #'id] ...) #`(#,part ...))]
                               [id:id #:when (not (lexical? (α id))) #'#'id]
                               [s #'#`'#,s]))

  ; ———

  (define (expand-for-errors stx) (local-expand stx 'expression '()))
  
  (require racket/lazy-require)
  (lazy-require [(submod "../functions/image.rkt"     nullary) (image:nullary-id?)]
                [(submod "../functions/non-image.rkt" nullary) (non-image:nullary-id?)])
  (define id-nullary? (∨ image:nullary-id? non-image:nullary-id?))

  (require (submod "syntax-error.rkt" binding-forms))
                 
  (define (definable? id) ((top?) . ⇒ . (not (identifier-binding id))))
  (define (compthink-defintion-context?) (or (top?) (in-with?)))

  (define (in-with?) (local [(define context (syntax-local-context))]
                       (and (pair? context) (expanding-for-with? (car context)))))

  (require (only-in syntax/kerncase kernel-form-identifier-list)
           (only-in syntax/context build-expand-context))
  (struct expanding-for-with ())
  (define kernel-forms (kernel-form-identifier-list)))


#| #%app fun |#

(begin-for-syntax
  (define-syntax-class callable #:attributes (name) (ρ (~or* name:id :fun-shape (name:id _ ...+)))))

(define-syntax-parser/⊥ compthink:#%app
  ; Special Non-call Forms
  [(_ . fun:fun-shape) (§ (#%fun . fun))]
  ; Call Forms
  ;  • is f expression even a valid expression
  [(_ . (f . _)) #:do [(expand-for-errors (α f))] #:when #false '⊥]
  ;  • allowed form for this call's f, inner problems screened by previous
  [(_ . (f:callable argument:expr ...))
   #:when ((null? (α argument)) . ⇒ . ((∧ identifier? id-nullary?) (α f)))
   (§ (base:#%app f argument ...))]
  ;  • allowed form for this call's f, inner problems screened by previous,
  ;     but this call is failed allowed-nullary screening by previous
  [(_ . (f:callable)) (⌢:call:nullary (α f))]
  ;  • disallowed form for this call's f
  [(_ . (~maybe (part . _))) (⌢:call:non:id/fun (or (α part) (void)))]
  #:error-module (submod "syntax-error.rkt" call))
;
(define-syntax-parser/⊥ (compthink:fun . _) (⌢:fun:header:bare)
  #:error-module (submod "syntax-error.rkt" binding-forms))
;
(define-syntax-parser #%fun
  [(_ . fun:binder-body) #:when (none? lexical? (α fun.parameter))
                         #:with anon (§ (base:λ fun.parameters fun.body))
                         #:with code (close (§ fun))
                         (§∘ (base:#%app wrap-λ anon (base:λ () code)))]
  [(_ . fun) (parameterize [(context (α fun))] (⌢:function this-syntax))])

#| Assuming |#

(begin-for-syntax (require (submod "syntax-error.rkt" conditionals)))

(define-syntax-parser/⊥ (compthink:assuming pre-condition:expr result:expr)
  (§ (base:if (base:#%app true? pre-condition) result (base:#%app ~s 'pre-condition)))
  #:arity ⌢:assuming:arity)

#;[:if-shape #:with (checked-condition ...)
             (for/list ([condition (α condition)])
               (§∘ condition (base:#%app if:checked-boolean (base:quote name) #,condition)))
             (§ (base:cond [checked-condition result] ... [base:else else-result]))]

#| time |#

(define-syntax-parser/⊥ (compthink:time expression:expr)
  (§ (base:#%app time (base:λ () expression) (base:quote expression)))
  #:arity ⌢:time:arity #:error-module (submod "syntax-error.rkt" miscellaneous))


#| check |#

(define-syntax-parser/⊥ compthink:check
  [(_ actual:expr expect:expr) #:when (top?) (§ (check-expect actual expect))]
  [_ #:when (not (top?)) (⌢:non-top this-syntax)]
  #:arity ⌢:check:arity #:error-module (submod "syntax-error.rkt" miscellaneous))


#| define |#

(define-syntax-parser/⊥ compthink:define
  [(_ . defining:binder-body)
   #:when (and (compthink-defintion-context?) (definable? (α defining.name)))
   #:with definition (§ (base:define . defining))
   #:attr redefine-check (and (interactions-top?) (§∘ (base:#%app define:interaction:redefine-check
                                                                  #'#,this-syntax #'defining.name)))
   #:attr record-definition (and (top?) (α defining.named-function?)
                                 (§∘ (base:#%app add-definition #'#,this-syntax)))
   (§ (base:begin {~? redefine-check} definition {~? record-definition}))]
  [_ #:when (not (compthink-defintion-context?)) (⌢:define:context)]
  [(_) (⌢:define:nothing)]   
  [:define-name-shape #:when (and #;(definitions-top?) (top?) (not (definable? (α name))))
                      (⌢:define:redefine (α name) (identifier-binding (α name)))]
  [(_ id:id ~rest :sequence) (⌢:define:variable:init:arity (α length) (α id))]
  [(_ (:id . _) . _) (⌢:function this-syntax)]
  [(_ part      . _) (⌢:define:non-binding (α part))])

(define-for-syntax (expanded-definitions definitions)
  (define int-def-ctx (build-expand-context (expanding-for-with)))
  (for/fold ([seen '()] [expands '()] #:result (reverse expands)) ([d definitions])
    (syntax-parse d
      [:define-name-shape (define seen′ (cons (α name) seen))
                          (if (distinct? seen′)
                              (values seen′ (cons (local-expand d int-def-ctx kernel-forms) expands))
                              (⌢:with:duplicate (α name)))]
      [_ (⌢:with:non-definition d)])))
;
(define-syntax-parser/⊥ compthink:local
  [:with/local-shape (with/local (expanded-definitions (α definition)) (α .body))]
  [(_)          (⌢:with:parts:none)]
  [(_ [] . _)   (⌢:with:definitions:none)]
  [(_ part . _) (⌢:with:non-sequence (α part))])
;
(define-syntax-parser/⊥ compthink:with
  [:with/local-shape (with/local (expanded-definitions (α definition)) (α .body))]
  [(_)          (⌢:with:parts:none)]
  #;[(_ _)        (⌢:with:definitions:none)]
  #;[(_ part . _) (⌢:with:non-sequence (α part))])
;
(define-for-syntax with/local
  (syntax-parser∗ [{expanded (body:expr)} #'(#%stratified-body (~@ . expanded) body)]
                  [{_ :sequence} (⌢:with:body:arity (α length))]))
;
(define (with-local-function f parameters body local-body)
  #`(compthink:with (compthink:define (#,f . #,parameters) #,body) #,local-body))

(define-for-syntax (⌢:function stx)
  (define/syntax-parse (_ . (~and :binder+body (~maybe anonymous:fun-shape))) stx)
  (cond [(not (α named?)) (⌢:define:function:non-id (α name))]
        [(null? (α parameter)) (⌢:parameters:none (α name) (and (α anonymous) #true) (α binder))]
        [else (for/fold ([seen '()]) ([parameter (α parameter)])
                (define seen′ (cons parameter seen))
                ((cond [(not (identifier? parameter)) ⌢:parameters:non-id]
                       [(lexical? parameter) ⌢:parameters:local:duplicate]
                       [(and (not (α anonymous)) (bound-identifier=? (α name) parameter))
                        ⌢:define:parameters:function]
                       [(not (distinct? seen′)) ⌢:parameters:duplicate]
                       [else (λ (_) seen′)])
                 parameter))
              (syntax-parse (α body-parts) [:sequence (⌢:function:body:arity (α length))])]))

#| for |#

(begin-for-syntax (define-pattern ~handle-compound [(_ pattern) #'(~and (~is-not id) ~! pattern)]))

(define-syntax-parser/⊥ compthink:for
  [(~handle-compound :for-form)
   #:with (a⒮′ body′ unary?) (if (identifier? (α a-id⒮))
                                 #'((base:#%app list a⒮) (base:#%app list body) #true)
                                 #'(a⒮ body #false))
   (§ (base:#%app for-loop 'name s a⒮′ (base:λ parameters body′) unary?))])

(define-syntax-parser/⊥ compthink:for/list
  [(~handle-compound :for/list-form) (§ (for/list ([e-id s] {~? {~@ #:when condition}}) body))])

#| or and if else |#

(begin-for-syntax (require (submod "syntax-error.rkt" conditionals)))

(define-syntaxes (compthink:and compthink:or)
  (local [(define (handler delegatee)
            (parser/⊥ [:or/and-shape
                       #:when ((length (α condition)) . >= . 2)
                       (§∘ (#,delegatee
                            . #,(for/list ([(condition i) (in-indexed (α condition))])
                                  (§∘ condition (base:#%app or/and:checked-boolean (base:quote name)
                                                            #,condition (base:quote #,i))))))]
                      #:arity ⌢:or/and:arity))]
    (values (handler #'base:and) (handler #'base:or))))

(define-syntax-parser/⊥ compthink:else [_ (⌢:else)])

(define-syntax-parser/⊥ compthink:if
  [:if-shape #:with (checked-condition ...)
             (for/list ([condition (α condition)])
               (§∘ condition (base:#%app if:checked-boolean (base:quote name) #,condition)))
             (§ (base:cond [checked-condition result] ... [base:else else-result]))]
  [(_)                        (⌢:if:nothing)]
  [(_ [(~is else-shape) . _]) (⌢:if:else:only)]
  [(_ [_ _])                  (⌢:if:clause:only)]
  [(_ . clauses:sequence) (define final (sub1 (α clauses.length)))
                          (for ([(clause i) (in-indexed (stx->list (α clauses)))])
                            ((syntax-parse clause
                               [[(~true (< i final) (~is else-shape))     . _] ⌢:if:else:early]
                               [[(~true (= i final) (~is-not else-shape)) . _] ⌢:if:else:missing]
                               [[_ _] void]
                               [:sequence (λ (clause) (⌢:if:clause:arity (α length) clause))]
                               [_ ⌢:if:non-clause])
                             clause))])
