#lang racket/base (require (submod "../../snack/contract.rkt" core))

(provide
 (contract-out
  [display-steps! ((listof sequence?) #:wait? boolean? #:interactions? boolean? #:top? boolean?
                                      . → . any #;void?)]))

(require (submod "../../configuration.rkt" step-display))
; See also the ‘print’ sub-module for token coloring.

(require (only-in "../../interpreter/interpreter.rkt" error? whole-error? error-expression error-exn)
         (only-in "../../interpreter/layout.rkt" chunks space-width))

(require racket/require
         "../../snack/lambda.rkt"
         (submod "../../snack/list.rkt" core)
         (multi-in "../../snack" ("string.rkt" "port.rkt")))
  
(define (display-steps! steps #:wait? wait? #:interactions? interactions? #:top? top?)
  #;(contracts [spacious? boolean?] [allow-wait? boolean?]
               [(display-step! any/c #;♯expression? string?) → any]
               [(maybe-wait) → any])
  (define spacious?   (not (preference:display:spacing:compact)))
  (define allow-wait? (not (preference:interaction:no-wait)))

  (unless interactions? (display/default-color message:header #:newline? #true))
  (when spacious? (newline))
      
  (define (display-step! step lead)
    (define lead-indent (string-length lead))
    ; For a whole error only uses the exception, for other error chunks the error-expression,
    ;  otherwise chunks the expression.
    ; Assumes new and next markup is balanced, and neither self-nests.
    (unless (whole-error? step)
      (let ([step (if (error? step) (syntax->datum (error-expression step)) step)])
        (parameterize ([space-width (a-space-width)])
          (for* ([(line index) (in-indexed (chunks step (- maximum-step-width lead-indent)))]
                 [stop? (in-value (index . >= . maximum-step-lines))] #:final stop?)
            (when (positive? lead-indent)
              (cond [(zero? index) (display/default-color lead #:newline? #false)]
                    [else #;(display/default-color (make-string indent-amount #\space)
                                                   #:newline? #false)
                          (display (make-string lead-indent #\space))]))
            (cond [stop? (display/default-color (message:elided) #:newline? #true)]
                  [else (for-each display-token line) (newline)])))))
    (when (error? step) (define e (error-exn step)) ((error-display-handler) (exn-message e) e)))
    
  (define (maybe-wait)
    (cond [(and allow-wait? wait?)
           (flush-output) ; Inside for efficiency of I/O.
           (with-handlers ([exn:fail? (λ₁ 'dont-care)]) (read-line))
           (consume-available!)]
          [spacious? (newline)]))

  (define (Δ s.0 s.1) s.1
    #;#;#;
    (local-require (only-in (submod "../../interpreter/token.rkt" layout) atom→token)
                   (only-in  "../../interpreter/private/common.rkt"
                             next? markup? markup-expression new?)
                   (only-in  (submod "../../interpreter/private/common.rkt" strings-and-images)
                             image? image-width rectangle above))
    (define-values (s′ _)
      (let Δ ([s.0 s.0] [s.1 s.1])
        (cond [(markup? s.0) (Δ (markup-expression s.0) s.1)]
              [(and (list? s.0) (list? s.1) (= (length s.0) (length s.1))
                    (not (equal? (first s.1) 'if)))
               (for/fold ([prefix '()] [in-shared-prefix? #true]) ([e.0 s.0] [e.1 s.1])
                 (define-values (e.1′ in-shared-prefix?′)
                   (if in-shared-prefix? (Δ e.0 e.1) (values e.1 #false)))
                 (values (append prefix (list e.1′)) in-shared-prefix?′))]
              [(equal? s.0 s.1)
               (values (if (image? s.1)
                           (above (rectangle (image-width s.1) 1 "solid" "gray")
                                  (rectangle 0                 7 "solid" "transparent"))
                           (string->symbol (make-string (string-length (atom→token s.1)) #\.)))
                       #true)]
              [else (values s.1 #false)])))
    s′)
  (local-require racket/stream)
  (local-require (only-in (submod "../../interpreter/private/common.rkt") cleanse))
  (local-require "../../snack/match.rkt")
  (local-require racket/control)
  
  (define final-step (void))
  (for*/and ([(top-level n) (in-indexed steps)]
             [first-step (in-value (stream-first top-level))] #:final (namespace? first-step))
    (cond [(namespace? first-step) (unless top? (match (cleanse final-step)
                                                  [`(define . ,_) (void)]
                                                  [(? error?) (abort)]
                                                  [expression (eval expression first-step)]))]
          [else (set! final-step first-step)
                (display-step! final-step (~a message:to-step-lead ␣))
                (flush-output)
                (for/and ([step (stream-rest top-level)]
                          [old-step top-level]
                          #;[n (in-range (sub1 (length (rest top-level))) -1 -1)])
                  (maybe-wait)
                  (cond #;[(n . >= . maximum-steps)
                           (display/default-color (message:elided message:too-many-steps)
                                                  #:newline? #true)
                           #false]
                        [else (display-step! (if #true #;(zero? n) step (Δ old-step step))
                                             (if spacious? (~a message:step-lead ␣) ""))
                              (set! final-step step)
                              #true]))
                (maybe-wait)]))
  #;(newline) #|For spacing, and also to produce void|# )

(module print racket/base (require (submod "../../snack/contract.rkt" core))

  (provide (contract-out [display/default-color (string? #:newline? boolean? . → . any)]
                         [display-token (token/c . → . any)]))

  (require (submod "../../private/print.rkt" step-display)
           (only-in (submod "../../shared-settings.rkt" step) preference:display:underline:next)
           (submod "../../interpreter/token.rkt" step-form))

  (require "../../snack/boolean.rkt"
           "../../snack/conditional.rkt"
           "../../snack/function.rkt")
  
  (define (display/default-color string #:newline? newline?)
    (print:display/default-color string) (when newline? (newline)))

  #;(contracts [in-next? boolean?]) (define in-next? #false)

  (define (underline-image image)
    (local-require (only-in mrlib/image-core make-image image-shape make-bb))
    (define image₁ (above (freeze (add1 (image-width image))
                                  (add1 (image-height image))
                                  image)
                          (rectangle 0 1 "solid" "black")
                          (rectangle (add1 (image-width image)) 0 "outline" "black")))
    (make-image (image-shape image₁) (make-bb (image-width     image₁)
                                              (image-height    image₁)
                                              (- (image-height image₁) 1))
                #false))
  
  (define (display-token token)
    (when (next-start? token) (set! in-next? #true))
    (define content (if (wrapped? token) (wrapped-content token) token))
    (define annotating? (preference:display:underline:next))
    (define underline-bold? (and annotating? ((∨ next-start? next-end?) token)))
    (define underline?      (and annotating? (not underline-bold?) (or in-next?
                                                                       (dereference? token))))
    (print:display/racket-style
     (if (underline? . ⇒ . (string? content)) content (underline-image content))
     (type-case content
       [delimiter-token? 'parenthesis]
       [(∨ boolean-token? number-token? box-token? void-token?) 'constant]
       [string-token? 'string]
       [image-token? 'image]
       [#:assert identifier-token? 'symbol])
     #:delta♯ (cond [underline-bold? print:delta:underline-bold]
                    [underline?      print:delta:underline]
                    [else            #false]))
    (when (next-end? token) (set! in-next? #false))))

(require 'print)
