#lang racket/base (provide step steps hide show)

#| Runtime |#

(require (only-in "current-definitions.rkt" get-definitions)) ; Shared state.

#| Expansion Time |#

(require (submod "../../snack/syntax.rkt" core)) ; Standard imports.

(require (only-in "forms.rkt" close)
         (only-in "../../shared/code.rkt" syntax→code′))

(begin-for-syntax
  
  (require (only-in "common.rkt" top?)
           (only-in "syntax-error-context.rkt" context)) ; Shared state for syntax errors.
  
  (require racket/lazy-require) ; Lazily require much of the static error handling.
  (lazy-require [(submod "syntax-error.rkt" step-form) (syntax-error ⌢:bare/weird)]
                [(submod "../../english.rkt" syntax)
                 (message:hide:out-of-context
                  message:show:out-of-context
                  message:hide:non:function/call message:hide:call:non-id
                  message:hide:call:nullary message:hide:call:argument:non-literal)])

  (require ; More specific imports.
           (only-in (submod "../../snack/list.rkt" core) rest first empty?)
           (only-in "../../snack/function.rkt" ∨)))

; Will delegate syntax check of code the user is stepping by massaging it into a local.
(require (only-in "forms.rkt" [compthink:with with]))

#| Delay load of interpreter and display to first invocation of step at runtime. |#
(require racket/lazy-require)
(module step-display racket/base (provide display-program-steps)
  (require (only-in "step-display.rkt" display-steps!)
           (only-in "../../interpreter/interpreter.rkt" definitions hiding program-steps))
  (define (display-program-steps program get-definitions
                                 #:wait? wait? #:interactions? interactions? #:top? top?
                                 #:hide hide #:namespace namespace)
    (display-steps! (parameterize ([definitions (get-definitions)] [current-namespace namespace])
                      (hiding hide (program-steps program)))
                    #:wait? wait? #:interactions? interactions? #:top? top?)))
(lazy-require [(submod "." step-display) (display-program-steps)])

#| ——— |#

(define-syntax (hide stx) (syntax-error stx (message:hide:out-of-context)))
(define-syntax (show stx) (syntax-error stx (message:show:out-of-context)))

(define-syntax (step  stx) (stepping stx #:interactive? #true))
(define-syntax (steps stx) (stepping stx))

(begin-for-syntax (define-splicing-syntax-class options #:literals (hide)
                    (pattern (~seq (~alt (~optional (~seq (~and hiding (hide to-hide:expr ...)))))
                                   ...))))

(define-for-syntax (stepping stx #:interactive? [interactive? #false])
  (step-error-check stx)
  (syntax-parse stx
    [(step-form:id (~optional (~and interactions #:interactions)) :options program ...)
     #`(#;void
        values
        (cond [#false #,(localize #'(program ...))]
              [else (display-program-steps
                     #:wait? '#,interactive?
                     #:interactions? '#,(if (attribute interactions) #'#true #'#false)
                     #:top? '#,(and (top?) #true)
                     #:hide '#,(if (attribute hiding) #'(to-hide ...) #'())
                     #:namespace (variable-reference->namespace (#%variable-reference))
                     (syntax→code′ #,(close #'(program ...))) get-definitions)]))]))

(define-for-syntax (localize stx)
  (syntax-parse stx #:datum-literals (define)
    [((~and (define . _) definition) ...+ . parts)
     #`(with definition ... #,(localize #'parts))]
    [(expression . parts) #`(cond [#false expression] [else #,(localize #'parts)])]
    [() #'#false]))

(define-for-syntax (step-error-check stx)
  (define-values (compthink-number? image? list-literal?)
    (values real? struct? (λ (v) (and (list? v) (not (empty? v)) (equal? (first v) 'list)))))
  (syntax-parse stx #:literals (hide)
    [(step-form:id (~optional #:interactions) the-options:options program ...)
     (define hiding (attribute the-options.hiding))
     (when hiding
       (define (error msg) (syntax-error hiding msg))
       (for ([hidden (rest (syntax->list hiding))])
         (unless (identifier? hidden)
           (syntax-parse hidden
             [()                             (error (message:hide:non:function/call (void)))]
             [(~not (~or :id (_ . _)))       (error (message:hide:non:function/call this-syntax))]
             [((~and (~not :id) non-id) . _) (error (message:hide:call:non-id #'non-id))]
             [(f)                            (error (message:hide:call:nullary #'f))]
             [(_ arg ...)
              (for ([arg (syntax->list #'(arg ...))]
                    #:unless ((∨ compthink-number? boolean? string? image? list-literal?)
                              (syntax->datum arg)))
                (error (message:hide:call:argument:non-literal arg)))]))))]
    [_ (parameterize ([context this-syntax]) (⌢:bare/weird))]))
