#lang racket/base

(module binding-forms racket/base
  (provide ⌢:define:variable:init:arity
           ⌢:define:redefine ⌢:define:context ⌢:define:nothing ⌢:define:non-binding
           ⌢:fun:header:bare
           ⌢:define:function:non-id
           ⌢:define:parameters:function ⌢:function:body:arity
           ⌢:parameters:none ⌢:parameters:non-id ⌢:parameters:duplicate ⌢:parameters:local:duplicate
           ⌢:with:parts:none ⌢:with:non-sequence
           ⌢:with:definitions:none ⌢:with:duplicate ⌢:with:non-definition
           ⌢:with:body:arity)
  (require racket/lazy-require)
  (lazy-require
   [(submod "..")
    (⌢:define:variable:init:arity
     ⌢:define:redefine ⌢:define:context ⌢:define:nothing ⌢:define:non-binding
     ⌢:fun:header:bare
     ⌢:define:function:non-id
     ⌢:define:parameters:function ⌢:function:body:arity
     ⌢:parameters:none ⌢:parameters:non-id ⌢:parameters:duplicate ⌢:parameters:local:duplicate
     ⌢:with:parts:none ⌢:with:non-sequence
     ⌢:with:definitions:none ⌢:with:duplicate ⌢:with:non-definition
     ⌢:with:body:arity)]))

(module conditionals racket/base
  (provide ⌢:or/and:arity
           ⌢:else
           ⌢:if:nothing
           ⌢:if:else:only ⌢:if:clause:only ⌢:if:else:early ⌢:if:else:missing
           ⌢:if:clause:arity ⌢:if:non-clause
           ⌢:assuming:arity)
  (require racket/lazy-require)
  (lazy-require
   [(submod "..") (⌢:or/and:arity
                   ⌢:else
                   ⌢:if:nothing
                   ⌢:if:else:only ⌢:if:clause:only ⌢:if:else:early ⌢:if:else:missing
                   ⌢:if:clause:arity ⌢:if:non-clause
                   ⌢:assuming:arity)]))

(module miscellaneous racket/base

  (provide (all-defined-out)
           ⌢:bare/weird ⌢:top:unbound
           ⌢:check:arity ⌢:time:arity)  
  (require racket/lazy-require)  
  (lazy-require [(submod "..") (⌢:bare/weird ⌢:top:unbound ⌢:check:arity ⌢:time:arity)])
  (define (message:unusable-character) "this character is not usable"))

(require "syntax-error-context.rkt")

(require (prefix-in message: (only-in (submod "../../english.rkt" syntax)
                                        function-call
                                        call:nullary
                                        call:non:id/fun)))
(define (⌢:call:nullary    f)    (syntax-error′ message:call:nullary f))
(define (⌢:call:non:id/fun part) (syntax-error′ message:call:non:id/fun part))
(define (syntax-error′ message/r [about (void)])
  (raise-syntax-error (string->symbol message:function-call)
                      (if (procedure? message/r) (message/r about) message/r)
                      (context)))
(provide ⌢:call:nullary ⌢:call:non:id/fun)
(module call racket/base (provide ⌢:call:nullary ⌢:call:non:id/fun)
  (require racket/lazy-require)
  (lazy-require [(submod "..") (⌢:call:nullary ⌢:call:non:id/fun)]))

(module* require   #false (provide syntax-error ⌢:bare/weird))
(module* step-form #false (provide syntax-error ⌢:bare/weird))

(require (submod "../../snack/contract.rkt" core)
         (only-in (submod "../../snack/contract.rkt" core) [any ⊥]))

(define arity/c (natural/c . → . ⊥))
(define (⊥/c domain) (domain . → . ⊥))

(provide (contract-out
          [⌢:bare/weird (→ ⊥)]
          ; We could add a property to note whether it's in function position.
          [⌢:top:unbound (→ ⊥)]
          [⌢:define:redefine (identifier? (♯ (or/c 'lexical non-empty/c)) . → . ⊥)]
          [⌢:define:redefine:user (⊥/c identifier?)]
          #;[define:interaction:redefine-check (syntax? identifier? . → . void?)]
          [⌢:define:context (→ ⊥)]
          [⌢:define:nothing (→ ⊥)]
          [⌢:define:variable:keyword (⊥/c syntax?)]
          [⌢:define:function:keyword (⊥/c syntax?)]
          [⌢:parameters:keyword      (⊥/c syntax?)]
          [⌢:define:non-binding      (⊥/c syntax?)]
          [⌢:define:function:non-id  (⊥/c syntax?)]
          
          [⌢:parameters:non-id       (⊥/c syntax?)]
          [⌢:define:parameters:function (⊥/c identifier?)]
          [⌢:parameters:duplicate       (⊥/c identifier?)]
          [⌢:parameters:local:duplicate (⊥/c identifier?)]
          [⌢:parameters:none (identifier? boolean? syntax? . → . ⊥)]
          
          [⌢:define:variable:init:arity (natural/c identifier? . → . ⊥)]
          [⌢:function:body:arity        arity/c]
          
          [⌢:with:parts:none       (→ ⊥)]
          [⌢:with:definitions:none (→ ⊥)]
          [⌢:with:non-definition (⊥/c syntax?)]
          [⌢:with:body:arity     arity/c]
          [⌢:with:non-sequence   (⊥/c syntax?)]
          [⌢:with:duplicate      (⊥/c identifier?)]
          
          [⌢:or/and:arity   arity/c]
          [⌢:check:arity    arity/c]
          [⌢:assuming:arity arity/c]
          [⌢:time:arity     arity/c]
          [⌢:big-bang:arity arity/c]
          
          [⌢:else (→ ⊥)]
          ;
          [⌢:if:nothing     (→ ⊥)]
          ;
          [⌢:if:else:only   (→ ⊥)]
          [⌢:if:clause:only (→ ⊥)]
          ;
          [⌢:if:else:early   (⊥/c syntax?)]
          [⌢:if:else:missing (⊥/c syntax?)]
          ;
          [⌢:if:clause:arity (natural/c syntax? . → . ⊥)]
          [⌢:if:non-clause   (⊥/c syntax?)]
          
          [⌢:fun:header:bare (→ ⊥)]))

(require (prefix-in message: (submod "../../english.rkt" syntax)))

(require (submod "../../snack/syntax.rkt" core))

(generate
 [[⌢:catch-all message:incorrect-form]
  [⌢:top:unbound message:top:unbound]
  [⌢:define:redefine:non-user (message:define:redefine:non-user name) detail]
  [⌢:define:redefine:user     (message:define:redefine:user     name) detail]
  ;
  [⌢:define:context message:define:context]
  [⌢:define:nothing message:define:nothing]
  [⌢:define:variable:keyword message:define:keyword          keyword]
  [⌢:define:function:keyword message:define:function:keyword keyword]
  [⌢:parameters:keyword      message:parameters:keyword      keyword]
  [⌢:define:non-binding      (message:define:non-binding      part) as-detail]
  [⌢:define:function:non-id  (message:define:function:non-id  part) as-detail]
  
  [⌢:parameters:non-id (message:parameters:non-id part) as-detail]
  [⌢:define:parameters:function message:define:parameters:function id]
  [⌢:parameters:duplicate       (message:parameters:duplicate       id) as-detail]
  [⌢:parameters:local:duplicate (message:parameters:local:duplicate id) as-detail]
  
  [⌢:function:body:arity (message:function:body:arity amount)]
  
  [⌢:parameters:none (message:parameters:none function anonymous?) header]
  
  [⌢:define:variable:init:arity (message:define:variable:init:arity amount variable)]
  
  [⌢:with:parts:none       message:with:parts:none]
  [⌢:with:definitions:none message:with:definitions:none]
  [⌢:with:non-definition  (message:with:non-definition part)]
  [⌢:with:body:arity      (message:with:body:arity amount)]
  [⌢:with:non-sequence    (message:with:non-sequence part)]
  [⌢:with:duplicate       (message:with:duplicate id) as-detail]
  
  [⌢:or/and:arity   (message:conditions:less-than-two actual)]
  [⌢:check:arity    (message:expressions:not-two actual)]
  [⌢:assuming:arity (message:expressions:not-two actual)]
  [⌢:time:arity     (message:expressions:not-one actual)]
  
  [⌢:big-bang:arity (message:big-bang:arity actual)]  
  
  [⌢:else message:else:context]
  ;
  [⌢:if:nothing      message:if:nothing]
  ;
  [⌢:if:else:only    message:if:else:only]
  [⌢:if:clause:only  message:if:clause:only]
  ;
  [⌢:if:else:early   message:if:else:early   clause]
  [⌢:if:else:missing message:if:else:missing clause]
  ;
  [⌢:if:clause:arity (message:if:clause:arity amount) clause]
  [⌢:if:non-clause   (message:if:non-clause   clause) as-detail]]
 
 [f:id (~or* message/r:id (message/r:id v₀:id (~optional v₁:id))) (~optional detail:id)]

 (define (f (~? v₀) (~? v₁) (~? [detail v₀] (~? detail)))
   (syntax-error (context) (~? (message/r v₀ (~? v₁)) message/r) (~? detail))))

(define (form-name) (syntax-parse (context) [(name . _) (attribute name)]))
(define (⌢:fun:header:bare) (syntax-error (context) (message:fun:header:bare (form-name))))

(define (⌢:define:redefine id previous-binding)
  (define-values (mp bp) (if (pair? previous-binding)
                             (module-path-index-split (car previous-binding))
                             (values #false #false)))
  ((if mp ⌢:define:redefine:non-user ⌢:define:redefine:user) id id))
#;
(define (define:interaction:redefine-check definition id)
  (when (or (identifier-binding id)
            ;; identifier-binding returns #f for variable bound at the top-level, check explicitly:
            (namespace-variable-value (syntax-e id) #true (λ () #false)))
    (syntax-error definition (message:define:redefine:user id) id)))

(require (submod  "../../snack/syntax.rkt" syntax-classes))
(define (⌢:bare/weird) (syntax-parse (context) [(~or* :bare :catch-all) unreachable]))
(define unreachable 'unreachable)
(syntax-class (bare id:id #:do [(⌢:bare #'id)]))
(define (⌢:bare id) (syntax-error id (message:bare id)))
(syntax-class (catch-all _ #:do [(⌢:catch-all)]))

#;(contracts [syntax-error ([syntax? string?] [(♯ syntax?)] . →∗ . ⊥)])
(define (syntax-error stx msg [detail #false])
  (raise-syntax-error (syntax-parse stx [((form:id . _) . _) (syntax-e #'form)] [_ #false])
                      msg stx detail))

#; 
(define (top:checked-identifier id)
  (namespace-variable-value (syntax-e id) #true (λ₀ (syntax-error id "this variable is undefined"))))
