#lang racket/base

(provide (struct-out /C) (all-defined-out))

(require "C.rkt")

(require (prefix-in message: (combine-in (submod "../../english.rkt" types)))
         (only-in (submod "custom.rkt" non-image) non-negative?))

(require (submod "../../snack/syntax.rkt" core))

(define number? real?)

#| Any |#
(define any/C (/C (λ (_) #true) ""))

#| Main Disjoint Patitioning of Types, except boolean and image. |#
(define number/C   (/C real?      message:number))
(define text/C     (/C string?    message:text))
(define function/C (/C procedure? message:function))
(define list/C     (/C (λ (•) (or (pair? •) (null? •))) message:list))

#| Standard Numeric Sub-Types |#
(define-values (integer/C natural/C positive-natural/C)
  (values (/C exact-integer?             message:integer)
          (/C exact-nonnegative-integer? message:natural-number)
          (/C exact-positive-integer?    message:positive-natural-number)))
(generate {[non-negative/C non-negative?           message:non-negative-number]
           [non-zero/C     (λ (•) (not (zero? •))) message:non-zero-number]
           [positive/C     positive?               message:positive-number]}
          [protected/C:id c:expr message]
          (define protected/C (/C (λ (•) (and (number? •) (c •))) message)))

(module* non-image #false (provide (all-defined-out))

  (require (prefix-in message: (submod "../../english.rkt" problems))
           (only-in (submod "custom.rkt" non-image) character? unary? binary?))

  (define non-empty/C (/C pair?
                          (message:list-length-at-least 1)))
  (define two/C       (/C (λ (•) (and (pair? •)
                                      (pair? (cdr •))))
                          (message:list-length-at-least 2)))
  (define three/C     (/C (λ (•) (and (pair? •)
                                      (pair? (cdr •))
                                      (pair? (cddr •))))
                          (message:list-length-at-least 3)))
  (define four/C      (/C (λ (•) (and (pair? •)
                                      (pair? (cdr •))
                                      (pair? (cddr •))
                                      (pair? (cdddr •))))
                          (message:list-length-at-least 4)))
  (define boolean/C  (/C boolean? message:boolean))

  #| Text-related Sub-Types |#
  (define character/C (/C character? message:character))
  (define unicode/C
    (/C (λ (•) (and (exact-nonnegative-integer? •) (or (<= • 55295) (<= 57344 • 1114111))))
        message:unicode))

  #| Function Sub-Types |#
  (define-values (unary/C binary/C)
    (values (/C (λ (•) (and (procedure? •) (unary?  •))) message:unary)
            (/C (λ (•) (and (procedure? •) (binary? •))) message:binary)))
  (define-values (unary-predicate/C binary-predicate/C)
    (values (/C (λ (•) (and (procedure? •) (unary?  •))) message:unary)
            (/C (λ (•) (and (procedure? •) (binary? •))) message:binary)))

  (define-values (list-index/C text-index/C)
    (values
     (/C (λ (collection i) (i . < .        (length collection))) message:list-index-too-large)
     (/C (λ (collection i) (i . < . (string-length collection))) message:text-index-too-large)))

  (define list-lower/C (/C (λ (collection i j) (<= i (length collection)))
                           message:list-range-index-too-large))
  (define list-upper/C (/C (λ (collection i j) (<= j (length collection)))
                           message:list-range-index-too-large))
  (define list-order/C (/C (λ (collection i j) (<= i j))
                           message:number-out-of-order))
  (define text-lower/C (/C (λ (collection i j) (<= i (string-length collection)))
                           message:text-range-index-too-large))
  (define text-upper/C (/C (λ (collection i j) (<= j (string-length collection)))
                           message:text-range-index-too-large))
  (define text-order/C (/C (λ (collection i j) (<= i j))
                           message:number-out-of-order))

  (define small-enough-for-random/C
    (/C (λ (•) (and (exact-positive-integer? •) (• . < . 4294967088))) message:smaller-number)))

(module* image #false (provide (all-defined-out))
  
  (require (prefix-in message: (only-in (submod "../../english.rkt" problems)
                                        ≤-something-of-the
                                        length-of-color-list))
           (only-in (submod "custom.rkt" image) image? color? colors?)
           (only-in racket/math exact-round)
           (only-in 2htdp/private/image-more image-width image-height)
           racket/local)
  
  (define-values (image/C color/C colors/C) (values (/C image?  message:image)
                                                    (/C color?  message:color)
                                                    (/C colors? message:list-of-colors)))

  (define angle/C (/C number? message:angle))
  (define percentage/C (/C (λ (•) (and (number? •) (<= 0 • 100))) "percentage"))

  (define font-size/C (/C (λ (•) (and (number? •) (<= 1 • 255))) message:font-size))

  (define-values (shading/C horizontal-position/C vertical-position/C) 
    (local [(define (member/C ℓ message) (/C (λ (•) (member • ℓ)) message))]
      (values (member/C message:shadings message:shading)
              (member/C message:horizontal-positions message:horizontal-position)
              (member/C message:vertical-positions   message:vertical-position))))

  (define-values (within-width/C within-height/C)
    (values
     (/C (λ (image amount) ((exact-round amount) . <= . (image-width  image)))
         (message:≤-something-of-the message:width  message:image))
     (/C (λ (image amount) ((exact-round amount) . <= . (image-height image)))
         (message:≤-something-of-the message:height message:image))))

  (define pixel-dimensions/C (/C (λ (colors width height) (= (length colors) (* width height)))
                                 message:length-of-color-list)))
