#lang racket/base (require (submod "meta-language.rkt" image))

(define-for-syntax default #'2htdp/private/image-more)

(define-syntax-parser create 
  [(_ :contracting binding)
   #:declare binding (alias/rename/from #'f #'here custom)
   (define/provide this-syntax (attribute binding.library) default #'binding.data)])

(module* nullary #false
  (require syntax/parse)
  (provide (prefix-out image: nullary-id?))
  (define-literal-set nullary (above beside overlay blank))
  (define nullary-id? (literal-set->predicate nullary)))

; Build some checks from contracts.rkt contracts via error.rkt combinators
; ————————————————————————————————————————————————————————————————————————
; Optional arguments.
; Cost of result presumably outweighs cost of contract check.
(define-values (check-1D-shape
                check-2D-shape
                check-sizes
                check:text→image)
  (values (arity-arguments-contract-checker 1 non-negative/C                shading/C color/C)
          (arity-arguments-contract-checker 2 non-negative/C non-negative/C shading/C color/C)
          (arity-arguments-contract-checker 0 non-negative/C non-negative/C)
          (arity-arguments-contract-checker 1 text/C font-size/C                      color/C)))
; This could possibly use optimization.
(define check-color-lens (arity-arguments-contract-checker 1 color/C percentage/C))
;
(define check-images (arguments-homogeneous-contract-checker image/C))
(define-values (check-horizontal-align check-vertical-align check-horizontal-vertical-align)
  (values (checker-fixed-homogeneous (list horizontal-position/C) image/C)
          (checker-fixed-homogeneous (list   vertical-position/C) image/C)
          (checker-fixed-homogeneous (list horizontal-position/C vertical-position/C) image/C)))
(define-values (check-within-width check-within-height)
  (values (binary-checker within-width/C) (binary-checker within-height/C)))
(define check-dimensions (contract-checker pixel-dimensions/C))
  
(map-macro create

           [(width  image/C) #:renames image-width]
           [(height image/C) #:renames image-height]
           
           [(scale        image/C non-negative/C) #:custom]
           [(scale-width  image/C non-negative/C) #:custom]
           [(scale-height image/C non-negative/C) #:custom]
           [(rotate       image/C angle/C)        #:custom]
           ;
           [(shrink    image/C) #:custom]
           [(enlarge   image/C) #:custom]
           ;
           [(thinner   image/C) #:custom]
           [(wider     image/C) #:custom]
           ;
           [(shorter   image/C) #:custom]
           [(taller    image/C) #:custom]
           ;
           [(counter   image/C) #:custom]
           [(clockwise image/C) #:custom]
           
           [(mirror    image/C) #:renames flip-horizontal]
           [(flip      image/C) #:renames flip-vertical]

           [(above   . images) check-images #:custom]
           [(beside  . images) check-images #:custom]
           [(overlay . images) check-images #:custom]

           [(render image/C) #:renames freeze #:custom]

           [(image? any/C) #:custom]

           [(url->image  text/C) [#:aka url→image] #:renames bitmap/url #:custom]

           [(image-left   image/C non-negative/C) check-within-width  #:renames left   #:custom]
           [(image-top    image/C non-negative/C) check-within-height #:renames top    #:custom]
           [(image-right  image/C non-negative/C) check-within-width  #:renames right  #:custom]
           [(image-bottom image/C non-negative/C) check-within-height #:renames bottom #:custom]

           [(text->image text . size-color) check:text→image [#:aka text→image] #:custom]
           ;
           [(circle   size . fill-color) check-1D-shape #:custom]
           [(square   size . fill-color) check-1D-shape #:custom]
           [(triangle size . fill-color) check-1D-shape #:custom]
           [(star     size . fill-color) check-1D-shape #:custom]
           [(rectangle width height . fill-color) check-2D-shape #:custom]
           [(ellipse   width height . fill-color) check-2D-shape [#:aka oval] #:custom]
           ;
           [(color? any/C) [#:aka colour?] #:custom]
           
           ; Legacy.
           [(cut-left   image/C non-negative/C) check-within-width  #:custom]
           [(cut-top    image/C non-negative/C) check-within-height #:custom]
           [(cut-right  image/C non-negative/C) check-within-width  #:custom]
           [(cut-bottom image/C non-negative/C) check-within-height #:custom]

           [(blank . dimensions) check-sizes #:custom]
           [(align-above  x-place          . images) check-horizontal-align          #:custom]
           [(align-beside          y-place . images) check-vertical-align            #:custom]
           [(align-overlay x-place y-place . images) check-horizontal-vertical-align #:custom]
           
           ; After Reading Week.
           [(red     . get/set) check-color-lens #:custom]
           [(green   . get/set) check-color-lens #:custom]
           [(blue    . get/set) check-color-lens #:custom]
           [(opacity . get/set) check-color-lens [#:aka α alpha] #:custom]

           [(colors? any/C) [#:aka colours?] #:custom]
           [(image->colors image/C) [#:aka image→colors image→colours image->colours] #:custom]
           [(colors->image colors/C natural/C natural/C)
            check-dimensions [#:aka colors→image colours→image colours->image] #:custom])
