#lang racket/base

#| Function (type) signature.

 Constrained by class  contracting : {(~or* f:id (f:id . _)) (~optional _:id)} .
 The  ~or*  cases are handled by  define/provide , where the latter case is parsed by class  signature
  which handles the variadic and non-variadic cases of that.

 Arity checking is automatic, except for any non-homogeneous part of a variadic rest parameter.
 The contract following the header is required for variadic, which must be binary to receive the
  function id and a list of *all* the arguments. |#

(module definer racket/base (provide definer)
  
  (require (only-in "C.rkt" /C-predicate))
  ;
  (require racket/lazy-require)
  (lazy-require [(submod "error.rkt" public) (check-arguments-contracts)])
  
  (require (submod "../../snack/syntax.rkt" core)
           (for-syntax (submod "../../snack/syntax.rkt" syntax-classes)))

  (define-macro (definer (f:id . formals) . _)
    #:with the-body (syntax-parse this-syntax
                      [(_ :variadic-formals c:id (~or* body:expr (~seq #:renames f′)))
                       #'{(c 'f (list* . arguments)) (~? body (apply f′ . arguments))}]
                      [(_ (f:id arg:id ...+) [~bracketed C:expr ...+] body:expr ...+)
                       #:with checks (for/list ([c (syntax->list #'(C ...))]
                                                [a (syntax->list #'(arg ...))]
                                                #:unless (equal? (syntax-e c) 'any/C))
                                       (list #'#%plain-app (list #'#%plain-app #'/C-predicate c) a))
                       (if (null? (syntax->list #'checks))
                           #'{body ...}
                           #'{(unless (and . checks)
                                (check-arguments-contracts 'f (list C ...) (list arg ...)))
                              body ...})])
    (begin (define (f . formals) . the-body) (provide f))))

(module help racket/base (provide contracting alias/rename/from define/provide)
  
  (require (for-template (submod ".." definer)
                         (only-in (submod "error.rkt" non-image)
                                  unary-predicate/C binary-predicate/C)
                         (submod "error.rkt" higher-order)
                         
                         racket/base (prefix-in base: (only-in racket/base apply #%app)))
           (submod "../../snack/syntax.rkt" core)
           (submod "../../snack/syntax.rkt" syntax-classes))

  (syntax-class (contracting #:splice {(~or* f:id (f:id . _)) (~optional _:id)}))

  (syntax-class ((alias/rename/from f there custom′)
                 #:splice {(~optional [~bracketed #:aka alias:id ...+])
                           (~optional (~seq #:renames original:id) #:defaults ([original f]))
                           (~optional
                            (~or* (~and #:custom (~bind [library (datum->syntax there custom′)]))
                                  (~seq #:from library)))}
                 #:with data #`{#,f original #,(temporary #'renamed-in) (~? [alias ...] [])}))

  (define-splicing-syntax-class signature #:attributes {contract/s app formals arguments checker}
    (pattern (~seq header:variadic-formals contract/s:id) #:with {{~optional checker}} '{}
             #:with app #'base:apply
             #:with formals #'header.formals #:with arguments #'header.arguments)
    (pattern (~seq (_:id contract:id ...+) (~optional checker:id))
             #:with app #'base:#%app #:with contract/s #'[contract ...]
             #:with formals (generate-temporaries #'contract/s) #:with arguments #'formals))
  
  (define (define/provide whole-specification library default binding-data)
    (syntax-parse binding-data
      [{f original imported-as {alias ...}}
       #:with import #`(local-require (only-in #,(or library default) [original imported-as]))
       (syntax-parse whole-specification
         [(_ _:id . _) #'(begin (provide f) import (define f (if (procedure? imported-as)
                                                                 (procedure-rename imported-as 'f)
                                                                 imported-as)))]
         [(_ :signature . _)
          (syntax-parse #'{contract/s arguments}
            #:literals {unary-predicate/C binary-predicate/C}
            [(~or* {((~or* (~and  unary-predicate/C (~bind [wrapper #'checked-unary-predicate]))
                           (~and binary-predicate/C (~bind [wrapper #'checked-binary-predicate])))
                     _ ...)
                    (p a ...)}
                   _)
             #:with (name ...) #'{f alias ...}
             #'(begin (definer (name . formals) contract/s (~? (checker 'name . arguments))
                        (begin import (~? (app imported-as (wrapper 'f p) a ...)
                                          (app imported-as . arguments))))
                      ...)])])])))

(module client racket/base (provide (all-from-out (submod "error.rkt" public))
                                    (for-syntax (all-from-out (submod ".." help)))
                                    (all-from-out (submod "../../snack/syntax.rkt" core)))
  (require (submod "error.rkt" public)
           (for-syntax (submod ".." help))
           (submod "../../snack/syntax.rkt" core)))

(module non-image racket/base (provide (for-syntax custom)
                                       (all-from-out (submod "error.rkt" non-image)
                                                     (submod ".." client)))
  (require (submod "error.rkt" non-image) (submod ".." client))
  (define-for-syntax custom '(submod "custom.rkt" non-image)))

(module image racket/base (provide (for-syntax custom) (all-from-out (submod "error.rkt" image)
                                                                     (submod ".." client)))
  (require (submod "error.rkt"     image) (submod ".." client))
  (define-for-syntax custom '(submod "custom.rkt" image)))
