#lang racket/base
(provide (struct-out range)
         (struct-out rectangle) build-rectangle join-rectangles*
         
         (rename-out (make-rectangular rectangular))
         trivial? intersect? shift resize
         width-height x-y x-y-width-height)

(require (only-in "struct.rkt" define/extract-struct))

(struct range (start end caret-space? color [rectangles #:mutable #;(or/c #f (listof rectangle))])
  #:transparent)

(struct rectangle (left top right bottom style color)
  #:transparent)
(define (build-rectangle left top right bottom style color)
  (unless (or (<= left right) (<= bottom top))
    (displayln (format "(~a ~a ~a ~a) ~a" left top right bottom))
    (local-require (only-in racket/gui/base message-box))
    (message-box "Nested Highlighting" "Bad `build-rectangle`."))
  (rectangle left top right bottom style color))

(define (join-rectangles* rectangular₀/false rectangles)
  (define-values (left₀ top₀ right₀ bottom₀)
    (if rectangular₀/false
        (x-y-width-height rectangular₀/false)
        (values #f #f #f #f)))
  (call-with-values
   (λ ()
     (for/fold ([left′ left₀] [top′ top₀] [right′ right₀] [bottom′ bottom₀])
       ([a-rectangle (in-list rectangles)])
       (define/extract-struct (left top right bottom) (rectangle a-rectangle))
       (if (and left′ top′ right′ bottom′)
           (values (min left′ left) (min top′ top) (max right′ right) (max bottom′ bottom))
           (values left top right bottom))))
   rectangular))


(struct rectangular (x y x′ y′) #:transparent)
(define (make-rectangular x y #:w [w #f] #:h [h #f] #:x [x′ #f] #:y [y′ #f])
  (rectangular x y (if x′ x′ (+ x w)) (if y′ y′ (+ y h))))

(require (only-in racket/match define/match match-define match-define-values))
(define/match (intersect? r₀ r₁)
  [((rectangular x₀ y₀ x₀′ y₀′) (rectangular x₁ y₁ x₁′ y₁′))
   (and (< x₀ x₁′) (< x₁ x₀′) (< y₀ y₁′) (< y₁ y₀′))])
(define (shift r δx δy)
  (match-define (rectangular x y x′ y′) r)
  (rectangular (+ x δx) (+ y δy) (+ x′ δx) (+ y′ δy)))
(define resize
  (case-lambda [(r from δw δh)
                (define-values (δx δy δx′ δy′)
                  (case from
                    [(corner₀) (values (- δw) (- δh) 0 0)]
                    [(centered) (values (/ δw -2) (/ δh -2) (/ δw 2) (/ δh 2))]
                    [(corner₁) (values 0 0 δw δh)]))
                (match-define (rectangular x y x′ y′) r)
                (rectangular (+ x δx) (+ y δy) (+ x′ δx′) (+ y′ δy′))]
               [(r from δ)
                (define-values (δx δy δx′ δy′)
                  (case from
                    [(x₀) (values (- δ) 0 0 0)]
                    [(y₀) (values 0 (- δ) 0 0)]
                    [(x₁) (values 0 0 δ 0)]
                    [(y₁) (values 0 0 0 δ)]))
                (match-define (rectangular x y x′ y′) r)
                (rectangular (+ x δx) (+ y δy) (+ x′ δx′) (+ y′ δy′))]))
(define/match (trivial? _) [((rectangular x y x′ y′)) (or (> x x′) (> y y′))])
(define/match (width-height _) [((rectangular x y x′ y′)) (values (- x′ x) (- y′ y))])
(define/match (x-y _) [((rectangular x y _ _)) (values x y)])
(define (x-y-width-height r)
  (match-define-values (x y) (x-y r))
  (match-define-values (width height) (width-height r))
  (values x y width height))

#;(define (intervals-overlap? l₀ r₀ l₁ r₁) (and (<= l₀ r₁) (<= l₁ r₀)))
