#lang racket/base

(define keybinding-request-message "Please enter a key specification:")
;
(define collapse-atomically:keybinding-change-label "Change Key for: Collapse Atomically")
;
(define-values (collapse-atomically:default-keybinding collapse-atomically:name)
  (values "c:s:space" "collapse sexp atomically"))

(define-values (collapse-each-compound:default-keybinding collapse-each-compound:name)
  (values "c:s:4" "collapse each compound sexp"))

(define-values (expand-one-level:default-keybinding expand-one-level:name)
  (values "c:s:5" "expand sexp one level"))

(define collapse-atomically:keybinding-pref 'drracket:collapse-atomic-keybinding)

(require "../../snack/preferences.rkt"
         "../../snack/lambda.rkt"
         (only-in "../sexp.rkt" expand-one-level collapse-each-compound toggle-collapse)
         (only-in "../../snack/keybinding.rkt"
                  add-binding/s global valid-keybinding? make-replaceable-keybinding))

(preferences:set-default ; Worth it to use ‘valid-keybinding?’ instead of ‘string?’ ?
 collapse-atomically:keybinding-pref collapse-atomically:default-keybinding string?)

(add-binding/s global
               collapse-each-compound:default-keybinding
               collapse-each-compound:name
               collapse-each-compound)
(add-binding/s global
               expand-one-level:default-keybinding
               expand-one-level:name
               expand-one-level)

#| Add an Edit menu item for changing the “Collapse Atomically” keybinding. |#

(provide tool@) (require racket/unit drracket/tool)
(define-unit tool@ (import drracket:tool^) (export drracket:tool-exports^)
  (define (phase1) (void))
  (define (phase2) (void))
  (local-require "../../snack/gui.rkt"
                 "../../snack/menu.rkt"
                 (submod "../../snack/class.rkt" implement))
  (drracket:get/extend:extend-unit-frame
   (mixin (drracket:unit:frame<%>) () (super-new)
     (define/override (edit-menu:between-find-and-preferences edit-menu)
       (super edit-menu:between-find-and-preferences edit-menu)
       (with-parent edit-menu
         (new separator-menu-item%)
         (new menu-item%
              [label collapse-atomically:keybinding-change-label]
              [callback (λ₂ (update-preference))])
         (new separator-menu-item%))))))

#| Keybinding preferences |#

(require (only-in racket/gui get-text-from-user)
         (except-in "../../snack/false.rkt" λ-case♯))

(define (update-preference)
  (let*/else
   ([choice (get-text-from-user collapse-atomically:keybinding-change-label keybinding-request-message
                                #false ; Should we set the parent?
                                (preferences:get collapse-atomically:keybinding-pref)
                                '(disallow-invalid) #:validate valid-keybinding?)]
    [#:else (void)])
   (update-keybinding choice)))

(define update-keybinding
  (let ([replace-keybinding
         (make-replaceable-keybinding global
                                      (preferences:get collapse-atomically:keybinding-pref)
                                      collapse-atomically:name
                                      toggle-collapse)])
    (λ (new-keybinding)
      (replace-keybinding new-keybinding)
      (preferences:set collapse-atomically:keybinding-pref new-keybinding))))
