#lang racket/base (provide compthink-keybindings@) (require drracket/tool racket/unit)

(require (submod "sexp.rkt" commands)
         (submod "text.rkt" commands)
         
         (only-in "drracket.rkt" more-definitions-or-interactions)
         (only-in "../snack/drracket.rkt" labels:definitions labels:interactions)
         "../snack/string-constants.rkt"

         (only-in "../snack/keybinding.rkt"
                  add-binding/s user remove-shortcut
                  alt/option mac-command default-modifier get-default-shortcut-prefix
                  find-keymap-for-function)
         (submod "../snack/keybinding.rkt" racket)
         
         (only-in "text.rkt" edit-sequence)
         "../snack/boolean.rkt"
         "../snack/function.rkt"
         "../snack/class.rkt"
         (submod "../snack/class.rkt" implement)
         racket/local

         (only-in "text.rkt" get-positions)
         (only-in "../snack/style.rkt" underlining)
         "../snack/string.rkt")

(define (drracket-menu-items edit-menu view-menu)

  ; Remove default shortcuts for Show/Hide of Definitions or Interactions ...
  (remove-shortcut view-menu labels:definitions)
  (remove-shortcut view-menu labels:interactions)
  ;  ... and bind to showing more definitions/interactions:
  (add-binding/s user (default-modifier "d") "show more of definitions"
                 (more-definitions-or-interactions #:definitions? #true))
  (add-binding/s user (default-modifier "e") "show more of interactions"
                 (more-definitions-or-interactions #:definitions? #false))
  
  ; Remove ‘ctrl-;’ Preferences shortcut from Edit menu (if it exists on the platform) so it can't
  ;  shadow our added “line comment/uncomment around selection” if that chooses ‘ctrl’.
  (remove-shortcut edit-menu (string-constant preferences-menu-item)))

(define (drracket-keybindings)

  (define add (curry add-binding/s user))
  
  ; Semicolon comment
  ; ≈≈≈≈≈≈≈≈≈≈≈≈≈≈≈≈≈
  (add (default-modifier "semicolon") "line comment/uncomment around selection" un/comment-selection)
  
  ; Sexp navigation and selection
  ; ≈≈≈≈≈≈≈≈≈≈≈≈≈≈≈≈≈≈≈≈≈≈≈≈≈≈≈≈≈
  ; Forward/Backward sexp
  ; —————————————————————
  (add (list (alt/option "right") "c:]") "go forward by sexp"  forward-sexp)
  (add (list (alt/option "left")  "c:[") "go backward by sexp" backward-sexp)
  ; Forward/Backward select sexp
  ; ————————————————————————————
  (add (list (alt/option "s:right") "s:c:]" "s:c:}" "c:}")
       "select forward by sexp"  select-forward-sexp)
  (add (list (alt/option "s:left")  "s:c:[" "s:c:{" "c:{")
       "select backward by sexp" select-backward-sexp)
  ; Hug
  ; ———
  (define-values (key:try-hug± name:try-hug±) (values "[" "try hug±"))
  (local [(define keymap:current (racket:get-keymap))
          (define keymap:try-hug± (new keymap:aug-keymap%))]
    (add-binding/s
     keymap:try-hug± key:try-hug± name:try-hug±
     (λ (ted event) (or (try-hug± ted event)
                        (let*♯ ([fn (hash-ref (send (racket:get-keymap) get-map-function-table)
                                              '|[| #false)])
                               (send (racket:get-keymap) call-function fn ted event #false)))))
    (send keymap:current chain-to-keymap keymap:try-hug± #true))
  
  ; Explode
  ; ———————
  (add '() "explode" explode)

  ; Auto-pairing
  ; ≈≈≈≈≈≈≈≈≈≈≈≈
  (add "s:|" "maybe insert #||# pair" maybe-insert-#||#-pair)
  (add "‘" "insert ‘’ around selection" (surrounder "‘" "’"))
  (add "“" "insert “” around selection" (surrounder "“" "”")))

(define (repl-return-keybindings keymap)
  (add-binding/s keymap "return"   "repl: submit if balanced" repl-maybe-submit)
  (add-binding/s keymap "s:return" "repl: insert newline"     repl-insert-newline))

(define (add-mac-history repl-keymap keymap)
  (send repl-keymap chain-to-keymap keymap #true)
  (let*♯ ([history-keymap (find-keymap-for-function repl-keymap "put-previous-sexp")])
         (send∗ history-keymap map-function
                [(mac-command "up")   "put-previous-sexp"]
                [(mac-command "down") "put-next-sexp"])))

(define-unit compthink-keybindings@ (import drracket:tool^) (export)

  (drracket:get/extend:extend-unit-frame
   (mixin (drracket:unit:frame<%>) () (super-new) (inherit get-edit-menu get-show-menu)
     (drracket-keybindings)
     (drracket-menu-items (get-edit-menu) (get-show-menu))))

  #| Interactions drracket:rep:text<%>. |#
  (drracket:get/extend:extend-interactions-text
   (mixin (drracket:rep:text<%>) () (super-new)
     
     (inherit get-keymap)
     (local [(define keymap (new keymap:aug-keymap%))]
       (repl-return-keybindings keymap)
       (let*♯ ([repl-keymap (get-keymap)]) (add-mac-history repl-keymap keymap)))

     (inherit get-start-position last-position)
     
     ; Automatic Formatting
     ; ————————————————————
     (inherit tabify-selection get-unread-start-point)
     (define (tabify-repl) (tabify-selection (get-unread-start-point) (last-position)))
     (define/override (do-paste start time) (super do-paste start time) (tabify-repl))
     (define/override (on-local-char event)
       (when (eqv? (send event get-key-code) #\return) (tabify-repl))
       (super on-local-char event))
     
     ; Prompt Text
     ; ———————————
     (define/override (get-prompt) "↪ ")

     ; Interactions Recording
     ; ——————————————————————
     (inherit get-user-thread get-top-level-window begin-edit-sequence end-edit-sequence change-style)
     
     (define (♯definitions)
       (define ♯tlw (get-top-level-window))
       (and (is-a? ♯tlw drracket:unit:frame<%>) (send ♯tlw get-definitions-text)))
     
     (define recording 'no)
     ;
     (define (style-prompt)
       (begin-edit-sequence #false #true)
       (define arrow-position (- prompt-position (string-length (get-prompt))))
       (when (positive? arrow-position)
         (change-style (underlining (case recording [(no) #false] [else #true]))
                       arrow-position (+ arrow-position 1)))
       (end-edit-sequence))
     ; ﹔
     (define/public (toggle-record)
       (begin-edit-sequence #false #true)
       (define prompt-start (- prompt-position (string-length (get-prompt))))
       (define allowing-edits? (send this get-allow-edits))
       (send this set-allow-edits #true)
       (send this delete prompt-start prompt-position)
       (set! recording (case recording [(no) 'comment] [else 'no]))
       (send this insert (get-prompt) prompt-start)
       (send this set-allow-edits allowing-edits?)
       (style-prompt)
       (end-edit-sequence))
     ;
     (define/augment (on-submit)
       (when (and (not (equal? recording 'no)) (get-user-thread) (thread-running? (get-user-thread)))
         (let*♯
          ([defs (♯definitions)])
          (define defs-last-position₀ (send defs last-position))
          (define-values (defs-start defs-end) (get-positions defs))
          (when (send defs can-insert? defs-last-position₀ (+ 1 (- (last-position) prompt-position)))
            (edit-sequence defs
              (unless (or (zero? defs-last-position₀)
                          (eqv? (send defs get-character (sub1 defs-last-position₀)) #\newline))
                (send defs insert "\n" defs-last-position₀ 'same))
              (send this copy-to defs prompt-position (last-position) (send defs last-position))
              (define defs-last-position₁ (send defs last-position))
              (cond [(equal? recording 'copy)
                     (send* defs
                       (tabify-selection defs-last-position₀ defs-last-position₁)
                       (scroll-to-position defs-last-position₁))]
                    [else (send defs set-position (add1 defs-last-position₀) defs-last-position₁
                                #false #false)
                          (un/comment-selection defs)
                          (send defs set-position defs-start defs-end
                                #false #false)])))))
       (inner (void) on-submit))
     ;
     (define prompt-position 0)
     ;
     ; Movement before Unread Start Point
     ; ——————————————————————————————————
     (inherit set-position)
     (define/override (insert-prompt) (super insert-prompt)
       (set! prompt-position (get-start-position))
       (style-prompt))
     (define/override (move-position code [extend? #false] [kind 'simple])
       (edit-sequence this
         (super move-position code extend? kind)
         (when (neither extend? ((get-unread-start-point) . < . (get-start-position)))
           (set-position (get-unread-start-point))))))))
