#lang racket/base (provide implode-whitespace
                           toggle-collapse expand-one-level collapse-each-compound
                           touching-sexps touching-sexp/false
                           (all-from-out "text.rkt"))

(require "text.rkt"
         (only-in framework racket:sexp-snip%)
         "../snack/class.rkt" (only-in (combine-in "../snack/function.rkt"
                                                   "../snack/match.rkt"
                                                   "../snack/boolean.rkt")
                                       match ¬ ⇒ neither))

#;(module+ main (require "../snack/repl.rkt") (repl racket racket/gui framework)) #;{ˢ𝑠ⓢ}

#| Event handlers for sexp navigation, that visit both the start and end of each sibling sexp,
    regardless of direction, and jump out to end/start of parent.

 See sexp.txt for details of jumping out.
 All are (racket-text%? → any) |#
;
(module* commands #false (provide #;remove-parens-backward
                                  try-hug± explode
                                  forward-sexp backward-sexp
                                  select-forward-sexp select-backward-sexp)

  ; Balanced open sexp delete.
  #;(add "s:backspace" "remove preceding balanced open delimiter" remove-parens-backward)
  #| Event handler to remove balanced delimeters that open immediately before the cursor.
    • analogous to DrRacket's remove-parens-forward |#
  #;(define-for-racket-text-event (remove-parens-backward a-racket-text) #;(racket-text%? → any)
      (define start (send a-racket-text get-start-position))
      (when (positive? start) (send a-racket-text remove-parens-forward (sub1 start))))
  
  #| At the exact start of a compound sexp, explode it, otherwise explode the enclosing sexp. |#
  (define-for-racket-text-event (explode rat)
    (define here (send rat get-start-position))
    (define-values (here⁺ _) (♯sexp-at⁺ rat here))
    (define start (cond [(and (equal? here here⁺) (compound-sexp? rat here)) here]
                        [else (define-values (♯↑-start ♯↑-end) (♯sexp-↑ rat here))
                              (and ♯↑-end ♯↑-start)]))
    (when start (edit-sequence rat
                  (let explode ([at start])
                    (define-values (_ ♯head-end) (♯sexp-↓ rat at))
                    (when ♯head-end (let loop ([previous-end ♯head-end])
                                      (define-values (♯start ♯end) (♯sexp-at⁺ rat previous-end))
                                      (when ♯end
                                        (explode ♯start)
                                        (define-values (_ new-end) (sexp-at♯ rat ♯start))
                                        (define-values (♯next-start ♯next-end) (♯sexp-at⁺ rat new-end))
                                        (when ♯next-end
                                          (when (same-line? rat new-end ♯next-start)
                                            (send rat insert #\newline ♯next-start))
                                          (loop new-end))))))
                  (define-values (_ new-end) (♯sexp-at⁺ rat start))
                  (define-values (♯next-start ♯next-end) (♯sexp-at⁺ rat new-end))
                  (cond [(and ♯next-end (same-line? rat new-end ♯next-start))
                         (send rat insert #\newline ♯next-start)
                         (define-values (♯↑-start ♯↑-end) (♯sexp-↑ rat start))
                         (if ♯↑-end
                             (send rat tabify-selection ♯↑-start ♯↑-end)
                             (send rat tabify-selection start (+ ♯next-start 1)))]
                        [else (send rat tabify-selection start new-end)])
                  (send rat set-position start))))
  
  (define tabify-limit 1000)
  
  (define (try-hug± rat [_ (void)])
    
    (define (tabify-short start end)
      (when (< (- end start) tabify-limit) (send rat tabify-selection start end)))
    
    (and (racket-text%? rat)
         (let*♯ ([here (unselected-position rat)])
                (define-values (𝑠⁺:start _) (♯sexp-at⁺ rat here))
                (and 𝑠⁺:start (cond [(sexp-first? rat 𝑠⁺:start)
                                     (define-values (𝑠↖:start 𝑠↖:end) (♯sexp-↑ rat 𝑠⁺:start))
                                     (and 𝑠↖:end (edit-sequence rat
                                                   (send rat remove-parens-forward 𝑠↖:start)
                                                   (tabify-short 𝑠↖:start (- 𝑠↖:end 2))))]
                                    [(𝑠⁺:start . <= . here)
                                     (define-values (𝑠⁺⁺:start 𝑠⁺⁺:end) (♯sexp-at⁺⁺ rat here))
                                     (define argument?
                                       (neither (send rat is-stopped?)
                                                (eq? 'symbol (send rat classify-position 𝑠⁺⁺:start))))
                                     (edit-sequence rat
                                       (send∗ rat
                                              (insert ")" 𝑠⁺⁺:end)
                                              (insert (if argument? "( " "(") 𝑠⁺⁺:start)
                                              (set-position (+ 𝑠⁺⁺:start 1)))
                                       (tabify-short 𝑠⁺⁺:start (+ 𝑠⁺⁺:end 2)))]
                                    [else #false])
                     #true))))
  
  #;{(define/public (remove-parens-forward start-pos)
       (let* ([pos (skip-whitespace start-pos 'forward #f)]
              [first-char (get-character pos)]
              [paren? (or (char=? first-char #\() (char=? first-char #\[) (char=? first-char #\{))]
              [closer (and paren? (get-forward-sexp pos))])
         (if (and paren? closer)
             (begin (begin-edit-sequence #t #f)
                    (delete pos (add1 pos)) (delete (- closer 2) (- closer 1))
                    (end-edit-sequence))
             (bell))
         #t))
     (define/public (kill-enclosing-parens begin-inner)
       (begin-edit-sequence #t #f)
       (define begin-outer (find-up-sexp begin-inner))
       (cond [begin-outer
               (define end-outer (get-forward-sexp begin-outer))
               (cond [(and end-outer (>= (- end-outer begin-outer) 2))
                      (delete (- end-outer 1) end-outer) (delete begin-outer (+ begin-outer 1))
                      (tabify-selection begin-outer (- end-outer 2))]
                     [else (bell)])]
             [else (bell)])
       (end-edit-sequence))}
  #;{(define-values (key:open command-name:open) (values "[" "maybe-insert-smart-open"))
     (local [(define delegate-name "maybe-insert-[]-pair-maybe-fixup-[]")
             (define paren-keymap (find-keymap-for-function (racket:get-keymap) delegate-name))
             (define paren-space-keymap (new keymap:aug-keymap%))]
       #:side-effects
       (add-binding/s
        paren-space-keymap key:open command-name:open
        (λ (racket-text event)
          (and (text%? racket-text)
               (local [(define-values (start end) (get-positions racket-text))]
                 #:side-effects
                 #;(send racket-text begin-edit-sequence #true #true)
                 (define handled? (send paren-keymap call-function delegate-name racket-text event))
                 (when (and handled?
                            (= start end)
                            (= (send racket-text get-start-position) (+ start 1))
                            (text-equal? racket-text "()" start))
                   ; “Note that, for a text% object, the character-inserting version
                   ;   of insert interferes with sequence-based undo groupings.”
                   (send racket-text begin-edit-sequence #true #false)
                   (send racket-text insert " ")
                   (send racket-text set-position (+ start 1))
                   (send racket-text end-edit-sequence))
                 #;(send racket-text end-edit-sequence)
                 handled?))))
       (send paren-keymap chain-to-keymap paren-space-keymap #true))}
  
  (define-for-racket-text-event (forward-sexp a-racket-text)
    (define here (send a-racket-text get-start-position))
    (define-values (𝑠⁺:start 𝑠⁺:end) (♯sexp-at⁺ a-racket-text here))
    (if 𝑠⁺:start
        (send a-racket-text set-position (if (here . < . 𝑠⁺:start) 𝑠⁺:start 𝑠⁺:end))
        (send♯/with a-racket-text (forward-sexp (find-up-sexp here)))))
  ;
  (define-for-racket-text-event (backward-sexp a-racket-text)
    (define here (send a-racket-text get-start-position))
    (define-values (𝑠⁻:start 𝑠⁻:end) (♯sexp-at⁻ a-racket-text here))
    (if 𝑠⁻:end
        (send a-racket-text set-position (if (𝑠⁻:end . < . here) 𝑠⁻:end 𝑠⁻:start))
        (send a-racket-text up-sexp here)))
  ;
  (define-for-racket-text-event (select-forward-sexp a-racket-text)
    (define-values (start end) (get-positions a-racket-text))
    (edit-sequence a-racket-text
      (send a-racket-text set-position end) ; extend end of selection
      (forward-sexp a-racket-text)
      (send a-racket-text set-position start (send a-racket-text get-start-position))))
  ;
  (define-for-racket-text-event (select-backward-sexp a-racket-text)
    (define end (send a-racket-text get-end-position))
    (edit-sequence a-racket-text
      (backward-sexp a-racket-text)
      (send a-racket-text set-position (send a-racket-text get-start-position) end))))

#| === |#

#;{Precondition:  (racket-text-and-position? rat position)}
#;(define (racket-text-and-position? t p)
    (and (racket-text%? t) (natural? p) (p . ≤ . (send t last-position))))
#;{Postcondition: (result . ⇒ . (racket-text-and-position? rat result))}

#| These don't default to get-start-position because in that case current uses usually have/need
    that position as well. |#

(define (sexp-at♯ rat ♯start) (values ♯start (send♯ rat get-forward-sexp ♯start)))
(define (♯sexp-at⁻ rat p) (sexp-at♯ rat (send rat get-backward-sexp p)))
(define (♯sexp-at⁺ rat p)
  (define ♯end (send rat get-forward-sexp p))
  (values (send♯ rat get-backward-sexp ♯end) ♯end))
(define (♯sexp-↑ rat p) (sexp-at♯ rat (send rat find-up-sexp p)))
(define (♯sexp-↓ rat p)
  (if (compound-sexp? rat p) (sexp-at♯ rat (send rat find-down-sexp p)) (values #false #false)))

(define (♯sexp-at⁺⁺ rat p)
  (define ♯end (send rat get-forward-sexp p))
  (values (and ♯end (♯back rat ♯end)) ♯end))
(define (♯back rat p)
  (define limit (send rat get-limit p))
  (let*♯ ([end-pos (send rat backward-match p limit)])
         (define min-pos (send rat backward-containing-sexp p limit))
         (and (min-pos . ⇒ . (end-pos . >= . min-pos)) end-pos)))

; Is the ♯sexp at⁺ position compound?
(define (compound-sexp? rat p)
  ; Next open is balanced, inside exists and is inside (i.e. before end of) this sexp.
  ; • find-down-sexp insufficient since it moves forward to find a compound sexp to descend into
  (let*♯ ([down (send rat find-down-sexp p)]) (down . < . (send rat get-forward-sexp p))))

#| ——— |#

; racket-text% natural → boolean
(define (sexp-first? racket-text 𝑠:start)
  (equal? 𝑠:start (send♯/with racket-text (find-down-sexp (find-up-sexp 𝑠:start)))))

#| Position of start of sexp “touching” position, using the earlier if two, otherwise #false.
 • a position doesn't touch a compound sexp from inside
 • any “prefix” of an sexp is touchable anywhere inside the prefix, including whitespace and comments

 {⋯∘⋯, (⋯)∘, ⋯∘} ⇔ valid previous-end ∈ {⋯⋯∘, (⋯)∘, ⋯∘} ⇒ use previous-start ∈ {∘⋯⋯, ∘(⋯), ∘⋯}
  otherwise, {∘~(⋯), ∘~⋯ } ⇔ valid next-start, then { ∘(⋯) ∘⋯ } ⇔ position is next-start |#
;
(define (touching-sexp/false a-racket-text position)
  (define-values (previous-start previous-end) (♯sexp-at⁻ a-racket-text position))
  (if (and previous-end (position . <= . previous-end))
      previous-start
      (let-values ([(next-start _) (♯sexp-at⁺ a-racket-text position)])
        (and next-start (next-start . = . position) position))))

#| Return two values:
    • start of sexp touching start, or just start if not applicable
    • end   of sexp touching end,   or just end   if not applicable |#
(define (touching-sexps a-racket-text start end)
  (values (or (touching-sexp/false a-racket-text start) start)
          (or (send♯ a-racket-text get-forward-sexp (touching-sexp/false a-racket-text end)) end)))

#| === |#

#;{(require racket/stream "iteration.rkt")
   (define (sexp-ends a-racket-text position)
     (define next (send a-racket-text get-forward-sexp position))
     (if next (stream-cons next (sexp-ends a-racket-text next)) empty-stream))
   (define (stream-findf f s) (for/first e (in-stream s) #:when (f e)))}

#| ‘implode-whitespace’ |#
(define-for-text-event (implode-whitespace a-text)
  (let*/else
   ([start (unselected-position a-text)] [#:else (send a-text delete)])

   (define-values (white-start white-end) (values (add1 (find-non-whitespace a-text (sub1 start) -1))
                                                  (find-non-whitespace a-text start +1)))
   (define-values (first-paragraph last-paragraph) (paragraphs a-text white-start white-end))
      
   (define first-paragraph-start (send a-text paragraph-start-position first-paragraph))
   (define prefix-white-paragraph
     (if (= first-paragraph-start white-start) first-paragraph (add1 first-paragraph)))

   (define last-paragraph-end (send a-text paragraph-end-position last-paragraph))
   (define suffix-white-paragraph
     (if (= last-paragraph-end white-end) last-paragraph (sub1 last-paragraph)))

   (edit-sequence a-text
     (define changed? #false)
     (cond [(and (< white-start white-end) (member (send a-text get-character white-end)
                                                   '(#\) #\] #\} #\|)))
            (send a-text delete white-start white-end) (set! changed? #true)] 
           [(and (< 0 white-start start) (member (send a-text get-character (sub1 white-start))
                                                 '(#\( #\[ #\{ #\|)))
            (send a-text delete white-start white-end)
            (set! changed? #true)])
     (unless (or changed? (and (racket-text%? a-text)
                               (send a-text tabify white-end)
                               (not (= (find-non-whitespace a-text white-start +1) white-end))))
       (cond [(<= prefix-white-paragraph suffix-white-paragraph)
              (define-values (start end)
                (paragraph-positions a-text suffix-white-paragraph))
              (send a-text delete start (add1 end))
              (set! changed? #true)]
             [(< (add1 white-start) white-end)
              (send a-text insert #\space white-start white-end)
              (set! changed? #true)]))
     (when (racket-text%? a-text)
       #;(define paragraph-end
           (send/with a-text (paragraph-end-position (position-paragraph white-start))))
       (define-values (spanning-sexp-start spanning-sexp-end)
         (♯sexp-↑ a-text start #;paragraph-end))
       (when spanning-sexp-end #;(app♯ < paragraph-end spanning-sexp-end)
         (send a-text tabify-selection spanning-sexp-start spanning-sexp-end))))))

#| === |#

; Collapse/expand s-expression touching the cursor, or
;  if a selection then collapse all selected s-expressions together as a unit.
(define-for-racket-text-event (toggle-collapse a-racket-text)
  (define-values (start-pos end-pos) (get-positions a-racket-text))
  (let*♯
   ([start-sexp (touching-sexp/false a-racket-text start-pos)]
    [end-sexp   (send a-racket-text get-forward-sexp start-sexp)])
   (define-values (start end) (cond [; Toggle sexp touching the cursor.
                                     (= start-pos end-pos) (values start-sexp end-sexp)]
                                    [; For selections assume start-pos and end-pos are in siblings.
                                     ; Expand selection to include both siblings, always collapse.
                                     else (touching-sexps a-racket-text start-pos end-pos)]))
   (if (collapsed? a-racket-text start end)
       (expand-sexp-snip a-racket-text (send a-racket-text find-snip end 'before))
       (collapse a-racket-text start end))))


; Expand collapsed sexp touching cursor by one level: expand it, then collapse its immediate children.
(define-for-racket-text-event (expand-one-level a-racket-text)
  (define start-pos (send a-racket-text get-start-position))        
  (let*♯
   ([start-sexp (touching-sexp/false a-racket-text start-pos)]
    [end-sexp (send a-racket-text get-forward-sexp start-sexp)])
   (when (collapsed? a-racket-text start-sexp end-sexp)
     (edit-sequence a-racket-text
       (expand-sexp-snip a-racket-text (send a-racket-text find-snip end-sexp 'before))
       ; Collapse all immediate children, backwards to avoid shifting positions.
       (collapse-compounds
        a-racket-text
        start-sexp
        (send♯ a-racket-text get-backward-sexp
               (get-forward*-sexp a-racket-text (send a-racket-text find-down-sexp start-sexp))))
       ; Move cursor to end of last s-expression within newly-expanded expression.
       ; Expansion moved it to the end, and collapsing keeps it current, so move to beggining,
       ;  then in, then all the way forward.
       (send♯/with a-racket-text (backward-sexp (get-start-position)))
       (send♯/with a-racket-text (down-sexp (get-start-position)))
       (send a-racket-text set-position (get-forward*-sexp a-racket-text))))))

; Collapse all compound sexps touching the selection, separately.
(define-for-racket-text-event (collapse-each-compound a-racket-text)
  (define-values (start-pos end-pos) (touching-sexps a-racket-text
                                                     (send a-racket-text get-start-position)
                                                     (send a-racket-text get-end-position)))
  (edit-sequence a-racket-text
    (collapse-compounds a-racket-text start-pos (send a-racket-text get-backward-sexp end-pos))))

#| === |#

(define (bracket? char) #;(any/c . -> . any/c) (memq char '(#\( #\{ #\[ #\) #\} #\])))

(define (matching-open-delimiter/nul char) #;(any/c . -> . (or/c #\( #\[ #\{ #\nul))
  (match char [#\) #\(] [#\] #\[] [#\} #\{] [_ #\nul]))

#| === |#

; collapsed? : text%? natural? natural? → boolean?
; Whether start and end are the start and end of a single collapsed snip.
; Typically used with sexps so usually a-text : racket-text%?.
; Presumably start and end ∈ [0, last-position].
(define (collapsed? a-text start end)
  (let*♯ ([snip-at-start (send a-text find-snip start 'after-or-none)]
          [snip-at-end   (send a-text find-snip end   'before-or-none)])
         (and (eq? snip-at-start snip-at-end) (is-a? snip-at-start racket:sexp-snip%))))

#| === |#

; The end position of the last sibling sexp from the current or given position onward.
(define (get-forward*-sexp text [position (send text get-start-position)])
  (let loop ([current position] [previous position])
    (if current (loop (send text get-forward-sexp current) current) previous)))

; For a position at the start of an S-expression: is it atomic?
(define atomic-sexp? (¬ compound-sexp?))

#| === |#

; Mostly copied from gui-lib/framework/private/racket.rkt.
(define (collapse text left-pos right-pos)
  (when (and left-pos right-pos)
    (let* ([right (send text get-character (sub1 right-pos))]
           ; If the selection is a compound sexp, frame the collapsed snip with its brackets.
           ; Decide if a selection is a compound sexp by looking at the right bracket only,
           ;  because prefixes such as ‘#;’ are counted as parts of sexps.
           [left-bracket (matching-open-delimiter/nul right)]
           [right-bracket (if (bracket? right) right #\nul)])
      (edit-sequence text
        (send∗ text split-snip [left-pos] [right-pos])
        (let ([snips (let loop ([snip (send text find-snip left-pos 'after-or-none)])
                       (if (snip . ⇒ . ((send text get-snip-position snip) . >= . right-pos))
                           null
                           (cons (send snip copy) (loop (send snip next)))))])
          (send∗ text
                 [delete left-pos right-pos]
                 [insert (new racket:sexp-snip% 
                              [left-bracket left-bracket]
                              [saved-snips snips]
                              [right-bracket right-bracket])
                         left-pos left-pos]))))))

; racket-text%? (is-a?/c racket:sexp-snip<%>) → any
(define (expand-sexp-snip a-text snip)
  (let ([snips (send snip get-saved-snips)])
    (edit-sequence a-text (let ([pos #;(racket-text%? (is-a?/c snip%) → (or/c #false natural?))
                                     ; #false if not in the text
                                     (send a-text get-snip-position snip)])
                            (send a-text delete pos (add1 pos))
                            (for ([snip (in-list (reverse snips))])
                              (send a-text insert (send snip copy) pos pos))))))

; racket-text%? natural? natural? → any
; Collapse each compound sibling sexp which starts in [start-pos, end-pos].
; Precondition: pos is at the front of an s-expression.
(define (collapse-compounds text start-pos end-ps)
  (when (and end-ps (end-ps . >= . start-pos))
    (let ([end (send text get-forward-sexp end-ps)])
      (unless (or (collapsed? text end-ps end) (atomic-sexp? text end-ps))
        (collapse text end-ps end)))
    (collapse-compounds text start-pos (send text get-backward-sexp end-ps))))
