#lang typed/racket/base/no-check

(provide fine-format)
(: fine-format : (Instance Racket:Text%) → AnyValues)
(define (fine-format a-text) (fix-dangling-parentheses a-text) (space-out a-text))

(module interact typed/racket/base/no-check

  (provide show-change)
  
  (struct configuration
    ([highlight-style : (U 'dot 'ellipse 'hollow-ellipse 'rectangle)]
     [highlight-color : (Instance Color%)]
     [dialog-title : String] [dialog-message : String]
     [dialog-cancel : String] [dialog-skip : String] [dialog-ok : String]))

  (define (the-color [alpha-fraction : Real])
    (define sl (editor:get-standard-style-list))
    (define s (and sl (send sl find-named-style "framework:syntax-color:scheme:error")))
    (define c (and s (send s get-foreground)))
    (cond [c (make-color (send c red) (send c blue) (send c green) 
                         alpha-fraction)]
          [else (make-color 128 128 128 alpha-fraction)]))

  (define the-configuration
    (configuration 'hollow-ellipse (the-color .5)
                   "Formatting" "Adding a space between parts." "Cancel" "Skip" "Ok"))

  (require (only-in typed/racket/gui/base message-box/custom make-color Color% frame% dialog%)
           (only-in typed/framework editor:get-standard-style-list Racket:Text%)
           (only-in typed/racket/class send is-a?)
           (only-in racket/match match-define))
  
  (define (show-change [a-text : (Instance Racket:Text%)]
                       [position-1 : Integer] [position-2 : Integer])
    (match-define (configuration style color title message cancel skip ok) the-configuration)
    (define (clamp [position : Integer]) (min (max 0 position) (send a-text last-position)))
    (define canvas♯ (send a-text get-canvas))
    (define window♯ (and canvas♯ (send canvas♯ get-top-level-window)))
    (define frame/dialog♯ (and (or (is-a? window♯ frame%) (is-a? window♯ dialog%))
                               window♯))
    (let ([start (clamp position-1)] [end (clamp position-2)])
      (send a-text scroll-to-position start #false end)
      (define un-highlight (send a-text highlight-range start end color #false 'low style))
      (define response (message-box/custom title message cancel skip ok frame/dialog♯))
      (or (void? un-highlight) (un-highlight))
      (case response [(2) 'skip] [(3) 'ok] [else 'cancel]))))

(require (only-in 'interact show-change)
         "../snack/boolean.rkt"
         (only-in "../snack/class.rkt" send♯/with let*♯)
         
         #;(only-in typed/framework racket:text% Racket:Text%)
         
         (only-in racket/string string-trim)
         (only-in typed/racket/class send is-a?))

(define (fix-dangling-parentheses [a-text : (Instance Racket:Text%)])
  (loop-over a-text (λ (position forward-position)
                      (unless forward-position (pull-end a-text position))
                      forward-position)))

(: pull-end : (Instance Racket:Text%) Integer → Any)
(define (pull-end a-text position)
  (let*♯ ([parent-end (send♯/with a-text (get-forward-sexp (find-up-sexp position)))]
          [inside-parent-end (sub1 parent-end)])
         (when (and (< position inside-parent-end)
                    (whitespace? (send a-text get-text position inside-parent-end)))
           (send a-text delete position inside-parent-end))))
;
(define (whitespace? [s : String]) (equal? (string-trim s) ""))


(: space-out : (Instance Racket:Text%) → Any)
(define (space-out a-text)
  (let/ec exit : Any
    (loop-over
     a-text (λ (position forward-position)
              (cond [(forward-position . ⇒ . (separated? a-text forward-position)) forward-position]
                    [else (case (show-change a-text (sub1 forward-position) (add1 forward-position))
                            [(ok) (send a-text insert #\space forward-position)
                                  (add1 forward-position)]
                            [(skip) forward-position]
                            [(cancel) (exit (void))]
                            [else (typecheck-fail #'here "should be unreachable")
                                  #;(ann (values) : Nothing)])])))))
;
(define (separated? [a-text : (Instance Racket:Text%)] [position : Integer])
  (define next-start (send♯/with a-text (get-backward-sexp (get-forward-sexp position))))
  (next-start . ⇒ . (position . < . next-start)))


(: loop-over : (Instance Racket:Text%) (Integer (U Integer False) → (U Integer False)) → Any)
(define (loop-over a-text proc)
  (let loop ([position 0])
    (let*♯ ([down-position    (send a-text   find-down-sexp position)]
            [forward-position (send a-text get-forward-sexp position)])
           (when (< down-position forward-position) (loop down-position)))
    (let ([forward-position (send a-text get-forward-sexp position)])
      (let*♯ ([forward-position (proc position forward-position)])
             (loop forward-position)))))
  
(define (<♯ v1 v2 . vs) (and (real? v1) (real? v2) (andmap real? vs) (apply < v1 v2 vs)))
