#lang racket/base

(require syntax/parse/define
         (for-syntax racket/base
                     (only-in racket/syntax format-id)
                     (only-in syntax/stx stx-map)))

(provide define-settings-struct)

#| Defines four names:
   - A struct called struct-id, with parent struct super-id and added fields field-ids.
   - A function make-struct-id, which takes one parameter for each field-id and creates an
     instance of struct-id with each field of super-id initialized to its corresponding
     default-super-value, and the field-ids initialized to the provided values.
   - A function struct-id->hash, which takes an instance of struct-id and returns an
     immutable hash mapping each field-id to its value in the parameter.
   - A function hash->struct-id, which takes an immutable hash as produced by struct-id->hash
     and returns an instance of struct-id with each field of super-id initialized to its
     corresponding default-super-value, and each field-id initialized to its value in the hash.

   Each default-value and default-super-value is evaluated only once, at the time of the
   definition. |#

(define-syntax-parser define-settings-struct
  [(_ struct-id:id super-id:id ([field-id:id default-value:expr] ...)
      default-super-value:expr ...)
   (with-syntax ([(field-accessor ...) (stx-map (λ (stx) (format-id stx "~a-~a" #'struct-id stx))
                                                #'(field-id ...))]
                 [(super-value-id ...) (stx-map (λ (_) #`#,(gensym)) #'(default-super-value ...))]
                 [make-id (format-id #'struct-id "make-~a" #'struct-id)]
                 [to-hash-id (format-id #'struct-id "~a->hash" #'struct-id)]
                 [from-hash-id (format-id #'struct-id "hash->~a" #'struct-id)])
     (define field-pairs
       (apply append (stx-map syntax->list #'(('field-id (field-accessor s)) ...))))
     #`(begin (struct struct-id super-id (field-id ...))
              (define make-id
                (let ([super-value-id default-super-value] ...)
                  (λ (field-id ...) (struct-id default-super-value ... field-id ...))))
              (define (to-hash-id s) (hash #,@field-pairs))
              (define from-hash-id
                (let ([field-id default-value] ...)
                  (λ (s) (make-id (hash-ref s 'field-id default-value) ...))))))])
