#lang racket/base

(provide racket-text%? racket:text% define-for-racket-text-event racket:text<%>
         racket:text-balanced? #;([(is-a?/c text%)] [number? (or/c false/c number?)] . ->* . boolean?)
         text% define-for-text-event
         edit-sequence
         unselected-position get-positions paragraph-positions paragraphs same-line?
         text-equal? in-text (all-from-out "../snack/iteration.rkt") find-non-whitespace
         selected-syntax/false
         (all-from-out "../snack/class.rkt"))
;
(require (only-in framework racket:text% racket:text<%> racket:text-balanced? text:basic<%>)
         (only-in racket/port with-input-from-string port->list)
         (only-in racket/gui/base text% frame% snip%)
         racket/require
         (multi-in "../snack" ("class.rkt" "iteration.rkt"))
         "../snack/lambda.rkt"
         (except-in "../snack/false.rkt" λ-case♯)
         (submod "../snack/syntax.rkt" core))

#;(module+ main (require (snack: repl)) (repl racket framework racket/class))

; racket:text<%> : interface?  implements : text:basic<%> mode:host-text<%> color:text<%>
; implode-whitespace.rkt sexp.rkt self
(define (racket-text%? v) (is-a? v racket:text<%>))
; self
(define (text%?        v) (is-a? v text%))

#| Define safe event handlers for texts or racket texts.
 Safe for any value and optional ignored second argument.
 If the target is of the right type, then this produces non-false to signal the event was handled,
  otherwise produces #false to allow other handlers if appropriate. |#
; sexp.rkt self
(define-macro (define-for-text-event (f:id text:id) body:expr ...+)
  (define (f text [_ (void)]) (and (text%? text) (let () body ... (void)))))
; sexp.rkt self
(define-macro (define-for-racket-text-event (f:id racket-text:id) body:expr ...+)
  (define (f racket-text [_ (void)]) (and (racket-text%? racket-text) (let () body ... (void)))))

(module* commands #false (provide surrounder un/comment-selection
                                  maybe-insert-#||#-pair
                                  repl-maybe-submit repl-insert-newline)
  (require racket/contract "../snack/function.rkt")
  (define (surrounder open close)
    (define insertable/c (or/c (is-a?/c snip%) string? char?))
    (define/contract (surround ted open close) (text%? insertable/c insertable/c . -> . void?)
      (define start (send ted get-start-position)) ; → natural?
      (edit-sequence ted (send∗ ted
                                (set-position (send ted get-end-position)) ; → natural?
                                (insert close)
                                (set-position start) ; necessary?
                                (insert open))))
    (define-for-text-event (the-surround ted) (surround ted open close))
    the-surround)
  
  (define-for-racket-text-event (un/comment-selection racket-text)
    (define-values (start end) (get-positions racket-text))
    ; For completely-selected end line, i.e. selection at start of its next line, don't include next.
    (define-values (first-paragraph last-paragraph)
      (paragraphs racket-text start (max start (sub1 end)))) ; in particular, max avoids error at 0 0
    (if (for/or ([paragraph (in-range first-paragraph (add1 last-paragraph))])
          ; Starts with non-‘;’?
          (define-values (paragraph-start paragraph-end) (paragraph-positions racket-text paragraph))
          (let*♯ ([non-blank (find-first (¬ char-blank?)
                                         (in-text racket-text paragraph-start paragraph-end))])
                 (not (equal? non-blank #\;))))
        ; racket:text<%>
        (send racket-text comment-out-selection start end)
        (send racket-text uncomment-selection   start end)))

  ; Add a space between auto-completed block comment delimiters.
  (define (maybe-insert-#||#-pair editor evt)
    (let*/else ([km (send editor get-keymap)] [#:else (void)])
               (edit-sequence editor
                 (cond [(text%? editor)
                        (define end (send editor get-end-position))
                        (send km call-function "maybe-insert-||-pair" editor evt #true)
                        (define position (send editor get-start-position))
                        ; If this is (probably) a block comment and was auto-balanced, we inserted
                        ;  one "|" before the selection and then "|#" after, so look for that "#".
                        (when (equal? (send editor get-character (+ end 2)) #\#)
                          (send∗ editor (insert #\space (add1 end)) (set-position position)))]
                       [else (send km call-function "maybe-insert-||-pair" editor evt #true)]))))

  ; If Interactions contains a complete sexp, submit it for evaluation, otherwise insert newline.
  (define-for-racket-text-event (repl-maybe-submit racket-text) ; much is from text:ports<%> api
    (if (racket:text-balanced? racket-text (send racket-text get-unread-start-point))
        (send∗ racket-text (insert "\n" (send racket-text last-position)) (do-submission))
        (repl-insert-newline racket-text)))
  ; Forced newline.
  (define-for-racket-text-event (repl-insert-newline racket-text) (send racket-text insert-return)))

; text%? string? natural? [natural?] → boolean
; implode-whitespace.rkt
(define (text-equal? ted text start [end (+ start (string-length text))])
  (and (= (string-length text) (- end start))
       ; get-text: no restrictions beyond natural for start and end, “invalid” produces ""
       (equal? text (send ted get-text start end))))

#;(text%? . → . (values natural? natural?))
(define (get-positions ted) (values (send ted get-start-position) (send ted get-end-position)))
#;(text%? . → . natural?)
(define (same-line? ted p₀ p₁) (= (send ted position-paragraph p₀)
                                  (send ted position-paragraph p₁)))

#| (racket:text-balanced? text [start end]) → boolean?
  text : (is-a?/c text%)
  start : number? = 0
  end : (or/c false/c number?) = #f |#

(define (selected-syntax/false text)
  (define-values (start end) (get-positions text))
  (and (racket:text-balanced? text start end)
       (let [(source-string (send text get-text start end))]
         (fail→false (list source-string (parameterize ((port-count-lines-enabled #true))
                                           (with-input-from-string source-string
                                             (λ₀ (port->list read-syntax)))))))))

(define (in-text text start end)
  (local-require (only-in racket/stream for/stream))
  (for/stream ([position (in-range start end)]) (send text get-character position)))

(define-macro (edit-sequence editor:expr body:expr ...+) ; preserve result?
  (let ([editor-value editor])
    (send editor-value begin-edit-sequence) body ... (send editor-value end-edit-sequence)))

#;(text%? . -> . any/c)
(define (unselected-position text)
  (define-values (start end) (get-positions text))
  (and (= start end) start))

#;(text%? natural? . -> . (values natural? natural?))
; If beyond the last paragraph, the last paragraph is used.
;  The methods also have an optional “visibility” flag.
(define (paragraph-positions text paragraph) (values (send text paragraph-start-position paragraph)
                                                     (send text paragraph-end-position   paragraph)))
;
(define (paragraphs racket-text start end) (values (send racket-text position-paragraph start)
                                                   (send racket-text position-paragraph end)
                                                   #;(send a-racket-text calc-last-para end)))

; Intended preconditions:
;   • a-text : text%?
;   • position : natural? ∈ [-1, last-position], which is inspected regardless of Δ
;   • Δ : (or/c -1 1)
;  Then return exact-integer? ∈ [-1, last-position].
(define (find-non-whitespace a-text position Δ)
  (define last (send a-text last-position)) #;natural?
  (let loop ([position position])
    (if (and (< -1 position last) (char-whitespace?
                                   #;(natural? . -> . char?) ; #\nul if at/past last-position
                                   (send a-text get-character position)))
        (loop (+ position Δ))
        position)))
