#lang scribble/manual

@(provide the-documented-variables the-documented-functions the-documented-forms)
@(require (for-syntax syntax/parse racket/base racket/local racket/function racket/math racket/list)
          racket/local)

@; The main audience is diverse, but in principle an interested precise student.

@; @-expression commenting-out:
@;   @;{ ‹any›* } : analogous to combination of expression and block comment
@;     • nests properly
@;     • body must parse correctly
@;   @; ‹anything-else-without-newline›*
@;      • also consumes leading spaces and tabs on the next line, apparently in analogy to TeX's %.

@; ★★★ Content Directed at Myself or Instructors ★★★
@; There are also temporary notes, until I decide exactly how I want to provide this material
@;  tailored to different audiences. Make a note by wrapping content with ‘self’ or ‘instructor’.
@;  They can be easily be omitted from the built documentation by switching which expression is
@;  commented out in each form's implementation.
@;
@(define-syntax-rule (self pre-content ...)
   "" #;(list "\n" (bold "Self note. ") pre-content ... "\n"))
@(define-syntax-rule (instructor pre-content ...)
   "" #;(list "\n" (bold "Instructor Note. ") pre-content ... "\n"))
@(define (inset . stuff) (apply #:style 'inset nested stuff))
@(define-syntax-rule (instructor-inset pre-content ...) (instructor (inset pre-content ...)))
@(define-syntax-rule (self-inset pre-content ...) (self (inset pre-content ...)))
@(define (itemlist-inset . items) (inset (apply itemlist items)))

@; ★★★ Title ★★★
@title{Computational Thinking : CSC104 (Fall@|~|2018)}
@author{@(author+email "Gary Baumgartner" "gfb@cs.toronto.edu")}
@; Appears at:
@;   • top of document just after Racket version number
@;   • top of side bar with expandable table-of-contents top-level sections
@;   • top of side bar indexing of all the sections and indexed identifiers
@;   • main Racket documentation table of contents
@;
@; To Do: The date [when it was provided] wasn't appearing: should it?

This page is a reference for the University of Toronto “Computational Thinking”
Teaching Language, used in the Fall 2018 offering of CSC@|~|104.
@instructor-inset{ Top-level sections are ordered somewhat chronologically.
        
 Included are top-level sections for each datatype.
 Each function that could appear naturally in a datatype's section, but is in another section,
 is still discoverable in the datatype's section: it's mentioned in the datatype's sub-section
 named “<Datatype> Functions Documented in Other Sections”.
 
 Where both out-of-section and in-section references are made, out-of-section references mention
 their section so the reader can choose to click vs. scroll, and stay oriented within the document.}

@; ★★★ Sectioning ★★★
@; Sections are up to three deep, and can't skip a [the second] level going in.
@; Some shorthands, ‘§’ ‘§§’ ‘§§§’ [\S for latex keybinding]:
@(define-values (§ §§ §§§) (values section subsection subsubsection))
@; All sections [as used here?] have the same form:
@;   (§ [#:tag tag #:tag-prefix tag-prefix #:style style] pre-content ...+)
@; Titles, indexed in the side-bar table of contents, get numbered dotted, up to two dots: “§.§§.§§§”.
@;  ★★ Referencing.
@; I'm using single strings as tags
@;{  @§[#:tag "an-identifying-string"]{The Title}  }
@;  which I'm referencing by
@;{  @secref[tag-string]  }
@;  which makes a blue link with the section's pre-content title, without numbering.

@(table-of-contents)

@; ★★★ Declare we're Documenting a Particular Module ★★★
@; For context help and integration into the main documentation, IIRC.
@declare-exporting[csc104/2018-fall]

@; ★★★ Documenting Functions, Variables, and Forms ★★★
@; Custom forms to use, which also kept [before recent hack] a record of everything documented.
@; We generated a section with all signatures, and compared those with language's actual exports.
@;  ★★ Exports recording everything documented [before recent hack].
@(provide the-documented-variables the-documented-functions the-documented-forms)
@; There are some ‘defidform/inline’s buried in there, inside some of ‘the-documented-forms’.
@; To Do: Investigate extracting tags, and undefined references to tags, e.g. via ‘raco’:
@;{raco setup: --- building documentation ---
 raco setup: 1 running: <pkgs>/csc104/scribblings/docs.scrbl
 raco setup: WARNING: undefined tag in <pkgs>/csc104/scribblings/docs.scrbl:
 raco setup:  ((lib "csc104/compthink-intermediate.rkt") #%app)}
@;   ★ Helpers to manage recording of the list of documented identifiers.
@(define-values (the-documented-variables the-documented-functions the-documented-forms)
   (values (list) (list) (list)))
@(define (documented-variable id contract-expr-datum)
   (set! the-documented-variables
         (append the-documented-variables ; Extra list wrap to distinguish from functions/forms:
                 `(((,id ,contract-expr-datum))))))
@(define (documented-function prototype result-contract-expr-datum)
   (set! the-documented-functions (append the-documented-functions
                                          `((,prototype → ,result-contract-expr-datum)))))
@(define (documented-form form-datum)
   (set! the-documented-forms (append the-documented-forms `((,form-datum)))))
@;  ★★ The custom forms to use.
@(define-syntax-rule (variable id contract-expr-datum pre-flow ...)
   #;(begin (documented-variable 'id 'contract-expr-datum)
            (defthing id contract-expr-datum pre-flow ...))
   (defthing id contract-expr-datum pre-flow ...))
@(define-syntax-rule (function+ prototype result-contract-expr-datum pre-flow ...)
   (defproc #:kind "function" prototype result-contract-expr-datum pre-flow ...))
@(define-syntax-rule (function prototype result-contract-expr-datum pre-flow ...)
   (begin (documented-function 'prototype 'result-contract-expr-datum)
          (defproc #:kind "function" prototype result-contract-expr-datum pre-flow ...)))
@(define-syntax-rule (function* ([prototype result-contract-expr-datum] ...) pre-flow ...)
   (begin (documented-function 'prototype 'result-contract-expr-datum)
          ...
          (defproc* #:kind "function" ([prototype result-contract-expr-datum] ...)
            pre-flow ...)))
@(define-syntax form
   (syntax-parser
     [(_ #:literals (literal ...) form-datum pre-flow ...)
      #'(begin (documented-form 'form-datum)
               (defform #:kind "special form" #:literals (literal ...) form-datum pre-flow ...))]
     [(_ form-datum pre-flow ...) #'(form #:literals () form-datum pre-flow ...)]))
@(define-syntax form/none
   (syntax-parser
     [(_ #:literals (literal ...) form-datum pre-flow ...)
      #'(begin (documented-form 'form-datum)
               (defform/none #:kind "special form" #:literals (literal ...) form-datum pre-flow ...))]
     [(_ form-datum pre-flow ...) #'(form/none #:literals () form-datum pre-flow ...)]))
@;   ★ Forms to “mention” a function by contract, when it's documented fully in another section.
@(define-syntax-rule (mention-function prototype result-contract-expr-datum section)
   @defproc[#:kind "function" #:link-target? #false prototype result-contract-expr-datum]{
 See @secref[section].})
@(define-syntax-rule (mention-function* ([prototype result-contract-expr-datum] ...) section)
   @defproc*[#:kind "function" #:link-target? #false ([prototype result-contract-expr-datum] ...)]{
 See @secref[section].})
@;   ★ To Do: make certain duplicate or undocumented references appear in the left column index.
@;{@(require (only-in scribble/core make-toc-target2-element))
 @make-toc-target2-element[#false "" '(list?-tag "list?-tag") @racket[list?]]}

@; ★★★ Including Files ★★★
@;  ★★ CSS and image: wrap them with these so ‘raco’ installs them correctly.
@(define (reference-file name) (collection-file-path name "csc104" "scribblings" "includes"))
@(define (include-image name #:scale [scale 1.0])
   (image #:scale scale (collection-file-path name "csc104" "scribblings" "includes" "images")))
@;  ★★ ‘require’: use non-string module paths.

@; ★★★ Specially Formatted Identifiers ★★★
@(require (only-in scribble/racket make-element-id-transformer))
@;  ★★ Subscripted.
@; Use ‘make-subscripts’ below to declare them for use. To Do: make them work in prototypes.
@(define-for-syntax (to-subscript stx)
   (define str (symbol->string (syntax->datum stx)))
   (define id (string->symbol (substring str 0 1)))
   (define sub (substring str 2))
   (datum->syntax stx `(elem (racket ,id) (subscript ,sub))))
@(define-syntax-rule (make-subscripts id ...)
   (begin (define-syntax id (make-element-id-transformer to-subscript)) ...))
@(make-subscripts a_1 a_n a_i b_1 b_n l_1 l_2 x_1 x_n X_1 X_n y_1 y_n Y_1 Y_n)
@;  ★★ Raw ‘=’ sign.
@; An ‘=’ sign without it being a reference to the identifier.
@(define-syntax === (make-element-id-transformer (λ (_) #'(racketplainfont "="))))

@; ★★★ Interaction Examples ★★★
@(require (only-in csc104/eval compthink-eval) (only-in #;scribble/eval "eval.rkt" interaction
                                                        [scribble-exn->string scribble-exn->string′]))
@; Use this form for Interactions examples, in order for them to reflect out language.
@(define-syntax-rule (interact input-expr ...) @interaction[#:eval (our-eval) input-expr ...])
@;   ★★ Image literals and adjustments.
@; To show image literals in the input expression, and fine-tune their borders, use an escape
@;  in the ‘show-datum’ of scribble/example's ‘(eval:alts show-datum eval-datum)’.
@; And images in the result images can be fined-tuned by fine-tuning the ‘eval-datum’ expression.
@; Anything used in an escape must be defined in the document's scope.
@; To use it unescaped in the result expression, it must be defined in the evaluator as well.
@(require (only-in "../language/without-gui.rkt"
                   circle square triangle star rectangle ellipse url->image
                   width height overlay rotate above beside cut-right))
@; Transformed outlined images can have part of their border outside their bounding box,
@;  which is fixed with this function [also defined in the evaluator].
@(define (fix-outline image)
   (overlay image (rectangle (+ 2 (width image)) (+ 2 (height image)) "solid" "transparent")))
@; Some literal images in the examples for use inline [also defined in the evaluator].
@(define-values (ellipse₁ ellipse₂ ellipse₃) (values (ellipse 30 15 "solid" "dimgray")
                                                     (ellipse 50 15 "solid" "gray")
                                                     (ellipse 70 15 "solid" "lightgray")))
@(define-values (lay-rectangle lay-ellipse) (values (rectangle 20 50 "solid" "teal")
                                                    (ellipse 50 20 "solid" "gold")))
@(define-values (star₁ star₂ star₃)
   (values (star 15 "solid" "brown") (star 20 "solid" "silver") (star 25 "solid" "gold")))
@(define a-circle (circle 10 "solid" "green"))
@(define our-eval
   (let [(our-eval
          (compthink-eval
           #| Image handling. |#
           ; Imported evaluator has a parameter for the margin around image results. |#
           (extra-margin 8) ; Following also defined outside evaluator: see note above. 
           (define (fix-outline image)
             (overlay image (rectangle (+ 2 (width image)) (+ 2 (height image))
                                       "solid" "transparent")))
           (define-values (ellipse₁ ellipse₂ ellipse-3) (values (ellipse 30 15 "solid" "dimgray")
                                                                (ellipse 50 15 "solid" "gray")
                                                                (ellipse 70 15 "solid" "lightgray")))
           (define-values (lay-rectangle lay-ellipse)
             (values (rectangle 20 50 "solid" "teal") (ellipse 50 20 "solid" "gold")))
           (define-values (star₁ star₂ star₃) (values (star 15 "solid" "brown")
                                                      (star 20 "solid" "silver")
                                                      (star 25 "solid" "gold")))
           (define a-circle (circle 10 "solid" "green"))))]
     (λ () our-eval)))
@; From "manual-style.css": .SCodeFlow img { margin-top: 0.5em; margin-bottom: 0.5em; }
@(require scribble/html-properties scribble/core)
@; A style that references the new style file.
@(define compthink-style
   (make-style "compthink" (list (make-css-style-addition (reference-file "image.css")))))
@; A dummy element that references the new style file so scribble will include it in the html.
@elem[#:style compthink-style]{}
@; Images as results are a bit cramped now, but it's fine for now: that can easily be fixed with a new
@;  form for those, or abandon this and manually make a few images and use the ‘image’ element.

@; ★★★ To Do: Document the rest more thoroughly as above ★★★
@;{(require scribble/example)}
@(require (only-in "../rewrite-error.rkt" exn→rewritten-message))
@(scribble-exn->string′
  (λ (e) (if (exn? e) (exn→rewritten-message e) (format "uncaught exception: ~s" e))))

@; Classify and determine formatting of all code and code fragment occurrences in documentation.

@;{Each identifier that might be referenced, from the module we're documenting or from others
 [ISL, ‘picturing-programs’, ‘2htdp/image’, or ‘racket’], should be ‘require’d ‘for-label’
 so that mentions of it in in a code documentation form automatically link to it. 
 The definition in the documentation also formats the identifier as a bold non-blue link to itself. 
 If an identifier is used in multiple forms, and documented with one of the “together” forms,
 then all links are to the first definition, which is formatted as usual, and the others link to it
 with a non-blue non-bold link.
 To separate the multiple forms, use a ‘/none’ version for all but the first.
 They will become italicized non-links.
 If they're tagged there with ‘#:literal (<id>)’, they become non-bold blue links to the first form,
 which is what the ISL documentation appears to do.
 Another option is manally formatting the non-link, e.g., ‘#,(racketplainfont "<id>")’.
 How do we find out all referenced identifiers that aren't labelled?}
@(require
   (for-label (except-in csc104/2018-fall
                         #%app #%module-begin #%top #%datum #%top-interaction)
              (only-in racket/draw color-database<%>)
              (only-in 2htdp/image underlay crop mode? pen)
              (only-in picturing-programs flip-other flip-main)
              (only-in lang/htdp-intermediate
                       sqrt string-ith explode implode
                       andmap ormap memf build-list argmin argmax
                       cons empty car cdr null fifth eighth cons? eq? eqv? member
                       error exit)))
@§{Acknowledgements}
Aliya Hameer implemented a significant portion of the 2017 Fall version of the language,
which this 2018 Fall version is based on.

@;@§{Installation}
@;{
 @§§{Preferences Automatically Set when First Installed}
 @(require (only-in string-constants string-constant))
 @itemlist-inset[@item{General / @(string-constant open-files-in-tabs) : ☑}
                 @item{General / @(string-constant interactions-beside-definitions) : ☑}
                 @item{General / @(string-constant test-coverage-summary) : ☐}
                 @item{Colors / Background / @(string-constant parenthesis-color-scheme) : Spring}
                 @item{Editing / Square Bracket / @(string-constant fixup-open-brackets) : ☑}
                 @item{Editing / Square Bracket / Cond-like keywords : Add big-bang (1)}
                 @item{Editing / General Editing / @(string-constant enable-automatic-parens) : ☑}
                 @item{Editing / Racket : ☑ all}
                 @item{Tools / Installed Tools : Skip Algol and Swindle}
                 @item{Font / Font Name : Source Code Pro (if installed)}
                 @item{Racket / Limit Memory... : 500 MB}]
}
@instructor-inset{
 I use “Source Code Pro” for a nice looking font that distinguishes similar glyphs well.

 The font size is adjustable with Command-+ and Command-- [on the Mac],
 which you can see in the “View” menu as well.
 
 For printing: on my Mac, 96 11-point characters fit the page width when printing to PDF
 [except some glyphs in my font, such as “★”, are slightly wider than the fixed-width].
 You can set a guideline to show up when you reach a certain column: a “View” menu item toggles this,
 but the column number is only adjustable via “DrRacket / Preferences / Editing / General Editing”.

 The color scheme for code is left as default [in CSC324 I make delimiters subtle,
 for reasons I won't go into here].}
@self{Assumes @racket[big-bang] isn't already in the user's Cond-like keywords list,
 and Algol and Swindle aren't already present in the user's Tools configuration.}

@;{
@§§{Toolbar Buttons}
@instructor{2018: One button toolbar does both of the following now, and PLT Stepper button removed.
 @inset{“Reindent All”: does the same as the DrRacket menu item “Racket / Reindent All”,
  which also has a keyboard shortcut you should encourage the students to use.
  It works on either the Definitions or Interactions: whichever has the focus.
  
  I always emphasize I'm doing a reindent when using the shortcut, but we want a more visible,
  but still convenient, indicator that we're frequently using it.
  Emphasize to the students that you are watching for any code movement, and taking a quick look
  afterwards at the code, to see if the actual structure matches our mental model of the structure.}
 @inset{“Finer Format”: goes through the Definitions and removes (non-comment) whitespace before any
  closing parentheses/brackets, and then does a second pass alerting the user to any lack of
  whitespace between a part and an opening parenthesis/bracket: tell students to examine whether it
  indicates a problem with their code's meaning, e.g.
  @;(length (reverse(L)) 
  @(local [(define-values (open-paren close-paren)
             (values (racketparenfont "(") (racketparenfont ")")))]
     (itemlist-inset
      (item (list open-paren
                  (racket length) " "
                  open-paren (racket reverse) open-paren (racketidfont "L") close-paren close-paren
                  close-paren
                  " : meaning"))
      (item (list open-paren
                  (racket length)
                  open-paren (racket reverse) " " (racketidfont "L") close-paren
                  close-paren
                  " : formatting"))))}}
}

@;{
@§§[#:tag "highlighting"]{Highlighting}
@instructor{The highlighting of code is also changed: rather than the line-based highlighting,
 it “boxes” the expressions. Out-of-the-box it has all the bells-and-whistles enabled, to
 maximally emphasize the structure of code. The highlighting is customizable with a new
 DrRacket menu item “View / Boxed Highlight”.}
@self{Shading out-of-the-box.}
}

@§{The Definitions and Interactions Areas}
@instructor-inset{
 The Toggle (“Hide” and “Show”) Definitions and Interactions menu items no longer have keyboard
 shortcuts: instead, those shortcuts have been rebound to do “show more”, as opposed to toggling.
 This matches what Danny and I still often unconsciously expect, even though we've experienced
 the toggling behaviour hundreds to thousands of times.
 
 The Definitions and Interactions are now laid out horizontally instead of the default vertical,
 since larger modern displays are wide rather than tall (and it tends to work better in lecture
 for other reasons, so students should see the same layout we use). This can be toggled with
 a “View” menu item, as well as with the shortcut key shown there.}

The Interactions Area is where you can enter an expression and have DrRacket immediately compute
its value.

When you press Return (which might be labelled “Enter” on your keyboard), DrRacket checks
whether you have written a complete expression (essentially: are all the parentheses “balanced”).
If not, DrRacket simply moves the cursor to a new line, so you can complete or edit the expression
you are creating. If you have written a complete expression, then pressing Return makes DrRacket
compute the result of the expression and show you its value.

If you are editing an expression and want to spread it over multiple lines while you work on it
(something you will see us do for more complex expressions to show their structure), then you
can hold down the Shift key while you press Return. Holding down the Shift key prevents DrRacket
from computing the expression, even if the expression looks complete.

@instructor-inset{
 This is new behaviour, more closely matching students' expectations.
 @link["https://groups.google.com/forum/#!topic/bootstrap-discuss/L0YQBpozK9s"]{Bootstrap}
 changed this as well.
 
 I spread expressions over multiple lines as early as the first day, when doing something like
 rotating of two images placed @racket[above] or @racket[beside] each other, and vice-versa,
 but students either miss the commentary (and just assume it word-wrapped), or if they do understand
 the commentary then I think they're likely to note the Return/Shift-Return distinction.

 [Return with no expression moves to a new line, since there's no complete expression:
 this has been thought about and kept intentionally].

 I've disabled being able to move up past the prompt with the keyboard, which was a major source
 of confusion/annoyance for students.
 You can still select with the mouse, or extend selection into it, then move around (selecting or not)
 by s-expression navigation, and do arbitrary selection movements.

 On the Mac, I've made Command-↑/↓ cycle through history.
 
 If the Definitions text is changed without running, then the Interactions is potentially out-of-sync
 with the definitions in the Definitions text. A yellow warning appears at the bottom of DrRacket
 in this case, but I still frequently miss that.}
@self{Important keybindings.
 @inset{
  We need to think about the majority of students (Windows), and their
  use of the labs, which probably means making the teach.cs keybindings more Windows-like if there's
  a difference.

  Keep @link["https://groups.google.com/forum/#!topic/bootstrap-discuss/L0YQBpozK9s"]{Bootstrap} in
  mind. In particular, they have a keybinding to bring up a window listing the keybindings.  
  @itemlist[
 @item{more/less definitions/interactions}
 @item{submit/newline in interactions}
 @item{previous/next in interactions history}
 @item{run: make the students aware of it for themselves, press the button in lecture for visibility}
 @item{break, kill: same comment as run --- stop button, with multiple pressessupposed to become kill}
 @item{save, save as}
 @item{tab or window creation, movement, close}
 @item{indent all: same comment as for run}
 @item{indent line or selection: useful at times to mimic watching indent-all working incrementally}
 @item{automatic starting bracket vs parenthesis, and closing delimiter}
 @item{force surround by delimiter: I don't know why this shouldn't always be the case}
 @item{collapse space: maybe bind to option-space, since similar to navigation?}
 @item{navingate by s-expression}
 @item{collapsing that we've added}
 @item{remove balanced delimiters}
 @item{comment manipulation}
 @item{find, find next}
 @item{context help: same comment as for run, but with context menu instead of button}
 @item{change font size, change layout}]}}
@instructor-inset{
 I've changed how option-left and option-right on the Mac navigate s-expressions
 (which we'll just refer to as left and right here).
 
 Danny, when you get back we'll take your habits, and modifier keys, into account,
 and also decide for Windows users.

 On the Mac, left and right move along the siblings s-expressions, (also when shifted,
 for selection) but no longer skip starts when going right, nor ends when going left.
 The loss is that the number of moves doesn't count the number of s-expressions, which
 I might not have realized gave me a “feel” for the size and number of components of
 an s-expression. The gain is that one can stop in front of an s-expression in order
 to go into it with option-down, rather than having to overshoot and go back. That's
 likely an improvement for students watching us navigate, so tips the balance.

 Also, rather than getting stuck at the beginning or ending delimiter, the left or
 right movement jumps just past it if past the first/last sibling. The loss is holding
 down one of the keys and waiting to get stuck in order to reach an end, and now if you
 overshoot you can't immediately invert. Leaving by going up to the start of the enclosing
 s-expression, close to the first component, where the main meaning is usually determined,
 is actually somewhat orienting even though it's momentarily in the wrong direction if one
 wants to continue to the right. I might reconsider and have option-up when at the end be
 the way to break out to the right (it's ambiguous in an empty s-expression, but then it's
 fine to just go up to the left), and option-down when right beside a closing delimiter
 go down into the end of it, rather than down into the next s-expression (here the ambiguity
 is an empty gap between two s-expressions, but again the default behaviour is fine).}
@self{
 @link["https://developer.apple.com/macos/human-interface-guidelines/input-and-output/keyboard/"]{
  Apple Human Interface Guidelines: Keyboard}.}
@§§{Inserting Comments}
There are four(!) ways to write something in your code so that what you write will be @italic{ignored}
by DrRacket when running the document. Something written for a human reader of your code,
not for DrRacket to interpret, is called a “comment”.

A semi-colon “@tt{;}” makes DrRacket ignore everything from it up to the end of the line it's on,
and the text will be given a different color than code.
You can put a semi-colon at the start of many lines at once, by selecting multiple lines and then
choosing the menu item “Racket / Commment Out with Semicolons” (turning something into a comment
is called “commenting it out”, and this kind of comment is called an “end-of-line comment”).
If you have many lines with semi-colons at the start, you can remove them all by selecting the lines
and then choosing the menu item “Racket / Uncomment”.

A comment box creates an area where text is still colored like code but otherwise ignored by DrRacket.
You can insert one with the menu item "Insert / Insert Comment Box", and then type inside it.
You can also comment out a part of your document by selecting it and then choosing the menu item
"Racket / Comment Out with a Box".

Another way to comment out a part of your document is to surround it with the delimiting pairs of
characters “@tt{#|}” and “@tt{|#}”. Like with semi-colons, the color will change. This kind of
comment is called a “block comment”.

Finally, to make DrRacket ignore a definition or expression, you can put the two characters “@tt{#;}”
in front of the expression. This can be a little hard to notice at first, since the color of the
definition or expression doesn't change.

@instructor-inset{
 PP describes them all except comment boxes in PP Section 3.4.3.

 An end-of-line comment with only whitespace in front gets indented along with other code.
 
 We'll probably have keybindings to toggle end-of-line comments, and to comment-out with a box.
 And we'll probably have comment-out with semi-colons insert a space after the semi-colons.

 We tend not to use block comments and expression comments, although I have used block comments in
 project code. They're both interesting to use “inline”, both in coding and teaching, and I use
 them a lot in CSC324. Expression comments show a strength of our syntax. At some point I'd like to
 get them shaded slightly to indicate that they're comments, in which case they might become
 clearer and more meaningful to play with in CSC104.}
@self{Block comments nest properly, which means the restrictive policies on their use in some
 other languages don't apply. Their drawback is that over multiple lines the indentation of their
 contents is underspecified, and moving their delimiters doesn't move their contents.

 The ‘@tt{|}’ is among the delimiters that are inserted/enclose balanced if balancing is turned on:
 if it's immediately after a ‘@tt{#}’ then the closing block-comment delimiter is used. At some
 point I'd like it changed to also put a space before the closing delimiters.}

@§{Seeing the Steps to Compute an Expression}
@instructor-inset{
 For now, the documentation in this section is just for you to get started playing with the new
 alternate stepper. I'm still mulling the “interface” to it (input and output), and will also
 make it more uniform wrt our language for its error detection and reporting.}

@;{To see the steps to compute an expression, you can use DrRacket's traditional Stepper, by pressing
 the “Step” button. To see a listing of all the steps, closer to our course's style, use one of
 the following @racket[step] or @racket[steps] forms.}
@form[(step expression/definition
            ...)]{
 Print the steps of computation, waiting for the user to press Enter/Return between steps. }
@form[(steps expression/definition
             ...)]{ Print the steps of computation without waiting. }
Limitations: these can't step into @racket[big-bang], @racket[require], and @racket[time] expressions.
@instructor-inset{
 Previoulsy we didn't emphasize short-circuiting of @racket[and] and @racket[or], nor test on it.
 Now, the ease of playing with steps at the level we test makes it reasonable to expect
 the correct model. As a concept: short-circuiting is useful, and can made more relevant as part of
 a larger discussion of meaning, denotation, and intention.

 2018: The stepper now sees all (and only) preceding definitions, including in Interations

 For the moment (because it happens accidentally), if the new stepper is used in the Interactions,
 it sees user definitions, and treats them the same as the language's definitions
 (i.e. can use them, but it black-boxes the function bodies). Otherwise, each use of @racket[steps]
 is in its own new environment. I'm still mulling the possibilities, and best default(s) for the
 scope of environments.

 We often ask for only subsets of the steps, e.g. allowing parallel stepping of argument expressions,
 or allowing reuse of a previously stepped expression's value without stepping it again. I'm mulling
 those as well.

 @;{I've made DrRacket's Stepper pretend that step expressions don't exist, and have it skip them.}

 The result value is void, which can't be used meaningfully, but it could be interesting eventually
 to allow a sub-expression to be a @racket[step] expression still producing its normal value,
 especially for debugging.}

@;@§{Using Functions}
@instructor{In the sections that document functions, the first sentence describing a function is
 declarative, and about the result value. This can be a bit awkward at times, and I'll revisit this
 decision later. Alternatives are doing one or more of:
 @itemlist-inset[@item{describing the function or expression as computing/producing the result}
                 @item{using the parameter names more}]}
@self-inset{
 Document: function call expression form, function literals.
 
 Decide our terminology.
 
 See the HtDP @link["https://docs.racket-lang.org/htdp/index.html#%28part._error-guidelines%29"]{
  Error Message Composition Guidelines} and the other sections there, including the papers.}

@§{Images}
@instructor-inset{
 The course used to rely on some functions from “Picturing Programs”,
 which were included via @racket[(require picturing-programs)].

 2018: Removed deprecation warning if requiring picturing-programs.}
@self-inset{
 Discussion of the relationship to 2htdp/image.
 
 Non-textual source files, @racket[url->image].}
@instructor-inset{
 @secref["geometric-shapes"] comes after many sections using its functions in their examples,
 to match the chronology of the course more closely, but are still used since they make for
 clearer stand-alone examples.}

@§§{Literal Images}
A “literal” image (as opposed to one computed by an image function) can be entered into DrRacket,
and used as-is as a value in code.

@§§§{Copying Images in DrRacket}
You can find functions for modifying images, as well as creating geometric shapes, in later sections.
But if you find an image that you would like to include, you can copy it into DrRacket from another
application, from a file on your computer, or from a web address.

@itemlist[
 @item{To copy an image from another application (for example a web browser), select/highlight it
  in the other application, using that application's copy mechanism (which should be a menu item
  and/or a keyboard shortcut):
  
  @(include-image #:scale 2/5 "copy-image.png")
  
  Then paste it into DrRacket (using DrRacket's copy mechanism, for example the DrRacket menu item
  “Edit / Paste”, or the associated keyboard shortcut):
  @(include-image #:scale 2/5 "paste-image.png")}

 @item{Sometimes, the copy-paste approach does not work for a particular application.
  If you can save the image from the application to a file on your computer, then you can insert it
  into DrRacket using the menu item “Insert / Insert Image...” and choosing the file:
  
  @(include-image #:scale 2/5 "insert-image.png")}

 @item{If the image is on the web, and you know its address (“URL”), you can use @racket[url->image]
  to download and refer to it. First copy the URL, for example from your browser address bar:
  
  @(include-image #:scale 2/5 "copy-url.png")

  Then use the following function (be sure to use the double-quotes, to make it a “text”):}]

@deftogether[
 (@defproc[#:kind "function" (url->image [url text?]) image?]
   @defthing[#:kind "alias" url→image url->image])]{
 The image at the web address @racket[url].
 @;{@interact[(url->image "https://racket-lang.org/logo.png")]}}

You can also save an image from DrRacket to a file, using the image context menu (right-click
on the image):

@inset{@(include-image #:scale 2/5 "save-image.png")}

@§§{Combining Images}

@instructor{Students forget:
 @itemlist-inset[@item{default alignment is centering}
                 @item{the ordering of @racket[above] and @racket[overlay]}
                 @item{images are combined by bounding box, not necessarily the visible parts}]}
@instructor{Selecting/highlighting an image can emphasize the bounding box aspect:
 @inset{@(include-image #:scale 2/5 "bounding-box.png")}}
@instructor{I extended all these functions to allow one or no arguments, for convenience.
 Preventing it doesn't seem obviously useful:
 students should be using image functions a lot, early on,
 and learning to form compound expressions carefully
 rather than reacting to error messages appearing.}
@self{Detail convenience, reference discussions elsewhere in these docs,
 simple benefit of uniformity with them.}

@function[(beside [image image?] ...) image?]{
 The single image formed by arranging all the images in a horizontal row,
 vertically centered, starting with the first image at the left.
 @interact[(eval:alts (beside #,star₁ #,star₂ #,star₃)
                      (beside   star₁   star₂   star₃))]}
@function[(above [image image?] ...) image?]{
 The single image formed by arranging all the images in a vertical column,
 horizontally centered, starting with the first image at the top,
 above the second image, above the third image, etc.
 @interact[(eval:alts (above #,ellipse₁ #,ellipse₂ #,ellipse₃)
                      (above   ellipse₁   ellipse₂   ellipse-3))]}

@function[(overlay [image image?] ...) image?]{
 The single image formed by arranging all the images in the same location,
 centered, with the first image on the second image, on the third image, etc.
 @interact[
 (eval:alts (overlay #,lay-rectangle #,lay-ellipse)
            (overlay   lay-rectangle   lay-ellipse))
 (eval:alts (overlay #,(ellipse 50 10 "solid" "red") #,(ellipse 40 20 "solid" "black") #,(ellipse 30 30 "solid" "red") #,(ellipse 20 40 "solid" "black") #,(ellipse 10 50 "solid" "red"))
            (overlay   (ellipse 50 10 "solid" "red")   (ellipse 40 20 "solid" "black")   (ellipse 30 30 "solid" "red")   (ellipse 20 40 "solid" "black")   (ellipse 10 50 "solid" "red")))
 #;(eval:alts (overlay #,(circle 5 "solid" "red") #,(circle 10 "solid" "black") #,(circle 15 "solid" "red") #,(circle 20 "solid" "black") #,(circle 25 "solid" "red"))
              (overlay   (circle 5 "solid" "red")   (circle 10 "solid" "black")   (circle 15 "solid" "red")   (circle 20 "solid" "black")   (circle 25 "solid" "red")))]
 @instructor{There is a @racketmodname[2htdp/image] function @racket[underlay],
  but it can be simulated by re-arranging the argument expressions, or @racket[reverse]ing the list
  when used with @racket[apply]. This approach works for @racket[beside] and @racket[above] as well,
  which don't even have reversed versions in @racketmodname[2htdp/image].}}
 
@function*[([(align-beside [alignment (or/c "top" "bottom" "center" "baseline")]
                           [image image?] ...) image?]
            [(align-above [alignment (or/c "left" "right" "center")]
                          [image image?] ...) image?])]{
 The image formed by placing the images @racket[beside], or @racket[above], each other,
 but aligned according to @racket[alignment] instead of automatically centered.
 @interact[
 (eval:alts (align-beside "top" #,(ellipse 15 60 "solid" "yellowgreen") #,(ellipse 15 40 "solid" "olivedrab") #,(ellipse 15 20 "solid" "darkolivegreen"))
            (align-beside "top"   (ellipse 15 60 "solid" "yellowgreen")   (ellipse 15 40 "solid" "olivedrab")   (ellipse 15 20 "solid" "darkolivegreen")))
 (eval:alts (align-above "left" #,(fix-outline (ellipse 60 15 "outline" "mediumslateblue")) #,(fix-outline (ellipse 40 15"outline" "slateblue")) #,(fix-outline (ellipse 20 16 "outline" "darkslateblue")))
            (align-above "left"   (ellipse 60 15 "outline" "mediumslateblue")   (ellipse 40 15"outline" "slateblue")   (ellipse 20 16 "outline" "darkslateblue")))]

 The @racket[alignment] @racket["baseline"] is meant for images produced by
 the @racket[text->image] function (from section @secref["type-conversion"]).
 Using it lines up the character “baselines”, ignoring descending parts of letters like @racket["p"]:
 @interact[(align-beside "baseline"
                         (text->image "play" 30 "tomato")
                         (text->image "a" 40 "blue")
                         (text->image "jig" 20 "darkgreen"))]
 @function[(align-overlay [horizontal-alignment (or/c "left" "right" "center")]
                          [vertical-alignment (or/c "top" "bottom" "center" "baseline")]
                          [image image?] ...) image?]{
  The image formed by arranging all the images in the same location,
  but aligned according to @racket[horizontal-alignment] and @racket[vertical-alignment]
  instead of automatically centered, with the first image on the second image, on the third image,
  etc.}
 @instructor{We usually use these functions much later in the course: occasionally for a nice
  recursive image, and for presenting a database as a table. But I sometimes use one earlier
  to make the triangular tower of images for the visual summation of @racket[(+ 1 2 3 ... n)].}
 @instructor{If used with our now-restricted @racket[apply], the extra argument can be added to
  the list of arguments with @racket[adjoin].}
 @instructor{I removed @racket[alignment] @racket["middle"], a synonym for @racket["center"].}
 @self{Reference the image guide, various online courses that ask for images: we should encourage
  students to publish their creations somehow.}}

@§§{Rotating, Scaling, Mirroring, Flipping, and Cutting}
@;{[(rotate-180 [image image?]) image?]}
@function*[([(rotate    [image image?] [angle number?]) image?]
            [(counter   [image image?])                 image?]
            [(clockwise [image image?])                 image?])]{
 The image, rotated @racket[angle] degrees counter-clockwise, 90 degrees counter-clockwise,
 @;{180 degrees,} or 90 degrees clockwise, around its center.
 @interact[(eval:alts (rotate #,(ellipse 20 40 "solid" "maroon") 45)
                      (rotate   (ellipse 20 40 "solid" "maroon") 45))
           (eval:alts (rotate #,(fix-outline (square 30 "outline" "black")) -10)
                      (rotate                (square 30 "outline" "black")  -10))]
 @interact[(define a-half-triangle (cut-right (triangle 40 "solid" "forestgreen")
                                              20))
           a-half-triangle
           (counter a-half-triangle)
           @;{(rotate-180 a-half-triangle)}
           (clockwise a-half-triangle)]
 @instructor{Students forget:
  @itemlist-inset[@item{direction of @racket[rotate]}
                  @;{@item{abbreviation meanings in “@racket[rotate-ccw]” and “@racket[rotate-cw]”}}
                  @item{meaning of “clockwise” and “counter-clockwise”}]}}

@function*[([(scale [image image?] [factor (and/c number? non-negative?)]) image?]
            [(shrink [image image?]) image?]
            [(enlarge [image image?]) image?])]{
 The image, scaled in both directions by the amount @racket[factor], 1/2, or 2.
 @interact[(eval:alts (scale #,(ellipse 20 15 "solid" "gold") 3)
                      (scale   (ellipse 20 15 "solid" "gold") 3))
           (eval:alts (shrink #,(ellipse 20 15 "solid" "gold"))
                      (shrink   (ellipse 20 15 "solid" "gold")))
           (eval:alts (enlarge #,(ellipse 20 15 "solid" "gold"))
                      (enlarge   (ellipse 20 15 "solid" "gold")))]}
@instructor{Students forget: that it scales in both dimensions, rather than the area as a whole.}

@function*[([(mirror [image image?]) image?]
            [(flip   [image image?]) image?])]{
 The image, mirrored left-to-right, or flipped upside-down.
 @interact[
 (eval:alts
  (beside (rotate #,(square 30 "solid" "tan") -30) (mirror (rotate #,(square 30 "solid" "brown") -30)))
  (beside (rotate   (square 30 "solid" "tan") -30) (mirror (rotate   (square 30 "solid" "brown") -30))))
 (eval:alts (above #,(star 25 "solid" "firebrick") (flip #,(star 25 "solid" "gray")))
            (above   (star 25 "solid" "firebrick") (flip   (star 25 "solid" "gray"))))]
 The image cannot contain an image created with
 the @racket[text->image] function (from section @secref["type-conversion"]):
 using @racket[mirror] or @racket[flip] with such an image reports an error.

 @;{@instructor{Students forget: these flip @italic{in} the indicated direction versus
   @italic{around that axis}.
   The description I wrote for the students above suggests better naming,
   but I don't think it's worth diverging here from Picturing Programs and HtDP2e
   since flipping is not a main cs concept.}}

 @instructor{There are also @racket[flip-main] and @racket[flip-other] in
  @racketmodname[picturing-programs], which I removed.}}

@function*[([(cut-bottom [image image?] [height (and/c number? non-negative?)]) image?]
            [(cut-top    [image image?] [height (and/c number? non-negative?)]) image?]
            [(cut-right  [image image?] [width (and/c number? non-negative?)]) image?]
            [(cut-left   [image image?] [width (and/c number? non-negative?)]) image?])]{
 A version of the image, with the specified height or width cut off from the
 top, bottom, left, or right of the image.
 
 If the number of pixels is too many for the size of the image, an error is reported.
 @interact[(eval:alts (cut-bottom #,(circle 15 "solid" "orange") 9)
                      (cut-bottom   (circle 15 "solid" "orange") 9))
           (eval:alts
            (above #,(fix-outline (circle 8 "outline" "blue")) (cut-top #,(fix-outline (star 24 "outline" "blue")) 13))
            (above                (circle 8 "outline" "blue")  (cut-top                (star 24 "outline" "blue") 13)))
           (eval:alts
            (beside (cut-right #,(triangle 40 "solid" "forestgreen") 20) (cut-left #,(triangle 20 "solid" "darkgreen") 10))
            (beside (cut-right   (triangle 40 "solid" "forestgreen") 20) (cut-left   (triangle 20 "solid" "darkgreen") 10)))]
 @;{@instructor{Students forget: it crops away, as opposed to keeps, the specified amount of pixels.}}
 @instructor{These come from Picturing Programs, whose versions expect a natural number of pixels.
  These now round instead, to make it easier to use very early in the course (e.g. when cropping away
  the @italic{computed} half an image).}
 @instructor{There's an all-direction @racket[crop] in @racketmodname[2htdp/image],
  which is more complicated in the simple situations we do cropping, so we didn't include it.}
 @self{Look at @racket[crop] from ISL: does it propagate non-integer sizes, in which case should we?}}

@§§{Image Properties}
@function*[([(width [image image?]) natural?]
            [(height [image image?]) natural?])]{
 Width, or height, of an image.
 @interact[(width (rectangle 5 30 "solid" "purple"))
           (height (rectangle 5 30 "solid" "purple"))
           (width (circle 200 "solid" "orange"))
           (height (above (circle 200 "solid" "orange")
                          (rectangle 5 30 "solid" "purple")))]

 @instructor{This is the “rendered” width and height.
  Images contain information about how they were constructed, so, e.g.,
  @racket[(scale 10 (square 1/10 "solid" "black"))] produces a 1-pixel square,
  even though the size of the unscaled square would render as 0 × 0
  (the “tracked” size is rounded for rendering).}

 @self{How the “structured” aspect of images impacts the course, especially in animations:
  accuracy versus performance, @racket[freeze],
  size and positioning during rotation.}}

@§§[#:tag "geometric-shapes"]{Geometric Shapes}

@instructor{We tell students that on tests they can draw with whatever color they want,
 and that only relative sizing matters.}

@instructor{I removed symbols as modes and colors (the course doesn't use the symbol type),
 transparency numbers as modes, “pens” as colors for outlines, and color objects as colors
 (see @racket[mode?], @racket[pen], and @racket[color] from @racketmodname[2htdp/image]).}

@instructor{Outline mode does something subtle with sizes, and can result in off-by-ones
 and missing edges: we try to avoid situations where that matters, and if we accidentally
 test in such a situation we pretend there is no subtlety.}

@self{Reference to the racket docs on nitty-gritty of this.}

@function[(circle [radius (and/c number? non-negative?)]
                  [shading (or/c "solid" "outline")]
                  [color color?]) image?]{
 A circle with the given radius, shading, and color.
 @interact[(circle 15 "outline" "green") (circle 10 "solid" "blue")]
 @instructor{Students forget: it takes a radius, not a diameter/width/height.}}
@function*[([(square [side-length (and/c number? non-negative?)]
                     [shading (or/c "solid" "outline")]
                     [color color?]) image?]
            [(triangle [side-length (and/c number? non-negative?)]
                       [shading (or/c "solid" "outline")]
                       [color color?]) image?])]{
 A square, or upward-pointing triangle, with the given side length, shading, and color.
 @interact[(square 20 "solid" "purple")
           (square 30 "outline" "brown")
           (triangle 30 "solid" "darkgreen")]
 @instructor{We don't ask students about heights (nor probably widths) of these triangles.}}
@function*[([(rectangle [width (and/c number? non-negative?)]
                        [height (and/c number? non-negative?)]
                        [shading (or/c "solid" "outline")]
                        [color color?]) image?]
            [(ellipse [width (and/c number? non-negative?)]
                      [height (and/c number? non-negative?)]
                      [shading (or/c "solid" "outline")]
                      [color color?]) image?])]{
 A rectangle, or an ellipse, with the given width, height, shading, and color.
 @interact[(rectangle 40 20 "outline" "black") (rectangle 20 40 "solid" "aquamarine")
           (ellipse 40 20 "outline" "maroon") (ellipse 20 40 "solid" "red")]
 @instructor{Students forget: @itemlist-inset[@item{meaning of “ellipse”, and even “rectangle”}
                                              @item{width-height order}]}}

@function[(star [distance (and/c number? non-negative?)]
                [shading (or/c "solid" "outline")]
                [color color?]) image?]{
 A five-pointed star, with the points @racket[distance] apart, with the given shading and color.
 @interact[(star 20 "solid" "yellow")]
 @instructor{We don't ask students about the size of stars.}}

@§§§{Color}
@deftogether[(@function+[(color? [v any/c]) boolean?] @defthing[#:kind "alias" colour? color?])]{
 Is the argument a text containing the name of a color,
 or a list of three or four percentages (numbers between 0 and 100 inclusive).
 @interact[(color? (list 12 56 34)) (color? (list 0 100 0 100)) (color? "orange") (color? "ORANGE")
           (color? "Black") (color? "bloo") (color? "transparent")]
 The list of recognized color names is shown in @racket[color-database<%>],
 except @racket["transparent"] is also considered a color.
 Color naming is case-insensitive.
 @;{(see @racket[text-ci=?] in @secref["approximate-equality"]).}
 A list of three or four percentages represents the color with that percentage of red, green, blue,
 respectively, and optionally how opaque to draw it (which defaults to 100 if unspecified).
 @instructor{As noted in @secref["geometric-shapes"]: I removed symbols, pens, and color structs
  as colors. Also, colors are now enforced to be in the database: ISL considers any text a color,
  defaulting black which was confusing.}}
@§§§{Pixels}
@function[(render [image image?]) image?]{
 The image, but without the hidden information about how it was constructed.
 Using this can improve the drawing speed, but loses information that which might end up making
 the image look worse if it is rotated or scaled up (if the original was scaled down).}
@deftogether[(@function+[(image->colors [image image?]) colors?]
               @defthing[#:kind "alias" image->colours image->colors]
               @defthing[#:kind "alias" image→colors image->colors]
               @defthing[#:kind "alias" image→colours image->colors])]{
 A list of the colors of the pixels in the image, row by row,
 with each color represented by a list of four percentages
 (see @racket[color?] for the meaning of the percentages).}
@deftogether[(@function+[(colors->image [colors colors?] [width natural?] [height natural?]) image?]
               @defthing[#:kind "alias" colours->image colors->image]
               @defthing[#:kind "alias" colors→image colors->image]
               @defthing[#:kind "alias" colours→image colors->image])]{
 An image built from the list of colors, each color represented by a list of four percentages,
 (see @racket[color?] for the meaning of the percentages)
 arranged row by row inside a rectangle of the given width and height.}
@deftogether[(@function+[(colors? [v any/c]) boolean?] @defthing[#:kind "alias" colours? colors?])]{
 Is the argument a list of lists of three or four numbers, where each number is a percentage
 (a number between 0 and 100 inclusive).}
@§§{Image Functions Documented in Other Sections}
@mention-function[(text->image [text text?] [font-size (and/c number? (between/c 1 255))]
                               [color color?]) image? "type-conversion"]
@mention-function[(image? [v any/c]) boolean? "data-types"]
@;{@mention-function[(image=? [image₁ image?] [image₂ image?]) boolean? "equality"]}

@§[#:tag "functions"]{Functions}
@§§[#:tag "function-types"]{Types of Functions}
@function*[([(unary? [f function?]) boolean?]
            [(binary? [f function?]) boolean?]
            @;{[(ternary-function? [f function?]) boolean?]
            [(quaternary-function? [f function?]) boolean?]}
              )]{ Can the function be called with one, or two, arguments, respectively?
 @interact[(unary? flip) (unary? rotate) (unary? +) (binary? flip) (binary? rotate) (binary? +)]}
@§§{Functions Operating on Functions, Documented in Another Section}
@mention-function[(function? [v any/c]) boolean? "data-types"]
@mention-function*[([(map [f unary?] [l list?]) list?]
                    [(apply [f function?] [l list?]) any/c]
                    [(repeats [f unary?] [seed any/c] [n natural?]) list?]
                    [(sift [p? unary?] [l list?]) list?]
                    [(map [f binary?] [l₁ list?] [l₂ list?]) list?])
                   "lists"]

@§{Numbers}
@instructor{Unapproximated ISL/racket numbers are represented by a rational (“exact”) datatype,
 versus approximated numbers (“inexact”, which internally are floating-point).
 @inset{
  In particular, decimal literals (e.g. @racketvalfont{1234567.89}) produce exact numbers,
  since they denote a rational (in particular, as an integer over a power of 10).
  An integer literal followed immediately by a @racketvalfont{/}, followed immediately by a natural
  number literal, is a numeric literal and produces an exact number.
  Exact non-integers print in DrRacket with a trailing “@racketresultfont{...}” or a
  “repeated decimal” overbar if necessary, and right-clicking on them allows you to switch the
  representation to a proper or improper fraction. I chose the decimal representation as the default,
  since students are very familiar with it, and any non-trivial math we do in this course is about
  the size of powers, where fraction representation can become unwieldy and unintuitive/intimidating
  for our audience.
  
  Inexact numberic literals have the prefix @racketvalfont{#i}, indicating that they were produced by
  (the computer or us) doing an approximation (see, e.g., the value shown for @racket[pi]).
  The sources of inexact numbers are: @racket[square-root], @racket[power], and @racket[pi].}}

@§§[#:tag "arithmetic"]{Arithmetic}
@instructor{Arities of @racket[+], @racket[*], @racket[-], and @racket[/]:
 @inset{Adding/multiplying one or no numbers becomes useful:
  @itemlist-inset[@item{with @racket[apply]}
                  @item{when commenting out parts of a sum/product during exploration}
                  @item{for uniformity during base representation discussion}
                  @item{for uniformity when discussing powers}]
  We don't test on the zero-arity conventions (and I probably avoid testing unary multiplication).

  ISL/racket require at least one argument for @racket[-] and @racket[/], but allow more than two.
  I removed the latter as being too error-prone, not very common, and easily expressed more clearly
  with addition and binary subtraction, or multiplication and binary division.
  For similar reasons, @italic{unary} division (reciprocal) was removed
  (and achievable via rational literal or binary division with numerator @racket[1]).}}
@instructor{Binary subtraction, because of its asymmetry with the unary form (and general asymmetry)
 is error-prone in prefix form.}
@instructor{The distinction between @racketvalfont{+7} and @racket[(+ 7)],
 @racketvalfont{-7} and @racket[(- 7)], and @racketvalfont{2/3} and @racket[(/ 2 3)],
 can be subtle for students.}
@instructor{Imaginary numbers, and fractional powers:
 @inset{
  There is a built-in datatype for Complex numbers, which we avoid.

  The ISL/racket version of @racket[square-root] allows the square root of negative numbers,
  which produces imaginary numbers: I've restricted the domain to avoid that.
  
  I've also restricted the domain of @racket[power] to avoid imaginary numbers,
  by disallowing a negative base (which is simple to describe, even though it excludes
  even powers of negative numbers). More discussion of this is where @racket[power] is documented.}}

@function[(+ [x number?] ...) number?]{ Sum of the numbers (the numbers added together).
 @interact[(+ 100 -1 5) (+ 1/3 1/3) (+ 7)]
 There is a convention in CS (and Math) that the sum of no numbers is @racket[0]: @interact[(+)]}
@function[(* [x number?] ...) number?]{ Product of the numbers (the numbers multiplied together).
 @interact[(* 5 -2 3) (* 3 4/3) (* 7)]
 There is a convention in CS (and Math) that the product of no numbers is @racket[1]: @interact[(*)]}
@function*[([(- [x number?]) number?] [(- [y number?] [x number?]) number?])]{
 With one argument: the negative of the number. @interact[(- 7)]
 With two arguments: the difference of the two numbers (the first number minus the second number;
 also phrased as: the second number subtracted from the first number). @interact[(- 7 2)]}
@function[(/ [x number?] [y number?]) number?]{
 The first number divided by the second number (the first number over the second number).
 @interact[(/ 12 2)] See also @racket[quotient] in @secref["integer-division"].}

@function*[([(inc [x number?]) number?] [(dec [x number?]) number?])]{
 One more, or one less, than the number. @interact[(inc 3) (dec 3)]
 @italic{Inc}rementing (adding @racket[1]) and @italic{dec}rementing (subtracting @racket[1])
 are somewhat common in programming. Using dedicated names for them is intended to make it easier
 for someone reading code to see at a glance which of the two operations is intended.
 They're also simple unary functions for us to use in examples, so we've abbreviated their
 names to make typing out examples about easy as if using @racket[(+ x 1)] or @racket[(- x 1)].
 
 If @racket[inc] and @racket[dec] were not already in the language, they could be defined this way:
 @(racketblock (define (inc x)
                 (+ x 1)))
 @(racketblock (define (dec x)
                 (- x 1)))
 @instructor-inset{Common uses:
  @itemlist-inset[@item{play with @racket[map], can be contrasted with @racket[apply]ing them}
                  @item{play with @racket[repeated]}
                  @item{in @racket[on-tick], and to decrement in recursions}]
  Although easily definable using core concepts, I left them in because they're used early and often,
  as part of many examples, and using their definitions doesn't re-inforce any important concept.

  The name “@racket[dec]” violates the spirit of “no abbreviations”,
  but I've left it as-is to make it less obtrusive in recursions.
  Of course the name also violates hyphenation-naming, but “@racketidfont{add-1}” is confusing.}}

@function[(squared [x number?]) number?]{Square of the number. @interact[(squared 5)]
 If @racket[squared] was not already in the language, it could be defined this way:
 @(racketblock (define (squared x)
                 (* x x)))
 @;{The name is abbreviated to avoid clashing with the @racket[square] function that
  produces an image of a square.}
 @instructor{@;{It's not commonly written out as a name in math and cs anyway,
   and we don't want an exception to all the shape names being “normal”.}
  Common uses:
  @itemlist-inset[@item{play: same as the @racketfont{(number → number)} functions
                     @racket[inc] and @racket[dec]
                     (and we don't implement it, for the same reasons)}
                  @item{fast exponentiation in public key cryptography}
                  @item{possibly distance/size/radius calculations}]}}
@function[(square-root [x (and/c number? non-negative?)]) number?]{ Square root of the number.
 @interact[(square-root 9) (square-root 2)]
 @instructor-inset{
  Renamed from @racket[sqrt] in ISL/racket, to follow the spirit of English naming,
  but more importantly: students often don't notice the “@racketidfont{t}”.
  The original name isn't a common abbreviation in math and cs, and we use the function infrequently.
  
  We don't use this much: I use it in “radius” calculations occasionally when constructing
  a background to put an image onto, and sometimes to mention number representation issues.}}
@function[(power [base (and/c number? positive?)] [exponent number?]) number?]{
 The first number (the “@racket[base]”)
 raised to the power of the second number (the “@racket[exponent]”).
 @interact[(power 3 2) (power 10 3)]
 @instructor-inset{
  Renamed from the non-English, non-standard, technical name “@tt{expt}” in ISL/racket.
                                                              
  As mentioned above, the domain is now restricted to avoid imaginary numbers.
  It also disallows powers of zero, since we don't use those, avoiding the zero to the power of zero
  technicality.
  
  Common uses:
  @itemlist-inset[@item{exponential/geometric growth (fibonacci, compound interest, chess-board
                     doubling story, global real-world processes such as pollution and population):
                     some of these can involve non-integer bases (e.g. compound interest) and/or
                     exponents (e.g. @racket[1/2] a year), and changes over time could plausibly use
                     negative exponents (e.g. @racket[-1] for a year ago.}
                  @item{base representation: powers of @racket[2] and @racket[10]
                     with natural exponent}
                  @item{natural number bases and exponents in public-key cryptography}]}}
@deftogether[(@variable[pi number?] @defthing[#:kind "alias" π pi])]{
 The ratio of a circle's circumference to its diameter. @interact[pi]
 @instructor{Common uses:
  @itemlist-inset[@item{awareness of inexact numbers for approximations, even just in case they occur}
                  @item{early example of pre-defined name of a non-function}]}}

@§§§[#:tag "integer-division"]{Whole-Number Division}
Whole-number division, also called “integer division”, is a version of division (@racket[/])
@margin-note*{If you are not familiar with the terms “integer” and “natural number”,
 see @secref["numeric-types"].}
that produces a whole number “@racket[quotient]” along with any leftover “@racket[remainder]”.

There is no general agreement on the meaning of whole-number division for negative numbers, so
for simplicity this course provides and uses it only for natural numbers.
@function*[([(quotient [m natural?] [n (and/c natural? positive?)]) natural?]
            [(remainder [m natural?] [n (and/c natural? positive?)]) natural?])]{
 The first number divided by the second number, rounded down to the nearest natural number,
 and the remainder.
 @interact[(code:line (quotient  17 5) (code:comment "Since 5 “goes into” 17 three times,"))
           (code:line (remainder 17 5) (code:comment "with two “left over”."))]
 If @racket[quotient] was not already in the language, it could be defined this way:
 @(racketblock (define (quotient m n)
                 (floor (/ m n))))
 @instructor{Common uses:
  @itemlist-inset[@item{awareness of existence, and variations, in PLs}
                  @item{base conversion/representation}
                  @item{public-key cryptography}
                  @item{implicit @racket[floor]ing for picky image functions such as
                     @racketidfont{cut-*} (but we're removing that pickiness)}
                  @item{faster Euclidean Algorithm}]
  
  See also @racket[even?] and @racket[odd?] in @secref["numeric-types"].}}
 
@§§{Rounding}
@function*[
 ([(round [x number?]) integer?] [(floor [x number?]) integer?] [(ceiling [x number?]) integer?])]{
 The integer that is closest to the number (@racket[round]),
 closest and less than or equal to the number (@racket[floor], also known as “round down”),
 or closest and greater than or equal to the number (@racket[ceiling], also known as “round up”).
 @interact[(round 16) (round 16.3) (round 16.8) (round 17)]
 @interact[(floor 16) (floor 16.3) (floor 16.8) (floor 17)]
 @interact[(ceiling 16) (ceiling 16.3) (ceiling 16.8) (ceiling 17)]
 @instructor-inset{
  I think I use these only for picky image functions, but we're removing the pickiness.
  They are however common cs/math concepts.
              
  Rounding uses the convention that ties (@racketresultfont{_.5}) round to even,
  which we might want to change (or else mention and not test on).
   
  Behaviour for negative numbers is probably not intuited correctly by many students,
  but we don't use that nor test on it. If we discuss these as cs concepts, then it's probably worth
  having negative numbers in the domain.}}

@§§[#:tag "numeric-comparison"]{Comparing Numbers}
@instructor-inset{
 “@racket[=]” is a very traditional name for numeric equality, so we stick with it rather than
 something like “@racketidfont{number=?}” (see @secref["comparison"]). But going the other way:
 it could be extended to all types.
 
 Inequalities accept more than two numbers, which is occasionally handy for three numbers,
 or with @racket[apply] to determine whether a list is sorted.
 Traditionally they (along with @racket[=]) don't work with zero or one argument,
 which was a teachable moment in lecture (I used @racket[or] to guard such a use of @racket[apply].
 
 Inequalities are the main difficulty with prefix notation,
 and it showed up in student errors with the Sneeze project.
 The predicates in @secref["numeric-types"] mitigate this somewhat (not completely for Sneeze IIRC).
 Perhaps we should teach and test a rote translation to, e.g., the phrase “Is _ less than _?”.
 @(racketblock (= x y z ...) === (and (= x y) (= y z) ...)
               (< x y z ...) === (and (< x y) (< y z) ...)
               etc.)}
@self{It's the main place where I would consider allowing an infix notation.}
See also @racket[non-negative?], @racket[positive?], @racket[negative?], and @racket[zero?]
in @secref["numeric-types"].
@function*[([(= [x number?] [y number?]) boolean?]
            [(= [x number?] [y number?] [z number?] ...) boolean?])]{
 With two arguments: are the two numbers the same? @interact[(= 2 3) (= 3 3)]
 With more than two arguments: are all the numbers the same? @interact[(= 3 3 2 3) (= 3 3 3 3 3)]
 See also @secref["comparison"] for a discussion of @racket[same?],
 which works both on numbers and other types of data.}
@function*[([(< [x number?] [y number?]) boolean?]
            [(< [x number?] [y number?] [z number?] ...) boolean?])]{                                                                    
 With two arguments: is the first number less (smaller) than the second number?
 @interact[(< 2 3) (< 3 3) (< 3 2)]
 With more than two arguments: are the numbers in increasing order
 (is each number less than the next number)?
 @interact[(< 2 3 5 7) (< 2 3 3 5 7) (< 2 5 3 7)]}
@function*[([(> [x number?] [y number?]) boolean?]
            [(> [x number?] [y number?] [z number?] ...) boolean?])]{
 With two arguments: is the first number greater (larger) than the second number?
 @interact[(> 2 3) (> 3 3) (> 3 2)]
 With more than two arguments: are the numbers in decreasing order
 (is each number greater than the next number)?
 @interact[(> 7 5 3 2) (> 7 5 3 3 2) (> 7 3 5 2)]}
@deftogether[(@function+[(<= [x number?] [y number?] [z number?] ...) boolean?]
               @defthing[#:kind "alias" ≤ <=]
               @function+[(>= [x number?] [y number?] [z number?] ...) boolean?]
               @defthing[#:kind "alias" ≥ >=])]{
 Is each number less than or equal, or greater than or equal, to the next number?
 @interact[(<= 2 3) (<= 3 3) (<= 3 2) (<= 2 3 3 3 5 7 7) (<= 2 5 3 7)]
 @interact[(>= 2 3) (>= 3 3) (>= 3 2) (>= 7 7 5 3 3 3 2) (>= 7 3 5 2)]}

@function*[([(maximum [x number?] [y number?] ...) number?]
            [(minimum [x number?] [y number?] ...) number?])]{
 Maximum (largest, greatest), or minimum (smallest, least), of the given numbers.
 @interact[(maximum 3 2 7 9 8) (minimum 3 2 7 9 8)]
 @instructor-inset{
  Renamed from ISL/racket @racketidfont{max} and @racketidfont{min}.
  Although the abbreviations are common in cs/math, we don't use them enough to warrant abbreviation.
  
  Common uses: @itemlist-inset[@item{exploring the Collatz sequence}
                               @item{sizing of layout}
                               @item{play with @racket[apply]}]
  
  They require at least one argument.
  I can imagine (or have used) them in a function taking a list of things (text, images) and
  calculating layout, which could naturally encounter a singleton list (especially for the
  databasing example). With zero arguments (which could also come up for the same reason),
  one would probably want to specify the context-dependent “default” by throwing in a specific
  extra value, or writing a guard clause.}}

@§§[#:tag "numeric-types"]{Types of Numbers}
@instructor-inset{
 Added the racket predicate @racket[natural?], since it comes up frequently, and is an important
 cs/math concept.

 Added a predicate @racket[non-negative?], since it comes up frequently in contracts.

 The implementation of @racket[natural?], and some other implementations and contracts,
 are good examples relying on the specialness of conditionals.
 
 Functions in the language have been restricted enough to avoid imaginary numbers, rarely
 produce inexact numbers, make the exact versus inexact distinction rarely important, and
 avoid producing the special IEEE “numbers” @racket[-inf.0], @racket[+inf.0], @racket[+nan.0],
 @racket[-nan.0], @racket[#i+0], and @racket[#i-0].
 At the moment, the cost-benefit of modifying the parsing of the language to remove complex literals
 and the special numbers is too high, since students are unlikely to enter them accidentally.

 Historical convention considers numbers, integers, and (less so) natural numbers to be “types”,
 as opposed to the other subsets of numbers that have predicates here. This is reflected in their
 domains, but I'm not certain of the SE and pedagogical significance of maintaining the distinction.

 The sign predicates can be less error-prone and clearer at a glance than using the binary comparisons
 in @secref["numeric-comparison"]. See, e.g., the discussion for @racket[inc].

 Common uses for the parity predicates:
 @itemlist-inset[@item{play with simple unary predicates (see, e.g., @racket[inc])}
                 @item{implement the Collatz sequence}]}
The predicate @racket[number?] in @secref["data-types"] asks whether a value is a number.
@margin-note*{If you have not learned the term “predicate” yet:
 it means a function that produces a boolean.
 If you have not learned the term “boolean” yet:
 it means one of the special values @racket[#true] or @racket[#false].}
Predicates in this section determine whether a value is a particular kind of number.
@function*[([(integer? [v any/c]) boolean?] [(natural? [v any/c]) boolean?])]{
                                                                              
 @racket[(integer? v)]: is @racket[v] an integer
 (one of the the numbers @racket[0], @racket[-1], @racket[1], @racket[-2], @racket[2], @racket[-3],
 ...)?
 
 @racket[(natural? v)]: is @racket[v] a natural number
 (one of the numbers @racket[0], @racket[1], @racket[2], @racket[3], ...)?
 
 @interact[(integer? "hello") (natural? "hello")] @interact[ (integer? 3/2) (natural? 3/2)]
 @interact[(integer? 2) (natural? 2)] @interact[(integer? -2) (natural? -2)]
 @interact[(integer? 0) (natural? 0)]
 If @racket[natural?] was not already in the language, it could be defined this way:
 @(racketblock (define (natural? v)
                 (and (integer? v) (non-negative? v))))}
@function*[([(negative? [x number?]) boolean?]
            [(zero? [x number?]) boolean?]
            [(non-negative? [x number?]) boolean?]
            [(positive? [x number?]) boolean?])]{
 @racket[(negative? x)]: is the number @racket[x] less than zero?
 
 @racket[(zero? x)]: is the number @racket[x] zero?
 
 @racket[(non-negative? x)]: is the number @racket[x] at least zero?
 
 @racket[(positive? x)]: is the number @racket[x] greater than zero?
 
 @interact[(negative? -2) (negative? 0) (negative? 2)]
 @interact[(zero? -2) (zero? 0) (zero? 2)]
 @interact[(non-negative? -2) (non-negative? 0) (non-negative? 2)]
 @interact[(positive? -2) (positive? 0) (positive? 2)]
 Notice these predicates expect a number (and report an error otherwise).
 
 If these functions were not already in the language, they could be defined this way:
 @(racketblock (define (negative? x)
                 (< x 0)))
 @(racketblock (define (zero? x)
                 (= x 0)))
 @(racketblock (define (non-negative? x)
                 (>= x 0)))
 @(racketblock (define (positive? x)
                 (> x 0)))
 See also @racket[=], @racket[<], @racket[>], @racket[<=], and @racket[>=]
 in @secref["numeric-comparison"], and @racket[same?] in @secref["comparison"].}

@function*[([(even? [i integer?]) boolean?] [(odd? [i integer?]) boolean?])]{                                        
 Notice these predicates expect an integer (and report an error otherwise).
 
 @racket[(even? i)]: is integer @racket[i] even (a multiple of @racket[2])?
 
 @racket[(odd? i)]: is integer @racket[i] odd (not a multiuple of @racket[2])?
 
 @interact[(even? 6) (odd? 6) (even? 7) (odd? 7)]
 
 See also @racket[remainder] in @secref["integer-division"].}

@§§{Generating Random Numbers}
@function[(random [n (and/c natural? positive?)]) natural?]{
 A random natural number less than the number.
 @; @interact[(random 104) (random 104) (random 104)] @; Feeling somewhat lucky.
 @instructor-inset{
  The ISL/racket version allows a random number generator as the argument, for reproducibility.
  We should discuss whether reproducibility is worthwhile, and how to express it.

  The argument actually needs to be below 4294967088, but that doesn't seem worth mentioning
  (It's been made to generate an understandable error message).}
 
 @self{Original contract generates this error:
  “expects (or/c (integer-in 1 4294967087) pseudo-random-generator?) or a pseudo-random-generator”.}}

@§§{Number Functions Documented in Other Sections}
@mention-function[(number? [v any/c]) boolean? "data-types"]
@mention-function*[([(number->text [x number?]) text?]
                    [(text->number [t text?]) (or/c number? #false)]) "type-conversion"]
@instructor{Approximate comparisons aren't mentioned here, until/unless we decide to keep them.
 Unicode conversions aren't mentioned either, since they're not conceptually “about” number,
 nor core to the course.}

@§{Definitions}

@form[(define name expression)]{
 Define a variable, @racket[name], with the value of the @racket[expression].
                          
 The variable cannot have the same name as another variable or a function,
 and the variable cannot be used inside @racket[expression].}

@form/none[(#,(racketplainfont "define") (name parameter-name parameter-name ...)
                                         expression)]{
 Define a function, @racket[name], with the “parameters” @racket[parameter-name ...],
 and the “body” @racket[expression].

 When the function is called, the values of the arguments are substituted into the body
 in place of the @racket[parameter-name]s, and the function produces the value of that new expression.

 The function cannot have the same name as another function or a variable.}

@self{What terms do we @;{and @racketmodname[picturing-programs]} use for
 variables, parameters, arguments, and calling.
 Also @racketmodname[picturing-programs] considers the “header” to include the
 “@racketfont{(define }”.}

@§§{Temporary Definitions}
@instructor{We occasionally use this in lecture and in optional-to-understand project code,
 but Danny and I keep thinking we'd like to make this core to the course.}
@self{Catalogue our uses.}
@form[(local [definition ...]
        expression)]{
 Produce the value of @racket[expression] (the “body”), in the context of
 some temporary @racket[definition]s.

 Each @racket[definition] must be a @racket[define] form.
 
 Each @racket[definition] occurs, and then the value of the body @racket[expression] is computed
 and used as the result of the whole @racket[local] expression.
 
 The @racket[definition]s last only while the @racket[local] result is being produced,
 and the names they define can only be referred to within the @racket[local].
 The @racket[definition]s can define a name that already has a definition outside the @racket[local],
 in which case the @racket[local] @racket[definition] takes precedence inside the @racket[local].}

@§[#:tag "texts"]{Texts}
@instructor-inset{Text as a third type (actually fourth after functions) is useful to:
 @itemlist-inset[
 @item{compare/contrast with images and numbers}
 @item{signal early another important domain of application for the target audience}
 @item{highlight the distinction between names and values (e.g. @racketvalfont{"pi"})}
 @item{highlight the distinction between values and representations (the double-quotes)}]

 Texts show in the Interactions as literals, in particular (along with the other
 datatypes) as valid code that can be copy-pasted into the Definitions or Interactions
 to reproduce the text. So they always have the double-quotes, and special characters are
 shown as escapes, not evaluated for their meaning. For example, entering a multi-line text will
 echo it back with @racketvalfont{\n} in it, not print it across multiple lines
 (except for word-wrapping when reaching the edge of the display).

 The only place I've had to address special characters was when reading html/xml for a fun
 lecture example to motivate nested lists, but they didn't affect what we set out to accomplish
 with the html.

 I have a convention of using single-curly-quotes @tt{‘’} in comments to delimit code from surrounding
 text, and where one might use double-quotes, use curly-double-quotes “” to subtly distinguish from a
 text literal. I don't know that it matters, but in CSC324 I like it as a continual reminder that we
 should be aware of and precise about syntax and semantics. I've made it so that entering an opening
 curly-quote or curly-double-quotes inserts/delimits with the closing version.

 I've removed the ISL/racket character type (with literals such as @racket[#\a] and @racket[#\space]),
 in favour of texts of length 1. Pedagogically, there isn't much point in distinguishing the two,
 and characters as a (data) type are less and less important (being mainly a resource optimization).
 All the functions from the
 @seclink["Characters"
          #:tag-prefixes '("htdp-intermediate")
          #:doc '(lib "scribblings/htdp-langs/htdp-langs.scrbl")]{
  @racketmodname[lang/htdp-intermediate] “Characters”} section were removed,
 and instead the text counterparts (two of which we added) were considered for inclusion,
 possibly after renaming.

 Modes and colors of geometric images are texts: I delay shapes a bit for that reason.

 We use @racket[text-length] and @racket[text-join] early, as two simple text functions,
 with distinct contracts @racketfont{(text → number)} and @racketfont{(text ... → text)}.
 Both test the value-representation distinction nicely.

 @racket[text-join] could seem pointless to students initially, since one can just type out
 the combined texts, but quickly becomes worthwhile in functions, that, e.g., construct phrases
 by filling in a template or adding punctuation.

 Danny uses @racket[sub-text] for examples as well, more than I do.
 
 Later, @racket[text-length] and @racket[text-join] are good for examples of @racket[map]ping
 and @racket[apply]ing, and just useful for text manipulation.

 When different types have functions with similar behaviour, there's often a “main” one for
 a particular type, and other versions have the name prefixed with “@racketidfont{<type>-}”.
 List functions @racket[length] and @racket[join] are both considered their main versions,
 and they are another opportunity to test students on distinguishing data-types.}
@self{
 @inset{2017W: Introduced in 3rd lecture (@tt{Jan-13.rkt}), catalyzed by shape function arguments.
  Discussed fully, except for indexing, including @racket[text-length] and @racket[text-join].}}

A text is a sequence of characters, entered by surrounding the characters with double-quotes.
The double-quotes are not considered part of the text itself,
or put another way: they are not characters @racket[in] the text itself. The double-quotes are
just there to know where the text starts and ends, so the characters they surround are not
interpreted as code.

There are various special “characters” in computing, for example the character to represent the
end of a line in text that spans multiple lines. And putting a double-quote itself inside a text
as one of its characters requires some way to indicate that the double-quote is not just signalling
the end of the text. We avoid these technicalities, except if they become necessary or useful
for a particular text processing task. If they come up naturally we will address them in that context,
so except where we explicitly state otherwise, you can safely assume none of our texts will contain
double-quotes as characters, nor span multiple lines.

@function[(text-length [t text?]) natural?]{
 The number of characters in the text.
 @interact[(text-length "hello world")]}
@function[(text-join [t text?] ...) text?]{
 A single text formed by joining the given texts.
 @interact[(text-join "hello" " " "world" ", " "good bye")
           (text-join "hello")
           (text-join)]}
@function[(sub-text [t text?] [i natural?] [j natural?]) text?]{
 The part of the text @racket[t] starting from the character at index @racket[i]
 (counting @italic{from zero}), up to @italic{but not including} the character at index @racket[j].
 @interact[(sub-text "hello world" 1 7) (sub-text "hello world" 2 2)]
 The index @racket[i] must be less than or equal to the index @racket[j], which must be
 less than or equal to the length of the text @racket[t], otherwise an error is reported.
 @instructor-inset{Pre-condition: @racket[(<= i j (text-length s))]
  Removed the two-argument single-index version that defaults to using the end of the text
  as the second index, i.e. @racketfont{(substring s i (string-length s))}.
  It's not useful enough, and implementing it would be more instructive.}}
@function[(sub-text? [t text?] [u text?]) boolean?]{
 Does the first text appear inside the second one?
 @interact[(sub-text? "it" "kitten")]
 @instructor{If we're keeping this only to use for single characters, we should consider
  instructively building that version instead, from @racket[element?] and @racket[text->list],
  or not building that and just arranging our code to work on lists of characters most of the time.}}
@function[(character? [v any/c]) boolean?]{ Is the argument a text whose length is one?
 @interact[(character? 1) (character? "a") (character? "abc")]}
@instructor{The predicate @racket[character?] is included for the moment since @racket[text->list]
 is currently included. The other functions using it in their contract are: @racket[character],
 @racket[character->unicode], @racket[unicode->character], which are not yet considered core.}
@function[(character [t text?] [i natural?]) character?]{
 The character in the text at the given index (counting @italic{from zero}).
 @interact[(character "hello world" 1)]
 If @racket[character] was not in the language, it could be defined this way:
 @racketblock[(define (character s i)
                (sub-text s i (inc i)))]
 @;{@instructor{We should consider removing this, as instructively buildable from @racket[sub-text].
   Renamed from ISL/racket's @racket[string-ith], for consistency with @racket[list-ref],
   and removed ISL/racket's @racketidfont{string-ref}, which produces an ISL/racket character.}}}
@;@§§[#:tag "text-non-core"]{Text Functions Not Traditionally Core to the Course}
@self{A listing of which functions from the
 @seclink["Strings" #:tag-prefixes '("htdp-intermediate")
          #:doc '(lib "scribblings/htdp-langs/htdp-langs.scrbl")]{
  @racketmodname[lang/htdp-intermediate] “Strings”} section we removed.
 @inset{Using the character data-type:
  @itemlist-inset[@item{@racketidfont{list->string} and @racketidfont{string->list}:
                     kept 1-letter string version of @racketidfont{string->list}}
                  @item{@racketidfont{string}: a non-list variadic @racketidfont{list->string}}
                  @item{@racketidfont{make-string}: constant string,
                     removed the generalized string version @racketidfont{replicate} as well}
                  @item{@racketidfont{string-ref}: we kept the 1-letter string version for now}]
  Others: @itemlist-inset[
 @item{@racketidfont{implode}: I recommend deriving/building it}
 @item{@racketidfont{string->symbol}: we don't use symbols}
 @item{@racketidfont{replicate}: constant texts don't seem very useful, easily built via lists}
 @item{@racketidfont{format}: too idiosyncratic, @racketidfont{text-join} should be sufficient}
 @item{@racketidfont{string-copy}: weird}]

  Not having the character type means that some functions don't need the “@racketfont{<type>-}”
  distinguisher, but except for the character type predicates (where we would probably want to rename
  @racket[text-numeric?] to @racketidfont{digits?} to avoid an obvious confusion) and case changers,
  it doesn't seem a major win anywhere.
  
  @;{Should @racket[text-contains?] be renamed to something like @racketidfont{is-subtext?}.}}}
@instructor{These aren't currently in the core of the course, but could be useful for projects.
 I've used some in lecture for public-key cryptography, e.g. @racket[character->unicode] and
 @racket[unicode->character] (original ISL names: @racket[string->int] and @racket[int->string]).}
@§§[#:tag "character-types"]{Types of Characters}
@instructor{Vacuous truth (e.g. all of these being true for the empty text) is a reasonable
 cs/math concept to add to the course.}
@function[(letters? [t text?]) boolean?]{ Does the text contain only letters?
 @interact[(letters? "Cat") (letters? "Cat?") (letters? "123")]}
@function[(digits? [t text?]) boolean?]{ Does the text contain only digits?
 @interact[(digits? "123") (digits? "-23")]}
@function*[([(upper-case? [t text?]) boolean?] [(lower-case? [t text?]) boolean?])]{
 Does the text contain only upper-case letters, or only lower-case letters?
 @interact[(upper-case? "CAT") (upper-case? "kITTY") (upper-case? "CAT?")
           (lower-case? "cat") (upper-case? "kITTY") (lower-case? "cat?")]
 See also @racket[upper-case] and @racket[lower-case] in @secref["case-conversion"].}
@function[(whitespace? [t text?]) boolean?]{
 Does the text contain only blanks (spaces, or other special “invisible” characters)?
 @interact[(whitespace? "  ") (whitespace? " a ") (whitespace? "\n")]}
@§§[#:tag "case-conversion"]{Case Conversion}
@function*[([(upper-case [t text?]) text?] [(lower-case [t text?]) text?])]{
 The upper, or lower, case version of the text.
 @interact[(upper-case "kIttY!") (lower-case "kIttY!")]
 See also @racket[upper-case?] and @racket[lower-case?] in @secref["character-types"].}
@§§{Alphabetic Order}
@instructor{It's possible we would want these to allow zero or one argument.}
@;{See also @racket[text=?] in @secref["equality"].}
@;{See @;{also @racket[text-ci=?] in @secref["equality"], and}
 also @racket[upper-case?] and @racket[lower-case?] in @secref["character-types"].}
@function[(alphabet-order? [t text?] [u text?] [v text?] ...) boolean?]{
 Are the texts in alphabetic order?
 
 Upper-case letters are considered to be alphabetically before lower-case letters
 (for historical reasons).
 @interact[(alphabet-order? "al" "beth" "dee") (alphabet-order? "Beth" "al" "beth" "dee")
           (alphabet-order? "zebra" "koala" "kitten") (alphabet-order? "kitten" "koala" "zebra")]}
@§§[#:tag "unicode"]{Unicode Numbering of Characters}
@deftogether[(@function+[(character->unicode [s character?]) natural?]
               @defthing[#:kind "alias" character→unicode character->unicode]
               @function+[(unicode->character [n natural?]) character?]
               @defthing[#:kind "alias" unicode→character unicode->character])]{
 The natural number representing a character, and vice-versa, according to the
 the @link["https:www.wikipedia.org/wiki/Unicode"]{Unicode encoding for characters}.
 @interact[(character->unicode "A") (unicode->character 65)
           (character->unicode "⚘") (unicode->character 9880)]}

@§[#:tag "comparison"]{Comparison}
The functions @racket[same?], and @racket[differ?], determine whether two values
are the same, or different.

See @racket[=] for the version of @racket[same?] restricted to numbers, which can also be used with
more than two numbers.
@function*[([(same?   [v₁ any/c] [v₂ any/c]) boolean?] [(differ? [v₁ any/c] [v₂ any/c]) boolean?])]{
 @interact[(same? (list 1 2) (list (dec 2) (+ 1 1))) (differ? (list 1 2) (list (dec 2) (+ 1 1)))]}
@instructor{[2018] Removed restricted and approximate versions, and the commentary:
 @inset{There are also restricted versions of @racket[equal?],
  that work only on values of a particular type, and report an error otherwise.
  In situations where you expect values to be of a particular type, then using the restricted versions
  of @racket[equal?] can be clearer to someone reading your code.
  Using these versions can also report some errors earlier (ones where your code produced values
  of the wrong type, and then compares them) and more closely related to the true source,
  making them more useful.}}
@;{@function[(image=? [image₁ image?] [image₂ image?]) boolean?]{
  Are the two images equal?
  @interact[(eval:alts (image=? #,(circle 5 "solid" "green") #,a-circle)
                       (image=?   (circle 5 "solid" "green")   a-circle))
            (eval:alts (image=? #,(circle 10 "solid" "green") #,a-circle)
                       (image=?   (circle 10 "solid" "green")   a-circle))]}
 @function[(text=? [t text?] [u text?] [v text?] ...) boolean?]{
  Are all the texts the same?
  See also @racket[text-ci=?] in @secref["approximately"] for case-insensitive comparison.
  @interact[(text=? "hello" "world") (text=? "bye" "bye")]}
 @function[(boolean=? [b1 boolean?] [b2 boolean?]) boolean?]{
  Are the two booleans equal? @interact[(boolean=? #true #false)]}}
@;{@§§[#:tag "approximately"]{Approximately}
 @function[(text-ci=?  [t text?] [u text?] [v text?] ...) boolean?]{
  Are all the texts the same, ignoring the distinction between upper and lower case?
  @interact[(text-ci=?  "hello" "HellO")]}
 @function[(=~ [x number?] [y number?] [z (and/c number? non-negative?)]) boolean?]{
  Are numbers @racket[x] and @racket[y] within @racket[z] of each other?
  @interact[(=~ 1.01 1.0 .1) (=~ 1.01 1.5 .1)]}
 @function[(equal~? [v any/c] [w any/c] [z (and/c number? non-negative?)]) boolean?]{
  Are @racket[v] and @racket[w] approximately the same:
  the same if numbers in @racket[v] and @racket[w] are considered the same
  when they are within @racket[z] of each other?
  @interact[(equal~? (list 1.01 1.0) (list 1.01 .99) .2)]}}

@§{Testing}
@form[(check-expect expression expected-expression)]{
 Checks that the first @racket[expression]
 produces the same value as the @racket[expected-expression].
 If the values are different, then @racket[check-expect] reports it as a failed test case.}
@instructor-inset{
 Jump to an error with Command-. . We need to teach students that errors stop evaluation,
 the difference between those and failing test cases, and also be aware of what happens when there's
 an error generated by a test case (in particular, the behaviour is asymmetric: it depends on
 whether it occcurred in the “actual” expression versus “expected” expression.

 There's infrastructure for testing code coverage, which we don't use but should consider.
 It's jarring in the first part of the course when there's an error, so I've disabled it for now.
 This stops a discrete, but still a bit distracting, status-line message.}
@form[(time expression)]{
 Produces the same value as @racket[expression],
 but also prints the amount of time it took to calculate the value.
 The time is reported in milliseconds, measured in three ways
 (the distinctions between the three times are not important for this course).}

@§{Booleans}
@instructor{Unlike for functions, there are no technicalities in having canonical literals for
 two special values. Variable names for them (or worse, in lieu of them) are an unnecessary
 indirection. The full (non-identifier) names for the literals are more than convenient enough
 to not need abbreviation, and the abbreviations are in fact hard to distinguish at a glance
 --- even moreso in student handwriting on tests. The abbreviations can also be capitalized,
 which has some precedent in cs/math but not familiar nor commmon enough, nor important conceptually,
 to warrant inclusion in the course --- and they're still harder to distinguish visually
 than one might like. At the moment, the cost-benefit of modifying the parsing of the language
 to remove the synonymous literals is too low, but the variables have been removed.}
@instructor{[2018] Removed the deprecation warnings for ‘true’ and ‘false’ from the language,
 and the following commentary:
 @inset{There are a few traditional alternatives for referring to these two values.
  If you use only materials referred to in this instance of the course, then the alternatives
  should not appear, and you can safely ignore the rest of this paragraph.
  But if you look outside this instance of the course (for example, to some previous versions
  of the course), you might encounter the synonyms @racket[true] or @racket[#t] for @racket[#true],
  and @racket[false] or @racket[#f] for @racket[#false].}}

The Boolean type has two values: @racket[#true] and @racket[#false].
@function[(not [b boolean?]) boolean?]{Negation of a boolean. @interact[(not #false) (not #true)]}
@§§{Conditional Expressions}
These forms are special because they selectively (“conditionally”) compute the value of
some of their component expressions.

Since these forms need to take control, they cannot be used as arguments.
For example, the following expressions each report an error:
@interact[(function? or) (apply if #true 1 2) (map and
                                                   (list #true #true)
                                                   (list #false #true))]
@form[(and expression expression expression ...)]{
 Do @italic{all} of the @racket[expression]s produce the value @racket[#true]?

 The value of each @racket[expression] is computed, in order. This process ends early if either:
 @itemlist-inset[
 @item{An @racket[expression] produces @racket[#false], in which case @racket[#false]
   is used immediately (“short-circuiting”) as the result of the whole @racket[and] expression.}
 @item{An @racket[expression] produces a non-boolean value, in which case an error is reported.}]}
@form[(or expression expression expression ...)]{
 Do @italic{at least one} of the @racket[expression]s produce the value @racket[#true]?

 The value of each @racket[expression] is computed, in order. This process ends early if either:
 @itemlist-inset[
 @item{An @racket[expression] produces @racket[#true], in which case @racket[#true]
   is used immediately (“short-circuiting”) as the result of the whole @racket[or] expression.}
 @item{An @racket[expression] produces a non-boolean value, in which case an error is reported.}]}
@form[@;#:literals (else)
 (if [condition-expression result-expression]
     [condition-expression result-expression]
     ... 
     [#,(defidform/inline else) result-expression])]{                                      
 The value of the @racket[result-expression] corresponding to
 the first @racket[condition-expression] that produces @racket[#true] or is @racket[else].
 
 The paired condition-and-result expressions are known as “if clauses”.

 People traditionally group a condition-result pair with square brackets
 (although parentheses can also be used) to:
 @itemlist-inset[@item{Emphasize that the pairing doesn't mean the usual function call.}
                 @item{Make it easier to see the two contained expressions,
                   which are themselves often parenthesized forms.}]
 The value of each @racket[condition-expression] is computed, in order, until either:
 @itemlist-inset[
 @item{A condition produces @racket[#true], or they were all @racket[#false]
   and the @racket[else] was reached, in which case  
   the value of the @racket[result-expression] for that clause is computed
   and its value is used as the result of the whole @racket[if] expression.}
 @item{A condition produces a non-boolean value, in which case an error is reported.}]}
@instructor{[2018] At least two clauses are now required, and we step without ever showing
 a single-else-clause intermediate step}
@self{Square brackets versus parentheses (versus braces) in the language: warnings? enforce? etc.
 @racket[big-bang] would be the other place to use them. And adjust DrRacket's formatting.}
@§§{Boolean Functions Documented in Another Section}
@mention-function[(boolean? [v any/c]) boolean? "data-types"]
@;{mention-function[(boolean=? [b1 boolean?] [b2 boolean?]) boolean? "equality"]}

@§[#:tag "data-types"]{Types of Data}
@instructor-inset{I expect the concept of a datatype, the names of the six main datatypes,
 and their meanings, to be a core learning outcome. It's also necessary for understanding the course
 without excessive reminders and redundant examples. The understanding can be operationally
 expressed/tested by students:
 @itemlist-inset[@item{writing contracts}
                 @item{reacting to, and predicting, errors reported by contracts}
                 @item{knowing the data-type predicates and their behaviours}]
 In particular: reminders of the names, names of the predicates, and behaviours of the predicates,
 would be removed from test reminders in the latter part of the course.
 
 We should consider requiring students learn some of the numeric sub-types,
 and learn to be specific about list and text sub-types. This could include “list-of” types.}

The main types of data (“data-types”) in this course are:
@itemlist-inset[@item{image} @item{number} @item{function} @item{text} @item{boolean} @item{list}]

@§§{Showing Data Types in DrRacket}
To get used to the names of our main data-types and distinguishing which values have which type,
you can make DrRacket show you the type of every result value in the Interactions,
by following these steps:
@(local [(define (include-image′ name) (include-image #:scale 2/5 name))]
   @inset{@itemlist[#:style 'ordered
                    
                    @item{Select the DrRacket menu item “Language / Choose Language ...”.
                
                      @(include-image′ "language-dialog.png")}
                    
                    @item{In the language chooser make sure that “CSC 104” is selected,
                      and then click the “Show Details” button.
                      
                      @(include-image′ "language-details.png")}
                    
                    @item{Select the option “Prefix result values with their types”,
                      then click the “Ok” button.
                      
                      @(include-image′ "language-details-show-types.png")}
                    
                    @item{The language shown in the lower left corner of DrRacket should now say
                      “CSC 104 custom”, and be highlighted (confirming that a language option was
                      changed).

                      @(include-image′ "language-changed.png")}
             
                    @item{Click the “Run” button so that the change takes effect.
                      The highlighting of “CSC 104 custom” should now go away,
                      and the Interactions area should now say the language is “CSC 104 [custom]”.
            
                      Enter a value into the Interactions area to see the new behaviour.

                      @(include-image′ "interactions-types.png")}]})

@instructor-inset{
 We should consider having this on by default at the start of the course, because it already
 emphasizes the distinction between two images and two images combined with @racket[above]
 or @racket[beside]. It might also add emphasis to the demonstration, and if students do it,
 of entering separate unparenthesized expressions. We should also consider a message to print
 in the latter situation, reminding students of how to perform an operation.
                  
 The colour for printing the type name is determined by the DrRacket Preference
 “Color / Background / Default Text”.}

@§§{Type Predicates}
Each of our main data-types has a corresponding predicate (a function producing a boolean)
that determines whether a value has that type.
The name of the predicate is the name of the data-type, suffixed with a “@racketidfont{?}”.
@function[(image? [v any/c]) boolean?]{
 Is the argument an image? @interact[(eval:alts (image? #,a-circle) (image? a-circle))]}
@function[(number? [v any/c]) boolean?]{
 Is the argument a number? @interact[(number? -1.04) (number? "hello")]
 See also @secref["numeric-types"].}
@function[(text? [v any/c]) boolean?]{
 Is the argument a text? @interact[(text? "hello") (text? 104) (text? "")]}
@instructor{See also @racket[character?] in @secref["texts"].}
@function[(boolean? [v any/c]) boolean?]{
 Is the argument a boolean? @interact[(boolean? #false) (boolean? #true) (boolean? 104)]}
@function[(list? [v any/c]) boolean?]{
 Is the argument a list?
 @interact[(list? (list 104 "hello world" (+ 1 2 3) (triangle 30 "solid" "blue")))
           (list? "hello world") (list? (+ 1 2 3))
           (list? (list 1 2 3)) (list? (list)) (list? (list 104))]
 See also @racket[empty?] in @secref["lists"].}
@function[(function? [v any/c]) boolean?]{
 Is the argument a function? @interact[(function? circle) (function? +) (function? pi)]
 See also @secref["function-types"].}
@instructor{The only exceptions to the question-mark convention for predicates in the course are
 the already-familiar numeric comparison names, and the dedicated boolean function @racket[not]
 and forms @racket[and] and @racket[or].}
@self{Resisting the urge to show the difference between the form of an expression and its value,
 and other concepts, in this @italic{reference}.
 Just making sure to show a general case first, and include each special case,
 with at least one value not of the type.
 Gather the various nice example expressions that superficially could appear to be a different type
 than they are, somewhere else.}

@§§[#:tag "type-conversion"]{Converting Between Types}
@deftogether[(@function+[(number->text [x number?]) text?]
               @defthing[#:kind "alias" number→text number->text])]{
 A text representation of the number. @interact[(number->text 104)]}
@deftogether[(@function+[(text->number [t text?]) (or/c number? #false)]
               @defthing[#:kind "alias" text→number text->number])]{
 The number corresponding to a text representation of a literal number,
 or @racket[#false] if the text is not in the form of a literal number.
 @interact[(eval:alts (text->number "-1.5") (inexact->exact -1.5))
           (text->number "2/3") (text->number "(+ 2 3)")]
 For functions to convert between a character and its Unicode number, see @secref["unicode"].}
@deftogether[(@function+[(text->image [text text?] [font-size (and/c number? (between/c 1 255))]
                                      [color color?]) image?]
               @defthing[#:kind "alias" text→image text->image])]{
 An image of the text, with a given font size and color. @interact[(text->image "Hi!" 30 "olive")]}
@deftogether[(@function+[(text->list [t text?]) (listof character?)]
               @defthing[#:kind "alias" text→list text->list])]{
 A list of the characters of the text. @interact[(text->list "a kitten")]
 @instructor-inset{
  This was renamed from ISL's @racket[explode].
  ISL/racket's @racketidfont{string->list} produces a list of (ISL/racket) characters.
  Removed ISL's inverse @racket[implode], since requiring the elements have length 1 doesn't seem
  worthwhile, and even potentially limiting. Instead we should use and expect the students to
  understand the instructive definition:
  @racketblock[(define (list->text texts)
                 (apply text-join texts))]}}

@§[#:tag "lists"]{Lists}
A list bundles (“contains”, “holds”) zero or more values (known as its “members”, “elements”,
or “items”), into a single “compound” value. One way to create a list is with @racket[list],
which is also how lists are shown in the Interactions.

The list with no elements is known as the “empty” list.
@instructor{Removed deprecation warning for ‘empty’ from the language, and the commentary:
            
 There are two traditional alternatives for referring to the empty list.
 If you use only materials referred to in this instance of the course, then the alternatives
 should not appear, and you can safely ignore the rest of this paragraph.
 But if you look outside this instance of the course (for example, to some previous versions
 of the course), you might encounter the synonyms @racket[empty] or @racket['()].}
@instructor{The HtDP curriculum builds lists from the empty list using @racket[cons]
 (a restricted version of @racket[adjoin]) so needs a way to refer to the empty list.
 As with booleans, they've moved away from a variable (@racket[empty]) referring to
 an unprintable hidden value, to a literal (@racket['()], which generalizes later
 when they cover quoting and quasiquoting). Their curriculum doesn't have @racket[list]
 initially, and shows the construction of the list as nested calls to @racket[cons],
 since that is how they create and process them initially). Since we start with the
 @racket[list] “abbreviation”, I made @racket[(list)] our canonical representation
 of the empty list, for uniformity.}
@self{@inset{
  At the moment, @racket['()] reports an error different than other uses of quote and quasiquote:
  @inset{@racketplainfont{function call:
    expects a function after the open parenthesis, but nothing's there}}
  --- but other uses of quote and quasiquote do indeed report an error, preventing that hard-to-debug
  program behaviour a student once encountered.

  I've removed the historribly named @racket[car], @racket[cdr], @racket[null], etc.,
  @racket[cons] and @racket[cons?] (although we should consider a version of the latter),
  the functions using @racket[eq?] or @racket[eqv?] for equality, and
  the poorly-behaved @racket[member]. Need to go through and record the rationale
  for any unmentioned removed list HOFs.}}
@function[(list [e any/c] ... ) list?]{ A list whose elements are the arguments.
 @interact[(list 104 "hello" #true) (list "hello") (list)]}
@function[(length (l list?)) natural?]{ The number of elements in a list.
 @interact[(length (list 104 "hello" #true))]}
@function[(first [l list?]) any/c]{ The first element of a (non-empty) list.
 @interact[(first (list 104 "hello" #true)) (first (list))]
 If @racket[first] was not already in the language, it could be defined this way:
 @(racketblock (define (first l)
                 (element l 0)))}
@function[(rest [l list?]) any/c]{ A version of the list, without the first element.
 @interact[(rest (list 104 "hello" #true)) (rest (list))]
 @instructor{@(racketblock (adjoin (first l) (rest l)) ≡ l)}}
@function[(empty? [v any/c]) boolean?]{
 Is the value the empty list?
 @interact[(empty? (list)) (empty? (list 1 2 3)) (empty? 104)]
 If @racket[empty?] was not already in the language, it could be defined this way:
 @(racketblock (define (empty? v)
                 (same? v (list))))
 @self{Should this be restricted to lists? Is it like @racket[natural?] or like @racket[zero?].}}
@function*[([(second [l list?]) any/c] [(third [l list?]) any/c] [(fourth [l list?]) any/c])]{
 The second, third, or fourth element of a list.
 @interact[(second (list 104 "hello" #true))
           (third (list 104 "hello" #true))
           (fourth (list 104 "hello" #true))]
 If @racket[second], @racket[third], and @racket[fourth] were not already in the language,
 they could be defined with @racket[element], for example:
 @(racketblock (define (third l)
                 (element l 2)))
 @instructor-inset{
  Another approach would be: @(racketblock (define (third l)
                                             (first (rest (rest l)))))

  Removed @racket[fifth] through @racket[eighth] as not very useful.}}
@function[(element [l list?] [i natural?]) any/c]{
 The element in a list at the given index, counting @italic{from zero}.
 @interact[(element (list 104 "hello" #true) 0) (element (list 104 "hello" #true) 1)
           (element (list 104 "hello" #true) 2) (element (list 104 "hello" #true) 3)]
 @instructor-inset{ The list of valid indices is: @racket[(range 0 (length l) 1)].
  @;{Should we choose a more meaningful name, such as @racketidfont{element-at}?}}}
@function[(join [l list?] ...) list?]{ A single list containing the elements from all the lists
 (also known as “the concatenation” of all the lists).
 @interact[(join (list 1 2) (list "a" "b")) (join (list 1 2)) (join)]}
@function[(adjoin [e any/c] [l list?]) list?]{
 A version of list @racket[l] with @racket[e] at the front.
 @interact[(adjoin 104 (list "hello" #true))]}
@function[(sub-list [l list?] [i natural?] [j natural?]) list?]{
 The elements of the list @racket[l] starting from the element at index @racket[i]
 (counting @italic{from zero}), up to @italic{but not including} the element at index @racket[j].
 @interact[(sub-list (list "hello" 123 (list 4 5 6) #false) 1 3)
           (sub-list (list "hello" 123 (list 4 5 6) #false) 2 2)]
 The index @racket[i] must be less than or equal to the index @racket[j], which must be
 less than or equal to the length of the list @racket[l], otherwise an error is reported.}
@function[(range [start number?] [end number?] [step number?]) (listof number?)]{
 The list of numbers from @racket[start] to @italic{but not including} @racket[end],
 changing by @racket[step].
 @interact[(range 2 10 3)]
 @instructor{There are no other pre-conditions on the arguments: there are lots of situations
  that can produce the empty list. Also, stepping by zero doesn't terminate, which I'll consider
  preventing.}}
@function[(reverse [l list?]) list?]{ A version of the list, with the elements listed in reverse order
 (the last element first, the second-last element second, etc.).
 @interact[(reverse (list 104 "hello world" (+ 1 2 3) (triangle 30 "solid" "blue")))]}
@function*[([(map [f unary?] [l list?]) list?]
            [(map [f binary?] [l₁ list?] [l₂ list?]) list?])]{
 Performs a function on each value in a list, one at a time, separately,
 and gathers the resulting values into a list.
 @interact[(map text-length (list "rick" "and" "morty"))
           (list (text-length "rick") (text-length "and") (text-length "morty"))]
 The algebraic step for this is:
 @(racketblock (map f (list a b c  _...))
               → (list (f a) (f b) (f c) _...))
 It can also be used with two lists of the same length, in which case
 it performs the function on pairs of elements from the two lists.
 @interact[(map +
                (list 1 2 3)
                (list 40 50 60))
           (list (+ 1 40) (+ 2 50) (+ 3 60))]
 The algebraic step for this is:
 @(racketblock (map f
                    (list a₁ b₁ c₁ _...)
                    (list a₂ b₂ c₂ _...))
               → (list (f a₁ a₂) (f b₁ b₂) (f c₁ c₂) _...))
 @instructor{The version with two lists comes up in base representation.
  I've removed the versions with more than two lists: we don't use those.}}
@function[(apply [f function?] [l list?]) any/c]{
 The @italic{one} result of calling @racket[f] where the arguments have been grouped into a list.
 @interact[(apply maximum (list 3 1 4 1 5 9 2))
           (maximum 3 1 4 1 5 9 2)]
 The algebraic step for this is:
 @(racketblock (apply f (list a b c _...))
               → (f a b c _...))
 @instructor{I've removed the variadic version that allows individual arguments to be included
  before the list of remaining arguments. Instead, any extra arguments can easily and instructively
  be added to the list of arguments with @racket[adjoin].}}
@function[(repeats [f unary?] [seed any/c] [n natural?]) list?]{
 The list with elements @racket[a], @racket[(f a)], @racket[(f (f a))], ..., of length @racket[n].
 @interact[(repeats squared 2 3)
           (list 2 (squared 2) (squared (squared 2)))]
 The algebraic step for this is:
 @(racketblock (repeats f a n)
               → (list a (f a) (f (f a)) ...))
 @instructor{I added this to the course a while ago, in particular to prepare for @racket[big-bang],
  but also generally for some iteration.}}
@function[(sift [p? unary?] [l list?]) list?]{
 A version of the list @racket[l], keeping only the elements for which @racket[p?]
 produces @racket[#true].
 @interact[(sift even? (list 3 1 4 1 5 9 2))]
 @instructor-inset{
  @;{Students forget it's inclusive: “filter in” versus “filter out”.}
  I have a process for showing intermediate processing by putting each element on its own line
  and then putting a comment beside it of @racket[#true] or @racket[#false].
  We should consider an official “step” for tests: circle elements for which the predicate is true.}
 @;{@instructor{@filter-steps}}}
@instructor{These aren’t currently in the core of the course, but could be useful for projects.}
@function[(sort [comparison binary?] [l list?]) list?]{
 A version of the list @racket[l], with the elements ordered according to predicate @racket[comparison].
 @interact[(sort > (list 3 1 4 1 5 9 2 6))
           (sort alphabet-order? (list "hello" "there" "friend"))]}
@deftogether[(@function+[(element? [e any/c] [l list?]) boolean?]
               @defthing[#:kind "alias" ∈ element?])]{
 Is @racket[e] an element of the list @racket[l]?
 @interact[(element? "hello" (list 2 "hello" #false))]}
@function[(remove [e any/c] [l list?]) list?]{
 A version of @racket[l], with all occurrences of @racket[e] removed.
 @interact[(remove "hello" (list 2 "hello" #true "hello"))
           (remove "bye" (list 2 "hello" #true "hello"))]}
@§§{List Functions Documented in Other Sections}
@;{
@mention-function[(list? [v any/c]) boolean? "data-types"]
@§§{Interesting List Functions Removed}
@function[(make-list [n natural?] [v any/c]) list?]{ A list of @racket[n] copies of @racket[v].
 @interact[(make-list 3 "hello")]
 @instructor{This could be useful to build one of the lists for two-list @racket[map]}.}
@;{@function[(remove [v any/c] [l list?]) list?]{
  A version of the list @racket[l], with the first occurrence of @racket[v] removed,
  or just @racket[l] if @racket[v] is not an element of @racket[l].
  @interact[(remove "hello" (list 2 "hello" #true "hello"))
            (remove "bye" (list 2 "hello" #true "hello"))]}}
@function[(assoc [v any/c] [l (listof (list/c any/c any/c))]) (or/c list? #false)]{
 The first pair in the list of two-element lists @racket[l], whose first element is @racket[v],
 otherwise @racket[#false].
 @interact[(assoc "hello" (list (list "world" 2) (list "hello" 3) (list "good" 0)))
           (assoc "goodbye" (list (list "world" 2) (list "hello" 3) (list "good" 0)))]
 @instructor{If we want something like this, we should discuss its name, and other possibilities
  for a richer dictionary API for lists of two-element lists.}}
 @instructor-inset{
  Removed @racket[build-list]: use @racket[map] and @racket[range] instead.

  If we find uses for @racket[andmap] and @racket[ormap] we should consider
  names such as “@racketidfont{for-all?}” and “@racketidfont{for-some?}”.

  If we want @racket[memf], we should make a better name, and have the function's behaviour
  be to return just the value.

  If we want @racket[argmin] and @racket[argmax] we should consider better names
  (even though those are traditional in cs/math),
  such as “@racket[minimizer]” and “@racket[maximizer]”.}
}

@§{Animation}
@instructor{Fuller documentation to come. Of note is that now the @racketidfont{to-draw} may be
 omitted, defaulting to @racket[identity] (which Picturing Programs calls @racketidfont{show-it}).}
@self{@itemlist-inset[
 @item{@racket[(to-draw draw-expr width-expr height-expr)]}
 @item{@racket[(on-tick tick-expr rate-expr limit-expr)]}
 @item{@racket[(state expr)]}
 @item{@racket[(on-key key-expr)]}
 @item{@racket[(on-mouse mouse-expr)]}
 @item{@racket[on-draw] deprecated}
 @item{@racket[(stop-when stop-expr last-scene-expr)]}
 @item{@racket[(record? r-expr)]}]}
@form[(big-bang start
        [on-tick updater]
        [to-draw drawer])]{
 Launch a window with an animation.
 
 The animation frames are the same as the images in the following list
 @racketblock[(map drawer (repeats updater start n))]
 except @racket[big-bang] does not actually create that list, but goes on forever,
 not limited to generating some particular number of frames @racket[n].
 For example, @interact[(define (rotated-star angle)
                          (rotate (star 25 "solid" "blue") angle))
                        (eval:alts (big-bang 0
                                     [on-tick inc]
                                     [to-draw rotated-star])
                                   (void))]
 creates an animation from the following images:
 @interact[(map rotated-star (repeats inc 0 5))]
 A @racket[big-bang] expression also produes a result value, after you close its animation window:
 the updated version of @racket[start] that was used to produce the last image shown.}
@form/none[(big-bang start
             [on-tick updater])]{
 The animation frames are the same as the images in the following list:
 @racketblock[(repeats updater start n)]
 except @racket[big-bang] does not actually create that list, but goes on forever,
 not limited to generating some particular number of frames @racket[n].
 For example,
 @racketblock[(big-bang (triangle 25 "solid" "blue")
                [on-tick clockwise])]
 creates an animation from the following images:
 @interact[(repeats clockwise (triangle 25 "solid" "blue") 5)]}
@form/none[(big-bang start
             [to-draw drawer])]{ The animation frames are all the same: @racket[(drawer start)].}
@form/none[(big-bang start)]{ The animation frames are all the same: the @racket[start] image.}

@§{Miscellaneous}
@function[(identity [v any/c]) any/c]{
 Just the argument, unchanged.
 @interact[(identity 104) (eval:alts (identity #,a-circle) (identity a-circle))]
 @instructor{We might use this with @racket[big-bang].}}
@deftogether[(@function+[(url->browser [url text?]) void?]
               @defthing[#:kind "alias" url→browser url->browser])]{
 Open the web page at the url address, in your default browser.}
@function[(current-seconds) natural?]{
 The current time in seconds (counting from some platform-specific starting time).
 @; @interact[(current-seconds)]
 @instructor{This could be used with @racket[big-bang], or measuring runtime
  (but see @racket[time]).}}

@§{Packages/Libraries}
@form[(require library-name)]{
 Make the definitions from a “library” (an installed package, or a special file)
 available to use in the current program.

 If we use a library in the course, then we will describe how to install it (if necessary),
 and give you the exact expression to write in order to use the library.}

@instructor{The definitions accessible from a library are the ones exported by
 its @racket[provide] expressions.
 In particular you wouldn't implement the library in the course's language,
 nor do we expect students to create libraries.
 
 @nested[#:style 'inset]{To prototype a library, you can use a file name
  (distinguished from installed libraries by using a text literal, i.e. double-quoted), which is
  searched for relative to the directory/folder of the file of the program that @racket[requires] it.
  Write in full @tt{#lang racket} and use @racket[provide] to export definitions you'd like to use.
 
  The name is restricted to avoid inconsistencies between platforms:
  it may contain only letters @litchar{a} through @litchar{z} (upper-case or lower-case),
  digits, @litchar{-}, @litchar{_}, and @litchar{/} (except at beginning or end),
  and @litchar{.} for file names.
 
  The import and export mechanism of racket is powerful and extensible (moreso than almost
  any other language). I've chosen two options to add that could cover many likely scenarios
  for the course:

  @itemlist[@item{importing only a subset of the exports of a library,
               optionally renaming some or all of those exports}
            @item{importing a whole library,
               renaming all of its exports by adding a chosen prefix.}]

  @form/none[#:literals (require only-in)
             (require (#,(defidform/inline only-in) library-name name-or-rename ...))]{
   Make only the definitions specified by @racket[name-or-rename ...] from a library
   available to use in the current program.
   A @racket[name-or-rename] may be of the form @racket[name new-name],
   in which case the @racket[name] is renamed to @racket[new-name] during import.}
                                                                        
  @form/none[#:literals (require prefix-in)
             (require (#,(defidform/inline prefix-in) prefix library-name))]{
   Make the definitions from a library available to use in the current program,
   but renaming all names during import by prefixing each one with @racket[prefix].}}}

@self{Describe for instructors the other direction: how to import the language's names
 into a library file.}
@self{@racketblock[(require (only-in racket take))
                   (require (only-in picuring-programs cut-left))
                   (require (only-in racket cake))
                   (require picturing-programs)
                   (require 2htdp/image)
                   (require lang/htdp-intermediate)]}
@;@§{Miscellaneous Removed for Now}
@instructor{We might want:
 @itemlist-inset[@item{@racket[error], if we start teaching manual contract checking}
                 @item{@racket[exit], for interactive projects,
                   but those are probably easily structured to end “normally”}]}
@;{@function[(error [v any/c] ...) void?]{
  Stop the program, reporting an error with a message combining the arguments.
  
  Long printed representations of the arguments are truncated and indicated with “...”.
  If the first value is a symbol it is suffixed with a colon.
  @interact[(define Zero (- 1 1))
            (cond [(zero? Zero) (error '/ "can't divide by 0")]
                  [else (/ 1 Zero)])]}
 @function[(exit) void?]{Stop the program.}}

@;{
 @§{Terminology}
 @itemlist[
 @item{Racket, DrRacket, CSC 104 language, Interactions area, Definitions area, prompt, Run}
 @item{evaluate, produce, compute, return, result, output}
 @item{expression, sub-expression, part}
 @item{function, operation, call, perform, input, argument}
 @item{comment, end-of-line comment, block comment, expression comment, comment box, comment out}
 @item{“name” “value” “representation” “literal”}
 @item{error is reported, test case, failed test case, actual, expected}
 @item{“data-type”, “type”}
 @item{“number” “image” “text” “boolean” “list”}
 @item{“definition” “parameter” “header” “body”}
 @item{“character” “integer” “natural” “non-negative” “sign”}
 @item{“quotient” “remainder” “floor” “ceiling” “base” “exponent” “numerator” “denominator”}
 @item{“odd” “even” “parity”}
 @item{“short-circuit” “conditional”}
 @item{“predicate”}
 @item{“condition” “clause” “question” “answer”}
 @item{“bracket” “parenthesis” “digit” “dobule-quote” “indent” “double-quote”}
 @item{“step” "intermediate step"}
 @item{“index” “element” “item” “member” “contain” “holds” “bundles” “compound” “atomic”}
 @item{“library”}
 @item{“empty list/text”}]
}