#lang racket/base #| Helpers for implementing DrRacket tools |#

(require (submod "contract.rkt" core))

(provide (contract-out
          [labels:definitions (list/c string? string?)] [labels:interactions (list/c string? string?)]
          [♯view-menu (any/c . → . (♯ (is-a?/c menu%)))] ; Safe on any/c, intended for text:basic<%>.
          [♯definitions-or-interactions-item ((is-a?/c menu-item-container<%>)
                                              #:definitions? boolean?
                                              . → . (♯ (is-a?/c labelled-menu-item<%>)))]
          [definitions-or-interactions-item-hidden? ((is-a?/c labelled-menu-item<%>) . → . boolean?)]
          
          ; Contracts use the most specific approximation, according to the Racket documentation,
          ;  of the actual classes, since this isn't in the drracket unit: should probably be a mixin.
          [language-menu
           ((is-a?/c menu-item-container<%>) . → . (♯ (is-a?/c labelled-menu-item<%>)))]
          [current-tab-next-settings
           ((is-a?/c frame% #;drracket:unit:frame<%>) . → . (not/c #false) #;language-settings)]
          [definitions
            ((is-a?/c frame% #;drracket:unit:frame<%>)
             . → . (is-a?/c text:basic<%>) #;drracket:unit:definitions-text<%>)]
          [interactions
           ((is-a?/c frame% #;drracket:unit:frame<%>)
            . → . (is-a?/c text:basic<%>) #;(is-a?/c racket:text%) #;drracket:rep:text%)]))

; interactions: broke its own contract
;   promised: (is-a?/c set-mode-mixin)
;   produced: (wrapper-object:boxed-highlight-mixin ...)
;   in: the range of
;       (->
;        (is-a?/c frame%)
;        (is-a?/c set-mode-mixin))
;   contract from: 
;       <pkgs>/csc104/snack/drracket.rkt
;   blaming: <pkgs>/csc104/snack/drracket.rkt
;    (assuming the contract is correct)
;   at: <pkgs>/csc104/snack/drracket.rkt:22.11


(require "string-constants.rkt"
         "menu.rkt" (only-in racket/gui/base frame%) (only-in framework racket:text% text:basic<%>)
         (submod "class.rkt" use) (submod "list.rkt" core) "false.rkt")


(define-values (labels:definitions labels:interactions)
  (values (list (string-constant show-definitions-menu-item-label)
                (string-constant hide-definitions-menu-item-label))
          (list (string-constant show-interactions-menu-item-label)
                (string-constant hide-interactions-menu-item-label))))
(define-values (labels:show labels:hide)
  (values (map first  (list labels:definitions labels:interactions))
          (map second (list labels:definitions labels:interactions))))

; get-top-level-window: editor:basic<%> method, for non-embedded editors,
;  (-> (or/c (is-a?/c frame%) (is-a?/c dialog%)))

; get-show-menu: drracket:frame:<%> method (-> is-a?/c menu%)

(define (♯view-menu an-editor)
  (and (is-a? an-editor text:basic<%>) ; Was getting error for Stepper window.
       (let*♯ ([frame-or-dialog (send an-editor get-top-level-window)])
              (and (is-a? frame-or-dialog frame%) (send frame-or-dialog get-show-menu)))))

(define (♯definitions-or-interactions-item view-menu #:definitions? definitions?)
  (♯item-by-label/s view-menu (if definitions? labels:definitions labels:interactions)))

(define (definitions-or-interactions-item-hidden? item) (member? (send item get-label) labels:show))

(define (language-menu bar) (app♯ ♯item-by-label/s bar (string-constant language-menu-name)))

(define (current-tab-next-settings a-drracket-frame)
  (send (definitions a-drracket-frame) get-next-settings))

; If we can get-current-tab, we can also get-interactions-text and get-definitions-text.
(define (definitions  a-drracket-frame) (send a-drracket-frame get-definitions-text))
(define (interactions a-drracket-frame) (send a-drracket-frame get-interactions-text))
