#lang racket/base #| Functions and Forms to Propagate False |#

(require (submod "contract.rkt" core))

(provide app♯ let*♯ let*/else case♯ λ-case♯ fail→false
         (contract-out [map♯ ((any/c . → . any/c) list? . → . (♯ list?))]))

#| (app♯ f a ...) produces false if any of f or a ... are false, otherwise calls f: (f a ....)

   (let*♯ ([id expr]
           ...)
    body ...+)

     Behaves like ‘let*’ but short-circuits producing false if any of the exprs produces are false.
     Instead of a body, the final clause can be [#:result expr], which at times looks nicer/clearer.

   (let*/else ([id expr]
               ...
               [#:else else-body ...])
     body ...+)

    Similar to ‘let*♯’, but short-circuits to the #:else clause instead of producing false.

   (fail→false body ...+) evaluates the body, short-circuiting with false an exn:fail is raised.

   (map♯ f l) is (map f l), but short-circuiting with false if f ever produces false. |#

(require (only-in "lambda.rkt" λ₁ λ-case♯)
         (submod "list.rkt" core)
         (submod "syntax.rkt" core))

(define-macro (app♯ f:expr a:expr ...)
  #:with (a′ ...) (generate-temporaries #'(a ...))
  (let*♯ ([f′ f] [a′ a] ... [#:result (f′ a′ ...)])))

(define-syntax-parser let*♯
  [(_ () body:expr ...+) #'(let () body ...)]
  [(_ ([#:result result:expr])) #'result]
  [(_ ([id₀:id expr₀:expr]) body ...+) #'(let ([id₀ expr₀]) (and id₀ (let () body ...)))]
  [(_ ([id₀:id expr₀:expr]
       [id:id expr:expr] ...
       [(~and id′ (~or* _:id #:result)) expr′:expr])
      body:expr ...)
   #'(let ([id₀ expr₀]) (and id₀ (let*♯ ([id expr] ... [id′ expr′]) body ...)))])

(define-macro (let*/else ([id₀:id expr₀:expr] [id:id expr:expr] ...
                                              [#:else else-body:expr ...+])
                         body:expr ...+)
  #:with (id- ... id′) #'(id₀ id ...)
  (let* ([id₀ expr₀] [id (and id- expr)] ...) (cond [id′ body ...] [else else-body ...])))

(define-macro (fail→false body:expr ...+) (with-handlers ([exn:fail? (λ₁ #false)]) body ...))

(define-macro (case♯ e:expr [(datum ...+) (~optional (~seq result:expr ...+))])
  (case e [(datum ...) (~? (~@ result ...) #true)] [else #false]))

(define (map♯ f l) (let loop ([l l] [fl '()])
                     (cond [(empty? l) fl]
                           [else (define fv (f (first l)))
                                 (and fv (loop (rest l) (adjoin-right fl fv)))])))
