#lang racket/base #| Miscellaneous Iteration |# (require (submod "contract.rkt" core))

#| ‘find-first’ and an extension of ‘for/first’

 ‘find-first’ produces the first element of a sequence that satisfies a predicate
    - it can be used with lists, although "list.rkt" exports a version for just lists

 ‘for/first’ : allows simply (for/first id sequence guard ...), for ‘find-first’-like contexts |#

(provide (rename-out [for/first′ for/first])
         (contract-out [find-first ((any/c . → . any/c) sequence? . → . any/c)]))

(require (submod "syntax.rkt" core))

(define (find-first f s) (for/first′ e s #:when (f e)))

(define-syntax-parser for/first′
  [(_ id:id sequence:expr (~seq (~and type (~or* #:when #:unless #:break #:final)) guard:expr) ...)
   #'(for/first ([id sequence] (~@ type guard) ...) id)]
  [(_ . parts) (syntax/loc this-syntax (for/first . parts))])

(require racket/sequence)

#| ‘define/for/fold’

 Version of ‘for/fold’ that also binds the accumulators. |#

(provide define/for/fold)

(define-macro (define/for/fold ([accum:id init:expr] ...+) (for-clause ...+)
                body-or-break ... body:expr)
  (define-values (accum ...) (for/fold ([accum init] ...) (for-clause ...)
                               body-or-break ... body)))

#| ‘in-overlap-pairs’

 (in-overlap-pairs (sequence a b c ...)) → (sequence (list a b) (list b c) ...) |#

(provide (contract-out [in-overlap-pairs (sequence? . -> . (sequence/c (list/c any/c any/c)))]))

(require racket/stream)

(define (in-overlap-pairs s)
  ; The sequence abstraction is annoying, especially if a sequence has side-effects.
  (define previous (void))
  (define has-previous? #false)
  (for/stream ([e s] #:when (or has-previous? (begin (set! has-previous? #true)
                                                     (set! previous e)
                                                     #false)))
    (define e- previous)
    (set! previous e)
    (list e- e)))

#| Extended version of typed ‘for/fold’ to allow the new ‘#:result id’ clause that untyped allows. |#
#;
(module typed typed/racket/base (provide (rename-out [for/fold′ for/fold]))
  
  (require (submod "syntax.rkt" core))

  (define-syntax-parser for/fold′
    [(_ ([accum:id spec:expr ...+] ... #:result result:id)
        (for-clause ...) body-or-break ... body:expr)
     #'(let-values ([(accum ...) (for/fold ([accum spec ...] ...)
                                           (for-clause ...) body-or-break ... body)])
         result)]
    [(_ ([accum:id spec:expr ...+] ...) (for-clause ...) body-or-break ... body:expr)
     #'(for/fold ([accum spec ...] ...) (for-clause ...) body-or-break ... body)]))

#| ‘until!’ : a simple loop until stateful condition, with no body |#

(provide until!)

(define-macro (until! condition:expr) (let loop () (unless condition (loop))))

