#lang racket/base

#| Forms for creating:

 1. A function of arity 0, 1, 2, or any, that:

      • ignores its (non-keyword) arguments, and
      • allows an empty body, defaulting to returning void

 2. A unary function with implicit parameter named ‘•’. |#

(provide (all-defined-out) (rename-out [case-lambda case-λ]))

(require (submod "syntax.rkt" core))
(define-simple-macro (λ₀ body:expr ...+) (λ ()  body ...))
(define-simple-macro (λ₁ body:expr ...+) (λ (_) body ...))
(define-simple-macro (λ₂ body:expr ...+) (λ (_₀ _₁) body ...))
(define-simple-macro (λ_ body:expr ...+) (λ _   body ...))
(define-syntax-parser λ• [(form body:expr ...+) #:with • (datum->syntax this-syntax '•)
                                                #'(λ (•) body ...)])
(define-simple-macro (λ-case♯ [(datum ...+)]) (λ• (case • [(datum ...) #true] [else #false])))
(define-simple-macro (λ-case [(datum:expr ...+) body:expr ...+] ...+
                             (~optional [(~literal else) else-body:expr ...]))
  (λ (v) (case v [(datum ...) body ...] ... (~? [else else-body ...]))))
(define-simple-macro (λ-case/identity [(datum:expr ...+) body:expr ...+] ...+)
  (λ (v) (case v [(datum ...) body ...] ... [else v])))
