#lang racket/base

(provide match match-define define/match λ-match
         λ-♯match ♯match λ-identity/match match/identity
         define/simple-match define/match₁ define/match₂)

(require (rename-in racket/match [match-lambda λ-match])
         (submod "syntax.rkt" core)
         (for-syntax racket/local (submod "syntax.rkt" syntax-classes))
         racket/local)

(define-syntax-parser define/match₁ 
  [(_ f:id (pattern result ...+) ...+) #'(define/match (f _) [(pattern) result ...] ...)]
  [(_ f:id [pattern result ...+] ...+ (~seq #:with definition:expr) ...+)
   #'(define f (local [definition ...] (λ-match [pattern result ...] ...)))])

(define-syntax-parser define/match₂ 
  [(_ f:id ([pattern₁ pattern₂] result ...+) ...+)
   #'(define/match (f _₁ _₂) [[pattern₁ pattern₂] result ...] ...)])

(define-macro (define/simple-match (f:id pattern ...+) body:expr ...+)
  (define f (match-lambda** [[pattern ...] body ...])))

(define-syntaxes (λ-identity/match match/identity λ-♯match ♯match)
  (values (syntax-parser [(_ clause ...+) #'(λ (v) (match/identity v clause ...))])
          (syntax-parser [(_ e:expr clause ...+) #`(match/derived e #,this-syntax clause ... [v v])])
          (syntax-parser [(_ clause ...+) #'(λ (v) (♯match v clause ...))])
          (local
            [(syntax-class (quoted-pattern ((~or* (~literal quote) (~literal quasiquote))
                                            . quoted-pattern)))
             (syntax-class (clauses #:splice {clause ...}
                                    #:with delegating-clause #'[e′ (♯match e′ clause ...)]))
             (define-splicing-syntax-class maybe-when (pattern (~optional (~seq #:when when:expr))))]
            (syntax-parser [(_ e:expr pattern:quoted-pattern clause ...)
                            #'(♯match e [pattern] clause ...)]
                           [(_ e:expr [pattern maybe-when:maybe-when result:expr ...] :clauses)
                            #`(match/derived e #,this-syntax
                                             [pattern #,@#'maybe-when #true result ...]
                                             delegating-clause)]
                           [(_ e:expr) #`(match/derived e #,this-syntax [_ #false])]))))


(module data racket/base
  (provide
   (all-defined-out)
   (rename-out [this-syntax this-data] [unsyntax undata] [syntax->datum data->datum]
               [data D] [unsyntax E]
               [define/syntax-parse define/data-parse])
   (except-out (all-from-out syntax/parse)
               this-syntax #;syntax #;unsyntax #;syntax->datum
               define/syntax-parse
               syntax-parse syntax-parser define-syntax-class))

  (require (submod "syntax.rkt" core) syntax/parse)

  (define-macro (♯data-parse d:expr (~optional (~seq #:literals [literal:id ...]))
                             [pattern:expr result ...] ...+)
    (syntax-parse d #:datum-literals (~? [literal ...] []) [pattern #true result ...] ... [_ #false]))
  (define-syntax-parser data-parse
    [(_ d:expr (~optional (~seq #:literals [literal:id ...])) [pattern:expr result ...+] ...+)
     #'(syntax-parse d #:datum-literals (~? [literal ...] []) [pattern result ...] ...)])
  (define-syntax-parser define-data-class
    [(_ id:id (~optional (~seq #:literals [literal:id ...])) . parts)
     #'(define-syntax-class id #:datum-literals (~? [literal ...] []) . parts)])

  (define-macro (with-data ([pattern:expr e:expr] ...) result:expr ...+)
    (data-parse (list e ...) [(pattern ...) result ...]))

  (define-syntax-parser with-data* 
    [(_ ([pattern₀:expr e₀:expr] [pattern:expr e:expr] ...) result:expr ...+)
     #'(with-data ([pattern₀ e₀]) (with-data* ([pattern e] ...) result ...))]
    [(_ () result:expr ...+) #'(let () result ...)])

  (define-macro (data template:expr) (syntax->datum (quasisyntax template)))
  (define-macro (define-data-parser header:expr . parts) (define (header d) (data-parse d . parts))))

(module expander racket/base (provide == (all-defined-out))

  (require (only-in racket/match define-match-expander ==)
           (submod "syntax.rkt" core)
           (for-syntax (only-in racket/syntax format-id)
                       (only-in racket/sequence in-syntax)))

  (define-macro (define-match-expander-parsers expander:id
                  [match-pattern-parse-clause ...]
                  [expression-parse-clause ...])
    (define-match-expander expander
      (syntax-parser match-pattern-parse-clause ...)
      (syntax-parser expression-parse-clause ...)))

  (define-macro (define-match-expander-parser expander:id parse-clause ...+)
    (define-match-expander expander (syntax-parser parse-clause ...)))

  (define-macro (define-simple-match-expander (expander:id . new-pattern:expr) pattern:expr)
    (define-match-expander-parser expander [(_ . new-pattern) #`pattern]))

  (define-syntax-parser : [(_ id:id ...+) #:with (id′ ...) (for/list ([id (in-syntax #'(id ...))])
                                                             (format-id id "~a:" id #:source id))
                                          #'(begin (define-match-expander-parsers id′
                                                     ([(_ . sub-forms) #'`(id . sub-forms)])
                                                     ([(_ . sub-forms) #'`(id . sub-forms)]))
                                                   ...)])

  (define-macro (catamorphism cat-id:id f:expr)
    (define-match-expander cat-id (syntax-parser [(_ an-id:id) #'(app f an-id)]))))
