#lang racket/base #| Intended to capture our uses of menus from racket/gui/base |#

(require (submod "contract.rkt" core))

(define string⒮/c (or/c string? (listof string?)))
(provide
 menu-item-container<%> selectable-menu-item<%> labelled-menu-item<%>
 menu% checkable-menu-item% separator-menu-item% menu-item%

 (contract-out [rename ♯item-by-label⒮ ♯item-by-label/s ((is-a?/c menu-item-container<%>)
                                                         string⒮/c
                                                         . → . (♯ (is-a?/c labelled-menu-item<%>)))])
 ensured-item
 boolean-preference-item%

 (contract-out
  [invoke-callback ((is-a?/c selectable-menu-item<%>) (is-a?/c event%) . → . void?)]
  [remove-shortcut ((is-a?/c menu-item-container<%>) string⒮/c . → . any)]
  [set-shortcut   ([(is-a?/c menu-item-container<%>) string⒮/c (♯ (or/c char? symbol?))]
                   [(♯ prefix/c)]
                   . →∗ . any)]))

(require "false.rkt" "iteration.rkt"
         "lambda.rkt"
         "boolean.rkt"
         (submod "list.rkt" core)
         (submod "class.rkt" implement)
         (only-in racket/gui/base
                  menu-item-container<%> selectable-menu-item<%> labelled-menu-item<%>
                  menu% checkable-menu-item% separator-menu-item% menu-item%
                  label->plain-label
                  event% control-event%)
         (submod "syntax.rkt" core))


(define tagged<%> (interface () tag-is?))
(define tagged (mixin () (tagged<%>) (init-field tag) (super-new)
                 (define/public (tag-is? a-tag) (equal? a-tag tag))))

#;(contracts [(tag→♯item (is-a?/c menu-item-container<%>) any/c) → (♯ (is-a?/c tagged<%>))])

(define (tag→♯item menu-container tag)
  (for/first item (send menu-container get-items) #:when (and (is-a? item tagged<%>)
                                                              (send item tag-is? tag))))

(define-macro (ensured-item the-parent:expr the-tag:expr the-label:expr
                            ((~literal new) %:expr [field:id initial:expr] ...))
  (or (tag→♯item the-parent the-tag)
      (new (tagged %) [tag the-tag] [parent the-parent] [label the-label] [field initial] ...)))


(define boolean-preference-item%
  (class checkable-menu-item% (init preference)
    (super-new [checked (preference)]
               [callback (λ (self _) (preference (send self is-checked?)))]
               [demand-callback (λ (self) (send self check (preference)))])))

(define (♯item-by-label⒮ menu-container label⒮)
  (define labels (flatten label⒮))
  (for/first item (send menu-container get-items)
    #:when (and (is-a? item labelled-menu-item<%>)
                (or (member (send item get-label) labels)
                    (member (send item get-plain-label) (map label->plain-label labels))))))

; Invoke menu-item% item's callback, using timestamp information from the event.
(define (invoke-callback item event)
  (send item command
        (new control-event% [event-type 'menu] [time-stamp (send event get-time-stamp)])))

(define prefix/c
  (and/c (listof (or/c 'alt 'cmd 'meta 'ctl 'shift 'option))
         (λ• ((equal? 'unix (system-type)) . ⇒ . (not (and (member 'alt •)
                                                           (member 'meta •)))))
         (λ• (equal? • (remove-duplicates •)))))

(define (remove-shortcut menu-bar label⒮) (set-shortcut menu-bar label⒮ #false))
;
; Set shortcut for a menu item by label: the first item matching a label in ‘label⒮’.
; When prefix, set a shortcut prefix.
(define (set-shortcut menu-bar label⒮ shortcut [prefix #false])
  (let*♯ ([menu-item (♯item-by-label⒮ menu-bar label⒮)])
         (send menu-item set-shortcut shortcut)
         (when prefix (send menu-item set-shortcut-prefix prefix))))
