#lang racket/base #| Initial Replacement for racket/stream |#

#| The core currently used by the step interpreter.

  With some fancier names [e.g. using \rightarrow and \ast], but more significantly:

    • a lazier append operation

    • options to create with specified parameters' [a la parameterize] values captured,
       since streams only capture the lexical environment, unlike generators |#

(require (only-in (submod "contract.rkt" core) contract-out → any/c and/c #;contracts natural/c))

(provide stream?
         (rename-out [stream′ stream]) stream-adjoin stream-join
         stream-first 
         (contract-out [stream-second ((stream-length-at-least/c 2) . → . any/c)])
         (rename-out [stream->list stream→list]))

(require (only-in racket/stream
                  stream? stream-empty?
                  stream stream*
                  stream-first stream-rest
                  stream->list)
         (submod "syntax.rkt" core)
         "boolean.rkt")

#;(contracts [stream-length-at-least? (stream? natural/c . → . boolean?)])

(begin-for-syntax (define-splicing-syntax-class parameters
                    (pattern (~optional (~seq #:parameters [parameter:id ...+]))
                             #:with (parameter′ ...) #'(~? (parameter ...) ())
                             #:with (captured ...) (generate-temporaries #'(parameter′ ...))
                             #:with captures #'([captured (parameter′)] ...)
                             #:with restores #'([parameter′ captured] ...))))

(define-macro (stream′ parameters:parameters element:expr ...)
  (let parameters.captures (stream (parameterize parameters.restores element) ...)))

(define-macro (stream-adjoin parameters:parameters element:expr ...+ tail:expr)
  (let parameters.captures (stream* (parameterize parameters.restores element)
                                        ...
                                        (parameterize parameters.restores tail))))

(define-macro (stream-join s₀:expr parameters:parameters s₁:expr)
  (stream-join′ s₀ (let parameters.captures (λ () (parameterize parameters.restores s₁)))))

(define (stream-join′ s₀ s₁) (if (stream-empty? s₀)
                                 (s₁)
                                 (stream-adjoin (stream-first s₀)
                                                (stream-join′ (stream-rest s₀) s₁))))

(define (stream-second s) (stream-first (stream-rest s)))

(define (stream-length-at-least/c n)
  (define ((check n) s) ((positive? n)
                         . ⇒ . ((not (stream-empty? s)) . and . ((check (sub1 n)) (stream-rest s)))))
  (and/c stream? (check n)))
