#lang racket/base

(module core racket/base (provide (all-defined-out)
                                  (all-from-out syntax/parse/define 'implementation)
                                  (rename-out [define-simple-macro define-macro])
                                  (for-syntax (all-from-out racket/base 'implementation)
                                              (all-defined-out))
                                  (all-from-out syntax/stx))
  
  (require (only-in racket/splicing splicing-local)
           syntax/parse/define syntax/parse
           (for-syntax racket/base)
           (only-in syntax/stx stx->list stx-map stx-null?))

  (module implementation racket/base (provide ♯syntax-parser ♯syntax-parse)
    (require syntax/parse)
    (define-syntax-rule (♯syntax-parser part ...) (syntax-parser part ... [_ #false]))
    (define-syntax-rule (♯syntax-parse  part ...) (syntax-parse  part ... [_ #false])))

  (require 'implementation (for-syntax 'implementation))

  (define-simple-macro (syntax-parser∗ . parts) (λ args (syntax-parse args . parts)))


  (define (lexical? id) (equal? 'lexical (identifier-binding id)))
  
  (define (distinct? stx) (not (check-duplicate-identifier (stx->list stx))))
  (define (temporary id) (car (temporary/s id)))
  (define (temporary/s id/s) (generate-temporaries (if (identifier? id/s)
                                                       (datum->syntax #false (list id/s))
                                                       (syntax->list id/s))))
  
  (define-for-syntax (relocate stx to) (datum->syntax stx (syntax-e stx) to))

  (define-simple-macro (map-macro macro:id (~and parts [part ...]) ...)
    #:with (one ...) (for/list ([a-parts (syntax->list #'(parts ...))])
                       (relocate #`(macro . #,a-parts) a-parts))
    (begin one ...))

  (define-simple-macro (generate {data
                                  ...+}
                                 pattern
                                 result
                                 ...+)
    (splicing-local [(define-simple-macro (m . pattern)
                       result
                       ...)]
      [m . data]
      ...)))

(require 'core)
(provide (all-from-out 'core))

(module syntax-classes/core racket/base
  (provide (all-defined-out)
           (except-out (all-from-out syntax/parse) ~or)
           (rename-out [attribute α] [pattern ρ] [free-identifier=? alias?]))
  (require syntax/parse syntax/parse/define (for-syntax racket/base))

  (define-simple-macro (define-pattern id:id clause:expr ...+)
    (define-syntax id (pattern-expander (syntax-parser clause ...))))

  (define-pattern ~maybe [(_ pat:expr) #'(~commit (~or* pat _))])

  (define-pattern ~same?
    [(_ e:expr) #'(~same? _ e)]
    [(_ p:expr e:expr) #'(~and p this (~fail #:unless (equal? (syntax-e #'this) (attribute e))))])

  (define-pattern ~true
    [(_ e:expr) #'(~fail #:unless e)]
    [(_ e:expr pattern:expr) #'(~and (~true e) pattern)])
  (define-pattern ~is   
    [(_ sc:expr) #'(~var _ sc)]
    [(_ e:expr sc:expr) #'(~parse (~is sc) e)])
  (define-pattern ~is-not   
    [(_ sc:expr) #'(~not (~var _ sc))]
    [(_ e:expr sc:expr) #'(~parse (~is-not sc) e)])

  ; See also (~literal literal-id [#:phase phase-expr])
  (define-pattern ~binding?
    [(_ id/symbol:expr) #'(~binding? _ id/symbol)]
    [(_ pat:expr id/symbol:expr) #'(~and an-id:id pat (~true (binding? #'an-id id/symbol)))])
  ;
  (define (binding? id id/symbol)
    (or (and (identifier? id/symbol)
             (free-identifier=? id id/symbol 0 (if (syntax-transforming?) 0 -1)))
        (and (symbol?     id/symbol) (equal? (syntax-e id) id/symbol))))

  (define-pattern ~capture [(_ literal:id as:id) #'(~and (~literal literal) (~bind [as literal]))])

  (define (syn v) (datum->syntax #false v))
  (define (slist . args) (syn args))
  (define (slist* . args) (syn (apply list* args)))
    
  (define-syntax-parser syntax-class
    [(_ (header #:splice pat directive ...))
     #'(define-splicing-syntax-class header (pattern (~seq . pat) directive ...))]
    [(_ (header pat directive ...)) #'(define-syntax-class header (pattern pat directive ...))]))

(module syntax-classes racket/base (provide (all-defined-out) (all-from-out syntax/parse syntax/stx))
  
  (require (except-in (combine-in syntax/parse syntax/parse/define) ~or)
           (only-in syntax/stx stx->list stx-map)
           (for-syntax racket/base))

  (define-syntax-parser syntax-class
    [(_ (header #:splice pat directive ...))
     #'(define-splicing-syntax-class header (pattern (~seq . pat) directive ...))]
    [(_ (header pat directive ...)) #'(define-syntax-class header (pattern pat directive ...))])

  (syntax-class (variadic-formals (_:id parameter:id ... . parameters:id)
                                  #:with formals   #'(parameter ... . parameters)
                                  #:with arguments #'(parameter ... parameters)))

  (syntax-class (sequence (_ ...) #:attr length (length (syntax->list this-syntax))))

  (define-syntax ~multiple
    (pattern-expander
     (syntax-parser
       [(_ parts-name pattern) #'(~and (~describe #:opaque (multiple-description parts-name)
                                                  _:multiple)
                                       (~seq pattern ...+))]
       [(_ pattern) #'(~and _:multiple (~seq pattern ...+))])))

  (define (multiple-description [parts-name "parts"])
    (format "at least two ~a, as a sanity check" parts-name))

  (define-splicing-syntax-class multiple #:description (multiple-description)
    (pattern (~seq _ _ ...+)))

  (define-syntax ~bracketed
    (pattern-expander
     (syntax-parser
       [[_ #:name part-name . pattern] #'(~and (~describe #:opaque (bracketed-description part-name)
                                                          _:bracketed)
                                               pattern)]
       [[_ . pattern] #'(~and _:bracketed pattern)])))

  (define (bracketed-description [part-name "parts"])
    (format "square bracketed ~a, as a sanity check" part-name))

  (define-syntax-class bracketed #:description (bracketed-description)
    (pattern _ #:when (equal? (syntax-property this-syntax 'paren-shape) #\[))))
