#lang racket/base (provide compthink-ui@)

(require (submod "../configuration.rkt" ui)
         (submod "../shared-settings.rkt" step)
         "log.rkt"
         "../snack/drracket.rkt"
         "../snack/menu.rkt"
         "../snack/button.rkt"
         (except-in "../snack/false.rkt" λ-case♯)
         "../snack/lambda.rkt"
         (submod "../snack/list.rkt" core)
         "../snack/function.rkt"
         "../snack/string.rkt"
         (submod "../snack/class.rkt" implement)
         drracket/tool racket/unit)

(module actions racket/base (provide (all-defined-out))

  (require (only-in racket/gui/base event% editor<%> control-event%)
           racket/snip images/flomap
           "space-out.rkt"
           "../snack/drracket.rkt"
           "../snack/menu.rkt"
           (except-in "../snack/false.rkt" λ-case♯)
           "../snack/lambda.rkt"
           (submod "../snack/class.rkt" use)
           "../snack/string-constants.rkt")
  
  (define (menu-event) (new control-event% [event-type 'menu]))

  (define (focus-shrink-image focused)
    (define here (send focused get-start-position))
    (define snip₀ (send focused find-snip here 'after #false))
    (define snip₁ (cond [(is-a? snip₀ image-snip%) snip₀]
                        [else (send focused find-snip here 'before #false)]))
    (when (is-a? snip₁ image-snip%)
      (define ♯bitmap₀ (send snip₁ get-bitmap))
      (when ♯bitmap₀
        (define fm₀ (bitmap->flomap ♯bitmap₀ #:unscaled? #true))
        (define fm₁ (flomap-scale fm₀ (sqrt 1/2)))
        (define bitmap₁ (flomap->bitmap fm₁ #:backing-scale 1.))
        (send snip₁ set-bitmap bitmap₁))))
  
  (define ((focus-keymap-command name) focused)
    ; Could call our custom functions instead.
    (define k (and (is-a? focused editor<%>) (send focused get-keymap)))
    (and k (send k call-function name focused (new event%) #true)))

  (define ((focus-help keymap-thunk) focused)
    (send (keymap-thunk) ; keymap that adds the function
          call-function "search-help-desk"
          ; checks it's editor<%> via drracket/drracket/private/rep.rkt get-frame
          focused
          (new event%) ; call-function needs an event%
          ; Not chaining prevents some exploits?
          #false))

  (define (action:insert-image this)
    (λ₁ (let*♯ ([insert-image-item (♯item-by-label/s (send this get-insert-menu) ; menu%
                                                     (string-constant insert-image-item))])
               (send insert-image-item command (menu-event)))))
  
  (define ((focus-history this up?) focused)
    (when (eq? focused (interactions this))
      (send (send focused get-keymap) call-function
            (if up? "put-previous-sexp" "put-next-sexp")
            focused (new event%) #true)))

  (define (action:switch-layout this)
    (λ₁ (let*♯ ([layout-item (♯item-by-label/s (send this get-show-menu) ; menu%
                                               (list (string-constant use-horizontal-layout)
                                                     (string-constant   use-vertical-layout)))])
               (send layout-item command (menu-event)))))
      
  (define (action:toggle-record this) (λ₁ (send (send this get-interactions-text) toggle-record)))

  (define ((focus-format this) focused)
    (cond [(eq? focused (interactions this)) (send focused tabify-all)]
          [else (define the-definitions (definitions this))
                (send the-definitions tabify-all)
                (fine-format the-definitions)])))
;
(require 'actions)

(define-unit compthink-ui@ (import drracket:tool^) (export)

  (define (our-lang? this) (equal? configuration:language-name
                                   (send (drracket:language-configuration:language-settings-language
                                          (current-tab-next-settings this))
                                         get-language-name)))

  (define frame-mixin
    (mixin (drracket:unit:frame<%>) () (super-new)
      (inherit get-menu-bar get-button-panel register-toolbar-buttons get-focus-object)

      (define/override (on-activate active?)
        (define (log s) (log-compthink-error (~a␣ "language-ui:frame-mixin:on-activate:" s)))
        (let*/else
         ([Language-menu (language-menu (get-menu-bar))] [#:else (log "no Language menu")])
         (define compthink-menu (ensured-item Language-menu 'our-language
                                              configuration:language-settings-label (new menu%)))
         (for ([(preference label) (in-dict step:labels)])
           (ensured-item compthink-menu (string->symbol label) label
                         (new boolean-preference-item% [preference preference]))))
        (super on-activate active?))

      (define (for-focus handler) (λ₁ (handler (get-focus-object))))
      (define (keymap-command name) (for-focus (focus-keymap-command name)))
      (define button-specs
        (list (list configuration:view:help         (for-focus
                                                     (focus-help
                                                      drracket:rep:get-drs-bindings-keymap)))
              (list configuration:view:shrink       (for-focus focus-shrink-image))
              (list configuration:view:line-comment (keymap-command
                                                     "line comment/uncomment around selection"))
              (list configuration:view:explode      (keymap-command "explode"))
              (list configuration:view:format       (for-focus (focus-format this)))
              (list configuration:view:insert-image (action:insert-image this))
              (list configuration:view:hug          (keymap-command "try hug±"))
              (list configuration:view:definitions  (keymap-command "show more of definitions"))
              (list configuration:view:layout       (action:switch-layout this))
              (list configuration:view:record       (action:toggle-record this))
              (list configuration:view:interactions (keymap-command "show more of interactions"))
              (list configuration:view:history-up   (for-focus (focus-history this #true)))
              (list configuration:view:history-down (for-focus (focus-history this #false)))))
      
      (define toolbar (get-button-panel))
      
      (container-edit-sequence
       toolbar
       (define buttons (map (curry apply (button toolbar)) button-specs))       
       ; For sorting by #:numbers, vertical vs horizontal toolbar layout, small vs large, show label.
       (register-toolbar-buttons buttons #:numbers (map (∘ view-position first) button-specs))
       
       (define/public (show/hide-compthink-buttons)
         (container-edit-sequence (get-button-panel)
                                  (for-each (if (our-lang? this) show-button hide-button) buttons)))
       (define/augment (on-tab-change old new)
         (show/hide-compthink-buttons)
         (inner (void) on-tab-change old new))
       (show/hide-compthink-buttons))))
  
  (define (definitions-text-mixin %) ; Refresh show/hide state of buttons when language changes.
    (class % (inherit get-top-level-window) (super-new)
      (define/augment (after-set-next-settings settings)
        (define top-level-window (get-top-level-window))
        (cond [top-level-window
               (container-edit-sequence (send top-level-window get-button-panel)
                                        (send top-level-window show/hide-compthink-buttons)
                                        (inner (void) after-set-next-settings settings)
                                        (send top-level-window sort-toolbar-buttons-panel))]
              [else (inner (void) after-set-next-settings settings)]))))

  (drracket:get/extend:extend-unit-frame frame-mixin)
  (drracket:get/extend:extend-definitions-text definitions-text-mixin))
